/*
 *  This software may only be used by you under license from AT&T Corp.
 *  ("AT&T").  A copy of AT&T's Source Code Agreement is available at
 *  AT&T's Internet website having the URL:
 *  <http://www.research.att.com/sw/tools/graphviz/license/source.html>
 *  If you received this software without first entering into a license
 *  with AT&T, you have an infringing copy of this software and cannot use
 *  it without violating AT&T's intellectual property rights.
 */

package att.grappa;
import java.io.*;
import java.util.*;

class PrintfParser

    implements GrappaConstants

{

    private boolean		alternate;
    private boolean		rightpad;
    private boolean		sign;
    private boolean		space;
    private boolean		zeropad;
    private boolean		trim;
    private int			precision;
    private int			width;
    private String		plus;
    private char		padding;
    private StringBuffer	scratch;

    ///////////////////////////////////////////////////////////////////////////
    //
    // Constructor
    //
    ///////////////////////////////////////////////////////////////////////////

    PrintfParser() {

	scratch = new StringBuffer();
    }

    ///////////////////////////////////////////////////////////////////////////
    //
    // printfParser
    //
    ///////////////////////////////////////////////////////////////////////////

    final int
    parse(char cfmt[]) {

	return(parse(cfmt, 0));
    }

    ///////////////////////////////////////////////////////////////////////////

    final int
    parse(char cfmt[], int n) {

	boolean	done;
	int	ch;

    //
    // Parse the conversion specification that starts at index n
    // in fmt.  Results are stored in the class variables and the
    // position of the character that stopped the parse is
    // returned to the caller.
    //

	alternate = false;
	rightpad = false;
	sign = false;
	space = false;
	zeropad = false;
	trim = false;

	for (done = false; n < cfmt.length && !done; n++) {
	    switch (cfmt[n]) {
		case '-': rightpad = true; break;
		case '+': sign = true; break;
		case ' ': space = true; break;
		case '0': zeropad = true; break;
		case '#': alternate = true; break;
		default: done = true; n--; break;
	    }
	}

	plus = (sign ? "+" : (space ? " " : ""));

	for (width = 0; n < cfmt.length && GrappaSupport.isdigit(ch = cfmt[n]); n++)
	    width = width*10 + (ch - '0');

	if (n < cfmt.length && cfmt[n] == '.') {
	    n++;
	    for (precision = 0; n < cfmt.length && GrappaSupport.isdigit(ch = cfmt[n]); n++)
		precision = precision*10 + (ch - '0');
	} else precision = -1;

	padding = (zeropad && !rightpad) ? '0' : ' ';

	return(n);
    }

    ///////////////////////////////////////////////////////////////////////////

    final StringBuffer
    buildChar(StringBuffer buf, int arg) {

	scratch.setLength(0);
	scratch.append((char)arg);
	return(strpad(buf, scratch.toString(), ' ', width, rightpad));
    }

    ///////////////////////////////////////////////////////////////////////////

    final StringBuffer
    buildExp(StringBuffer buf, double arg, boolean upper) {

	double		exp;
	double		base;
	double		val;
	int		sign;

	precision = (precision >= 0) ? precision : 6;

	val = arg;
	sign = (val >= 0) ? 1 : -1;
	val = (val < 0) ? -val : val;

	if (val >= 1) {
	    exp = Math.log(val)/LOG10;
	    base = Math.pow(10, exp - (int)exp);
	} else {
	    exp = Math.log(val/10)/LOG10;
	    base = Math.pow(10, exp - (int)exp + 1);
	}

	scratch.setLength(0);
	scratch.append(upper ? "E" : "e");
	scratch.append(exp > 0 ? '+' : '-');

	strpad(scratch, ("" + (int)(exp > 0 ? exp : -exp)), '0', 2, false);
	if (padding == '0' && precision >= 0)
	    padding = ' ';

	return(strpad(buf, doubleToString(sign*base, scratch.toString()), padding, width, rightpad));
    }

    ///////////////////////////////////////////////////////////////////////////

    final StringBuffer
    buildFlex(StringBuffer buf, double arg, boolean upper) {

	double		exp;
	double		val;
	double		ival;
	StringBuffer	retbuf;
	int		iexp;
	int		pr;

	trim = true;

	val = arg;
	ival = (int)arg;
	val = (val < 0) ? -val : val;


	if (val >= 1) {
	    exp = Math.log(val)/LOG10;
	} else {
	    exp = Math.log(val/10)/LOG10;
	}

	iexp = (int)exp;
	precision = (precision >= 0) ? --precision : 5;



	if (val == ival) {
	    if (alternate) {
		if (precision < 0 || iexp <= precision) {
		    precision -= iexp;
		    retbuf = buildFloat(buf, arg);
		} else retbuf = buildExp(buf, arg, upper);
	    } else {
		if (precision < 0 || iexp <= precision) {
		    precision = -1;
		    retbuf = buildInteger(buf, (int)arg);
		} else retbuf = buildExp(buf, arg, upper);
	    }
	} else if (iexp < -4 || iexp > precision)
	    retbuf = buildExp(buf, arg, upper);
	else retbuf = buildFloat(buf, arg);

	return(retbuf);
    }

    ///////////////////////////////////////////////////////////////////////////

    final StringBuffer
    buildPoint(StringBuffer buf, java.awt.geom.Point2D parg, boolean upper) {

	double[]		arg = { 0, 0 };
	double[]		exp = { 0, 0 };
	double[]		val = { 0, 0 };
	double[]		ival = { 0, 0 };
	int[]			iexp = { 0, 0 };
	StringBuffer		retbuf = null;
	int			orig_precision;
	int			pr;


	trim = true;

	arg[0] = parg.getX();
	arg[1] = (Grappa.negateStringYCoord?-parg.getY():parg.getY());
	val[0] = arg[0];
	val[1] = arg[1];
	orig_precision = precision;

	for(int i=0; i<2; i++) {
	    precision = orig_precision;
	    ival[i] = (int)val[i];
	    val[i] = (val[i] < 0) ? -val[i] : val[i];

	    if (val[i] >= 1) {
		exp[i] = Math.log(val[i])/LOG10;
	    } else {
		exp[i] = Math.log(val[i]/10)/LOG10;
	    }

	    iexp[i] = (int)exp[i];
	    precision = (precision >= 0) ? --precision : 5;

	    if (val[i] == ival[i]) {
		if (alternate) {
		    if (precision < 0 || iexp[i] <= precision) {
			precision -= iexp[i];
			retbuf = buildFloat(buf, arg[i]);
		    } else retbuf = buildExp(buf, arg[i], upper);
		} else {
		    if (precision < 0 || iexp[i] <= precision) {
			precision = -1;
			retbuf = buildInteger(buf, (long)arg[i]);
		    } else retbuf = buildExp(buf, arg[i], upper);
		}
	    } else if (iexp[i] < -4 || iexp[i] > precision)
		retbuf = buildExp(buf, arg[i], upper);
	    else retbuf = buildFloat(buf, arg[i]);

	    if(i == 0) {
		retbuf = retbuf.append(',');
		buf = retbuf;
	    }
	}

	return(retbuf);
    }

    ///////////////////////////////////////////////////////////////////////////

    final StringBuffer
    buildSize(StringBuffer buf, java.awt.geom.Dimension2D parg, boolean upper) {

	double[]		arg = { 0, 0 };
	double[]		exp = { 0, 0 };
	double[]		val = { 0, 0 };
	double[]		ival = { 0, 0 };
	int[]			iexp = { 0, 0 };
	StringBuffer		retbuf = null;
	int			orig_precision;
	int			pr;


	trim = true;

	arg[0] = parg.getWidth();
	arg[1] = parg.getHeight();
	val[0] = arg[0];
	val[1] = arg[1];
	orig_precision = precision;

	for(int i=0; i<2; i++) {
	    precision = orig_precision;
	    ival[i] = (int)val[i];
	    val[i] = (val[i] < 0) ? -val[i] : val[i];

	    if (val[i] >= 1) {
		exp[i] = Math.log(val[i])/LOG10;
	    } else {
		exp[i] = Math.log(val[i]/10)/LOG10;
	    }

	    iexp[i] = (int)exp[i];
	    precision = (precision >= 0) ? --precision : 5;

	    if (val[i] == ival[i]) {
		if (alternate) {
		    if (precision < 0 || iexp[i] <= precision) {
			precision -= iexp[i];
			retbuf = buildFloat(buf, arg[i]);
		    } else retbuf = buildExp(buf, arg[i], upper);
		} else {
		    if (precision < 0 || iexp[i] <= precision) {
			precision = -1;
			retbuf = buildInteger(buf, (long)arg[i]);
		    } else retbuf = buildExp(buf, arg[i], upper);
		}
	    } else if (iexp[i] < -4 || iexp[i] > precision)
		retbuf = buildExp(buf, arg[i], upper);
	    else retbuf = buildFloat(buf, arg[i]);

	    if(i == 0) {
		retbuf = retbuf.append(',');
		buf = retbuf;
	    }
	}

	return(retbuf);
    }

    ///////////////////////////////////////////////////////////////////////////

    final StringBuffer
    buildBox(StringBuffer buf, java.awt.geom.Rectangle2D parg, boolean upper, boolean dimensioned) {

	double[]		arg = { 0, 0, 0, 0 };
	double[]		exp = { 0, 0, 0, 0 };
	double[]		val = { 0, 0, 0, 0 };
	double[]		ival = { 0, 0, 0, 0 };
	int[]			iexp = { 0, 0, 0, 0 };
	StringBuffer		retbuf = null;
	int			orig_precision;
	int			pr;


	trim = true;

	if(!dimensioned) {
	    arg[0] = parg.getX();
	    arg[1] = parg.getY();
	    arg[2] = arg[0] + arg[2];
	    arg[3] = arg[1] + arg[3];
	    arg[1] = (Grappa.negateStringYCoord?-arg[1]:arg[1]);
	    arg[3] = (Grappa.negateStringYCoord?-arg[3]:arg[3]);
	} else {
	    arg[0] = parg.getX();
	    arg[1] = (Grappa.negateStringYCoord?-parg.getY():parg.getY());
	    arg[2] = parg.getWidth();
	    arg[3] = parg.getHeight();
	}
	val[0] = arg[0];
	val[1] = arg[1];
	val[2] = arg[2];
	val[3] = arg[3];
	orig_precision = precision;

	for(int i=0; i<4; i++) {
	    precision = orig_precision;
	    ival[i] = (int)val[i];
	    val[i] = (val[i] < 0) ? -val[i] : val[i];

	    if (val[i] >= 1) {
		exp[i] = Math.log(val[i])/LOG10;
	    } else {
		exp[i] = Math.log(val[i]/10)/LOG10;
	    }

	    iexp[i] = (int)exp[i];
	    precision = (precision >= 0) ? --precision : 5;

	    if (val[i] == ival[i]) {
		if (alternate) {
		    if (precision < 0 || iexp[i] <= precision) {
			precision -= iexp[i];
			retbuf = buildFloat(buf, arg[i]);
		    } else retbuf = buildExp(buf, arg[i], upper);
		} else {
		    if (precision < 0 || iexp[i] <= precision) {
			precision = -1;
			retbuf = buildInteger(buf, (long)arg[i]);
		    } else retbuf = buildExp(buf, arg[i], upper);
		}
	    } else if (iexp[i] < -4 || iexp[i] > precision)
		retbuf = buildExp(buf, arg[i], upper);
	    else retbuf = buildFloat(buf, arg[i]);

	    if(i < 3) {
		retbuf = retbuf.append(',');
		buf = retbuf;
	    }
	}

	return(retbuf);
    }

    ///////////////////////////////////////////////////////////////////////////

    final StringBuffer
    buildFloat(StringBuffer buf, double arg) {

	double	val;
	int	sign;

	precision = (precision >= 0) ? precision : 6;
	val = arg;

	if (padding == '0' && precision >= 0)
	    padding = ' ';
	return(strpad(buf, doubleToString(val, ""), padding, width, rightpad));
    }

    ///////////////////////////////////////////////////////////////////////////

    final StringBuffer
    buildHex(StringBuffer buf, int arg, boolean upper) {

	String	str;

	scratch.setLength(0);

	str = (upper) ? Integer.toHexString(arg).toUpperCase() : Integer.toHexString(arg);

	if (precision > str.length()) {
	    if (alternate)
		scratch.append(upper ? "0X" : "0x");
	    strpad(scratch, str, '0', precision, false);
	    strpad(buf, scratch.toString(), ' ', width, rightpad);
	} else {
	    if (zeropad && !rightpad && precision < 0) {
		if (alternate) {
		    if (width > 2) {
			strpad(scratch, str, '0', width-2, rightpad);
			buf.append(upper ? "0X" : "0x");
			buf.append(scratch.toString());
		    } else {
			buf.append(upper ? "0X" : "0x");
			buf.append(str);
		    }
		} else strpad(buf, str, '0', width, rightpad);
	    } else {
		if (alternate) {
		    scratch.append(upper ? "0X" : "0x");
		    scratch.append(str);
		    str = scratch.toString();
		}
		strpad(buf, str, ' ', width, rightpad);
	    }
	}

	return(buf);
    }

    ///////////////////////////////////////////////////////////////////////////

    final StringBuffer
    buildInteger(StringBuffer buf, long arg) {

	String	str;
	String	sign;
	long	val;

	scratch.setLength(0);

	val = arg;
	sign = (val >= 0) ? plus : "-";
	str = "" + ((val < 0) ? -val : val);

	if (precision > str.length()) {
	    strpad(scratch, str, '0', precision, false);
	    scratch.insert(0, sign);
	} else {
	    scratch.append(sign);
	    scratch.append(str);
	}

	if (padding == '0' && precision >= 0)
	    padding = ' ';

	return(strpad(buf, scratch.toString(), padding, width, rightpad));
    }

    ///////////////////////////////////////////////////////////////////////////

    final StringBuffer
    buildOctal(StringBuffer buf, int arg) {

	String	str;

	scratch.setLength(0);

	if (alternate)
	    scratch.append('0');

	scratch.append(Integer.toOctalString(arg));
	if (precision > scratch.length()) {
	    str = scratch.toString();
	    scratch.setLength(0);
	    strpad(scratch, str, '0', precision, false);
	}

	if (padding == '0' && precision >= 0)
	    padding = ' ';

	return(strpad(buf, scratch.toString(), padding, width, rightpad));
    }

    ///////////////////////////////////////////////////////////////////////////

    final StringBuffer
    buildString(StringBuffer buf, String arg) {

	String	str;

	if (precision > 0) {
	    if (precision < arg.length())
		str = arg.substring(0, precision);
	    else str = arg;
	} else str = arg;

	return(strpad(buf, str, padding, width, rightpad));
    }

    ///////////////////////////////////////////////////////////////////////////
    //
    // Private methods
    //
    ///////////////////////////////////////////////////////////////////////////

    private String
    doubleToString(double val, String exp) {

	String	sign;
	double	whole;
	double	power;
	double	frac;

    //
    // Building the resulting String up by casting to an int or long
    // doesn't always work, so we use algorithm that may look harder
    // and slower than necessary.
    //

	scratch.setLength(0);

	sign = (val >= 0) ? plus : "-";
	val = (val < 0) ? -val : val;

	whole = Math.floor(val);

	if (precision != 0) {
	    power = Math.pow(10, precision);
	    frac = (val - whole)*power;
	    scratch.append((long)whole);
	    String tail = (""+((long)Math.round(frac)));
	    if(trim) {
		int len = tail.length();
		int extra = 0;
		while(extra < len && tail.charAt(len-extra-1) == '0') extra++;
		if(extra == len) {
		    if(exp.length() > 0) {
			tail = ".0";
		    } else {
			tail = "";
		    }
		    precision = 0;
		} else if(extra > 0) {
		    scratch.append('.');
		    tail = tail.substring(0,len-extra);
		    precision -= extra;
		} else {
		    scratch.append('.');
		}
	    } else {
		scratch.append('.');
	    }
	    if (precision > 0 && (power/10) > frac) {
		strpad(scratch, tail, '0', precision, false);
	    } else scratch.append(tail);
	    scratch.append(exp);
	} else {
	    scratch.append((long)whole);
	    if (alternate && exp.length() == 0)
		scratch.append('.');
	    scratch.append(exp);
	}

	if (zeropad && !rightpad) {
	    String str = scratch.toString();
	    scratch.setLength(0);
	    strpad(scratch, str, '0', width - sign.length(), false);
	}

	scratch.insert(0, sign);
	return(scratch.toString());
    }

    ///////////////////////////////////////////////////////////////////////////

    private StringBuffer
    strpad(StringBuffer buf, String str, int ch, int width, boolean right) {

	int	len;
	int	n;

	if (width > 0) {
	    if ((len = width - str.length()) > 0) {
		if (right)
		    buf.append(str);
		for (n = 0; n < len; n++)
		    buf.append((char)ch);
		if (!right)
		    buf.append(str);
	    } else buf.append(str);
	} else buf.append(str);

	return(buf);
    }

    ///////////////////////////////////////////////////////////////////////////
}

