
/*
 *  This software may only be used by you under license from AT&T Corp.
 *  ("AT&T").  A copy of AT&T's Source Code Agreement is available at
 *  AT&T's Internet website having the URL:
 *  <http://www.research.att.com/sw/tools/graphviz/license/source.html>
 *  If you received this software without first entering into a license
 *  with AT&T, you have an infringing copy of this software and cannot use
 *  it without violating AT&T's intellectual property rights.
 */

package att.grappa;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;

class TableField
    implements GrappaConstants
{
    private Dimension size = new Dimension();
    private Rectangle bounds = new Rectangle();
    private Rectangle textBounds = null;
    private TableField[] subFields = null;
    private int subFieldsUsed = 0;
    private boolean orientLR = false;
    private String idTag = null;
    private String text = null;
    private TableField parent = null;

    /**
     * Creates an empty <code>TableField</code> instance.
     */
    TableField() {
	//super();
    }

    void setParent(TableField prnt) {
	parent = prnt;
    }

    TableField getTopMost() {
	TableField topper = this;
	while(topper.getParent() != null)
	    topper = topper.getParent();
	return topper;
    }

    TableField getParent() {
	return parent;
    }
     
    String getText() {
	return text;
    }
    
    String getIdentifier() {
	StringBuffer buf = new StringBuffer();
	if(isLR()) {
	    buf.append("LR:");
	} else {
	    buf.append("TB:");
	}
	buf.append(fieldCount());
	buf.append('(');
	buf.append(text);
	buf.append(')');
	TableField prnt = getParent();
	while(prnt != null) {
	    buf.append(',');
	    buf.append(prnt.fieldCount());
	    buf.append('(');
	    buf.append(prnt.getText());
	    buf.append(')');
	    prnt = prnt.getParent();
	}
	return buf.toString();
    }

    /**
     * Get the bounding box of this element
     *
     * @return the bounding box of this element
     */
    Rectangle getBounds() {
	return bounds;
    }

    void setBounds(int x, int y, int width, int height) {
	bounds.setBounds(x,y,width,height);
    }

    void setBounds(Rectangle r) {
	bounds.setBounds(r);
    }

    /**
     * Get the size of this object.
     *
     * @return the size of this object.
     */
    Dimension getSize() {
	return size;
    }

    void setSize(int width, int height) {
	size.setSize(width,height);
    }

    void setSize(Dimension d) {
	size.setSize(d.width,d.height);
    }

    boolean hasFields() {
	if(subFields == null || subFields.length == 0 || subFieldsUsed == 0) {
	    return false;
	}
	return true;
    }

    synchronized int subfields(int len) {
	if(len < 1) return 0;
	subFields = new TableField[len];
	return subFields.length;
    }

    int fieldCount() {
	if(subFields == null) {
	    return 0;
	}
	return subFieldsUsed;
    }

    synchronized void addField(TableField tf) {
	// can cause exception
	subFields[subFieldsUsed++] = tf;
    }

    TableField fieldAt(int nbr) {
	if(nbr < 0 || nbr >= subFieldsUsed) return null;
	return subFields[nbr];
    }

    boolean isLR() {
	return orientLR;
    }

    void setLR(boolean lr) {
	orientLR = lr;
    }

    String getId() {
	return idTag;
    }

    void setId(String id) {
	idTag = null;
	if(id == null) return;
	char[] array = id.toCharArray();
	boolean hadNBSP = false;
	for(int i = 0; i < array.length; i++) {
	    if(array[i] == GrappaSupportRects.NBSP) {
		array[i] = ' ';
		hadNBSP = true;
	    }
	}
	if(hadNBSP) idTag = new String(array,0,array.length);
	else idTag = id;
    }

    Dimension sizeFields() {
	return sizeUpFields(this);
    }
  
    private Dimension sizeUpFields(TableField tf) {
	//System.err.println(tf.getIdentifier());
	int fc = tf.fieldCount();
	if(fc == 0) {
	    if(tf.getTextBounds() != null) {
		tf.setSize(tf.getTextBounds().getSize());
	    } else {
		tf.setSize(0,0);
	    }
	} else {
	    Dimension dtmp = null;
	    Dimension dim = new Dimension();
	    for(int cnt = 0; cnt < fc; cnt++) {
		dtmp = sizeUpFields(tf.fieldAt(cnt));
		if(tf.isLR()) {
		    dim.width += dtmp.width;
		    dim.height = (dim.height > dtmp.height) ? dim.height : dtmp.height;
		} else {
		    dim.width = (dim.width > dtmp.width) ? dim.width : dtmp.width;
		    dim.height += dtmp.height;
		}
	    }
	    tf.setSize(dim);
	}
	//System.err.println("Size:"+tf.getSize()+";"+tf.getIdentifier());
	return tf.getSize();
    }

    Dimension resizeFields(Dimension sz) {
	resizeUpFields(this,sz);
	return this.getSize();
    }
  
    void resizeUpFields(TableField tf, Dimension sz) {
	//System.err.println(tf.getIdentifier());

	Dimension delta = new Dimension(sz.width - tf.getSize().width, sz.height - tf.getSize().height);
	tf.setSize(sz);

	int fc = tf.fieldCount();
    
	if(fc == 0) {
	    //System.err.println("Size:"+tf.getSize()+";"+tf.getIdentifier());
	    return;
	}

	// adjust children, if any
	double incr = 0;
	if(tf.isLR()) {
	    incr = (double)delta.width / (double)fc;
	} else {
	    incr = (double)delta.height / (double)fc;
	}
	TableField tfield = null;
	int amt = 0;
	// reuse old space under new name for readability
	Dimension newSz = delta;
	for(int cnt = 0; cnt < fc; cnt++) {
	    tfield = tf.fieldAt(cnt);
	    amt = (int)Math.floor(((double)(cnt+1))*incr) - (int)Math.floor(((double)cnt)*incr);
	    if(tf.isLR()) {
		newSz.setSize(tfield.getSize().width+amt,sz.height);
	    } else {
		newSz.setSize(sz.width,tfield.getSize().height+amt);
	    }
	    resizeUpFields(tfield,newSz);
	}
    }

    void positionFields(Point pos) {
	posFields(this,pos);
    }

    private void posFields(TableField tf, Point pos) {
	tf.setBounds(pos.x,pos.y,tf.getSize().width,tf.getSize().height);
	// clip spillage outside outer-most bounds
	Rectangle b1 = tf.getBounds();
	Rectangle b2 = tf.getTopMost().getBounds();
	int tmpi;
	tmpi = Math.max(b1.x,b2.x);
	b1.width = Math.min(b1.x+b1.width,b2.x+b2.width) - tmpi;
	b1.x = tmpi;
	tmpi = Math.max(b1.y,b2.y);
	b1.height = Math.min(b1.y+b1.height,b2.y+b2.height) - tmpi;
	b1.y = tmpi;

	int fc = tf.fieldCount();

	if(fc == 0) {
	    return;
	}

	TableField tfield = null;
	for(int cnt = 0; cnt < fc; cnt++) {
	    tfield = tf.fieldAt(cnt);
	    posFields(tfield,new Point(pos));
	    if(tf.isLR()) {
		pos.x += tfield.getSize().width;
	    } else {
		pos.y += tfield.getSize().height;
	    }
	}
    }

    void setTextBounds(String str, Node node) {
	int lines = 1;
	boolean cwFont = false;
	double[] fontwidth = { GrappaSupportRects.constantFontwidth };
	    
	String fontname = node.getAttribute(FONTNAME_ATTR).getStringValue().toLowerCase();
	if(fontname.startsWith("courier") || fontname.startsWith("monospaced")) {
	    cwFont = true;
	} else if(fontname.startsWith("helvetica") || fontname.startsWith("sansserif")) {
	    fontwidth = GrappaSupportRects.helveticaFontwidth;
	} else {
	    fontwidth = GrappaSupportRects.romanFontwidth;
	}

	char[] array = str.toCharArray();
	double fwidth = 0;
	double xwidth = 0;
	int value = 0;
	for(int i = 0; i < array.length; i++) {
	    if(array[i] == GrappaSupportRects.NBSP) {
		array[i] = ' ';
	    }
	    if(array[i] == '\\' && (i+1) < array.length) {
		if(array[i+1] == 'n' || array[i+1] == 'l' || array[i+1] == 'r') {
		    lines++;
		    i++;
		    if(fwidth > xwidth)
			xwidth = fwidth;
		    fwidth = 0;
		    continue;
		}
	    }
	    value = array[i] - 32;
	    fwidth += (cwFont) ? fontwidth[0] : ((value >= 0 && value < fontwidth.length) ? fontwidth[value] : 0 );
	}
	if(fwidth > xwidth)
	    xwidth = fwidth;
	int height = ((Integer)node.getAttributeValue(FONTSIZE_ATTR)).intValue();
	int width = (int)Math.round((double)height * xwidth);
	textBounds = new Rectangle(0,0,width,height*lines);
	text = str;
    }

    Rectangle getTextBounds() {
	return(textBounds);
    }

    void debugID() {
	int fc = fieldCount();

	if(fc == 0) {
	    return;
	}

	TableField tfield = null;
	for(int cnt = 0; cnt < fc; cnt++) {
	    tfield = fieldAt(cnt);
	    tfield.debugID();
	}
    }

}

