package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

import java.util.List;

/**
 * This queue will be used by root operators in a query plan and 
 * is not meant or intended to be used by any operators anywhere else in the query plan.  
 * Placing this queue in the middle of the query plan will cause problems with execution 
 * and lead to unneccessary debug time.  <p><br>
 *
 * With that said, here is what this queue does.  This queue will record the
 * time that tuples are enqueued and then keep track of the time that
 * each tuple spent in the system.  The tuple will then be discarded (to save space)
 * and only the time it spent in the system will be left. Therefore, this
 * queue will not actually store any tuples, but rather, just the time that tuples
 * spent in the system.  <p>
 *
 * The XATStreamApplication and will make use of these queues and will be able to
 * report (to the statistics gatherer) the average age of the tuples in the queue. <p>
 * 
 * For now, it will simply record the oldest time stamp (min), but this can be changed in
 * the future as ideas are changed.   <p>
 *
 * This class extends queue so that the rest of the system can use this functionality seemlessly
 * (like operators and statistics gatherer).  Several methods will have to be reimplemented in order
 * to make this work.
 * <ol>
 * <li> Enqueue will now only enqueue the time that the tuple spent in the system, not the
 * actual tuple
 * <li> Dequeue returns a list of Longs, not Tuples.  Only 1 dequeue method was reimplemented
 * because the dequeue that returns a Tuple[] is not valid.  
 * <li> getQueueSize (both methods) now reference the timeStamps list, not the tuple list
 * </ol>
 *
 * @author: Brad Pielech
 */
public interface XATEndUserApplicationQueue extends XATQueue{

/**
 * Enqueues multiple tuples.  This is done by appending the list
 * to the internal link list.  
 * 
 * The TUPLE_COUNT property is updated to reflect the newly added tuples.  
 * 
 * @param tuples XATTuple[] 
 * @param caller the owner of the queue who wishes to enqueue some tuples
 */
public void enqueue(XATTuple[] tuples, Object caller);
/**
 * Append 1 tuple to the end of the queue
 * 
 * @param tuple edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple
 * @param caller the owner of the queue who wishes to enqueue some tuples
 */
public void enqueue(XATTuple tuple, Object caller);
/**
 * Enqueues multiple tuples.  This is done by appending the list
 * to the internal link list.  So tuples are added to the end of the list in the order returned
 * by the list's iterator
 * 
 * The TUPLE_COUNT property is updated to reflect the newly added tuples.  
 * 
 * @param tuples List 
 * @param caller the owner of the queue who wishes to enqueue some tuples
 */
public void enqueue(List tuples, Object caller);
	/**
	* Returns the number of tuples that are in the queue.
	* @param caller Object 
	* @return int the number of tuples on the queue
	*/
	int getQueueSize();
	/**
	* Returns the number of tuples that are available for this given customer.  This
	* is different than the TUPLE_COUNT property that corresponds to the total amount
	* of tuples in the whole queue.  This just returns the tuples available to one customer. 
	* @param caller Object 
	* @return int the number of tuples available to this customer.   
	*/
	int getQueueSize(Object Customer);
/**
 * Dequeues a number of tuples in a list.
 *
 * Creation date: (5/19/2003 3:14:29 PM)
 * @return java.util.List
 * @param howMany int
 * @param caller java.lang.Object
 */
List peekN(int howMany, Object caller);
/**
 * This will remove all tuples from the queue.  Note, this
 * method obviously can create havoc if it is used indiscriminately, but
 * it is a huge speedup for the XATApplication. 
 * Creation date: (4/13/2003 1:38:36 AM)
 */
public void removeAllTuples();
}
