package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

import java.util.ArrayList;
import java.util.List;
/**
 * This queue will be used by root operators in a query plan and 
 * is not meant or intended to be used by any operators anywhere else in the query plan.  
 * Placing this queue in the middle of the query plan will cause problems with execution 
 * and lead to unneccessary debug time.  <p><br>
 *
 * With that said, here is what this queue does.  This queue will record the
 * time that tuples are enqueued and then keep track of the time that
 * each tuple spent in the system.  The tuple will then be discarded (to save space)
 * and only the time it spent in the system will be left. Therefore, this
 * queue will not actually store any tuples, but rather, just the time that tuples
 * spent in the system.  <p>
 *
 * The XATStreamApplication and will make use of these queues and (the Application) will be able to
 * report (to the statistics gatherer) the average age of the tuples in the queue. <p>
 * 
 * For now, it will simply record the oldest time stamp (min), but this can be changed in
 * the future as ideas are changed.   <p>
 *
 * This class extends queue so that the rest of the system can use this functionality seemlessly
 * (like operators and statistics gatherer).  Several methods will have to be reimplemented in order
 * to make this work.
 * <ol>
 * <li> Enqueue will now only enqueue the time that the tuple spent in the system, not the
 * actual tuple
 * <li> Peek returns a list of Longs, not Tuples.  Only 1 peek method was reimplemented
 * because the peek that returns a Tuple[] is not valid.
 * <li> Because Dequeue is not used (no need to reimplement the whole increment cursor methods),
 * peek was used instead.  Therefore, a call to removeAllTuples should be made
 * <li> getQueueSize (both methods) now reference the timeStamps list, not the tuple list
 * </ol>
 * 
 * @author: Brad Pielech
 */
public class XATEndUserApplicationQueueImp extends XATMemoryQueueImp implements XATEndUserApplicationQueue{
	/** 
	 this list will keep consist of Longs that report how long each
	 enqueued tuple was in the system 	
	*/
	private java.util.ArrayList timeStamps;
/**
 * XATEndUserApplicationQueue constructor comment.
 */
public XATEndUserApplicationQueueImp() {
	super();
	timeStamps = new ArrayList();
}
/**
 * XATEndUserApplicationQueue constructor comment.
 * @param owner java.lang.Object
 */
public XATEndUserApplicationQueueImp(Object owner) {
	super(owner);
	timeStamps = new ArrayList();
}
/**
 * Enqueues an array of tuples.
 * 
 * Note: the provide is reserved for future usage.
 *
 * Creation date: (4/27/2003 1:23:37 AM)
 * @param tuples edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple[]
 * @param provideer java.lang.Object
 */
public synchronized void enqueue(XATTuple[] tuples, Object provideer) {
	long now = System.currentTimeMillis();
	
	for(int i = 0; i < tuples.length; i++){
		//try {
			/** bug: enterTime is the relative to starttime
			 *  so should use the fomula "difference = now - enterTime - starttime".
			 * */
			//long enterTime = tuples[i].getTimeStamp().getMinTimeStamp();
			long enterTime = tuples[i].getMinTimeStamp();
			long difference = now - enterTime;
			//store the time that the tuple was in the systemws
			this.timeStamps.add(new Long(difference));	
		//} catch (XATTimeStampException e) {
			//hrm..
		//}
		
	}
}
/**
 * calls enqueue with the tuple array in order to cut down on redundant code.
 *
 * @param tuple edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple
 * @param provider java.lang.Object
 */
public synchronized void enqueue(XATTuple tuple, Object provider) {
	XATTuple[] array = new XATTuple[1];
	array[0] = tuple;
	enqueue(array, provider);
}
/**
 * calls the enqueue with the tuple[] so we can reduce redundant code 
 * 
 * 
 *
 * Creation date: (4/27/2003 1:23:37 AM)
 * @param tuples java.util.List
 * @param provider java.lang.Object
 */
public synchronized void enqueue(java.util.List tuples, Object provider) {
	enqueue((XATTuple[]) tuples.toArray(new XATTuple[] {}), provider);
}
	/**
	* Returns the number of tuples that are in the queue.
	* @param caller Object 
	* @return long the number of tuples on the queue
	*/
public synchronized int getQueueSize() {
	return this.timeStamps.size();
}
/**
 * Return the number of tuples in the queue for the given customer.
 * 
 * Creation date: (4/27/2003 1:23:37 AM)
 * @return int
 * @param customer java.lang.Object
 */
public synchronized int getQueueSize(Object customer) {	
	return this.timeStamps.size() - this.getCursor(customer);
}
/**
 * This does the same as the XATQueue.peekN, but it works over the
 * timeStamps list instead.  This will not return tuples! This will return
 * a list of the times that the tuples spent in the system.  This is
 * important to keep in mind.  This is why this class is only meant for root operators-
 * connecting this class to another operator will cause some major ClassCastExceptions
 *
 * @return java.util.List
 * @param howMany int
 * @param caller java.lang.Object
 */
public List peekN(int number, Object customer) {

	//we assume only 1 customer and that customer always starts at the bottom. 
	int cursor = 0;

	//check validity
	if(cursor + number > this.timeStamps.size())
		throw new XATQueueException("Not enough tuples to dequeue!");
		
	//get tuples
	List result = new ArrayList(number);
	for(int i = 0; i < number; i++)
		result.add(this.timeStamps.get(cursor + i));		
	return result;
}
/**
 * This will remove all tuples from the queue.  Note, this
 * method obviously can create havoc if it is used indiscriminately, but
 * it is a huge speedup for the XATApplication. 
 * Creation date: (4/13/2003 1:38:36 AM)
 */
public synchronized void removeAllTuples() {
    this.timeStamps.clear();   
}
}
