package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

import java.util.ArrayList;

/**
 * <p>The <tt>XATTimeStamp</tt> Interface defines the methods used in defining
 * timestamps for a tuple. A tuple may have as many timestamps as it
 * has columns. This interface describes the ways in which the timestamps
 * may be accessed.</p>
 *
 *
 * <p>For example:<br><br>
 *
 * An <tt>XATTuple</tt> may have 8 columns, which we will number 0-7. Each column or group of
 * sequential columns can have an associated timestamp. <br><br>
 *
 * Any number of timestamps may be defined over these columns as long as:
 *
 * <ol> <li> The columns do not overlap. <li> The group of columns are sequential. </ol>
 *
 * <br>In the example, a the following timestamps could be created:
 *
 * <br><br>TimeStamp1: Columns 0-4
 * <br>TimeStamp2: Columns 5-5
 * <br>TimeStamp3: Columns 6-7
 *
 * <br><br>The following groups of timestamps would not be allowed:
 *
 * <br><br>TimeStamp1: Columns 4-0 -- The columns must be in ascending order
 * <br>TimeStamp2: Columns 5-6 -- This would be fine
 * <br>TimsStamp3: Columns 6-7 -- This would not be allowed because column 6 already has a timestamp defined.
 *
 * </p><p>
 * The following describes how to GET/SET XATTimeStamps:
 *
 * <br><b>GETTING TIMESTAMPS</b><br>
 *
 * <br>When a <tt>XATTimeStamp</tt> object is received you can get the timestamp for
 * any column in the <tt>XATTuple</tt>, or also the max, min, number of timestamps.
 *
 * <br><br>To get the XATTimeStamp object from the tuple, use the following command:
 *
 * <br><br>XATTuple.getXATTimeStamp();
 *
 * <br><br>You can now get various timestamps using the methods in the <tt>XATTimeStamp</tt> interface.
 * Some included methods are:
 *
 * <br><br>XATTimeStamp.getMaxTimeStamp();
 * <br>XATTimeStamp.getMinTimeSTamp();
 * <br>XATTimeSTamp.getTimeStamp(int column);
 *
 * <br><br>Note: If a timestamp is not set for a particular column in an <tt>XATTuple</tt>
 * getTimeStamp will throw an <tt>XATTimeStampException</tt>.
 *  </p><p>
 * <b>SETTING TIMESTAMPS</b>
 *
 * <br><br>A timestamp can be set for any column or range of columns in a XATTuple.
 * A sample timestamp would be:
 *
 * <br><br>XATTimeStamp.setTimestamp(1,1) which would set column 1 of XATTuple to the current time.
 * <br>XATTimeStamp.setTimestamp(2,5,a long time) which would set colums 2-5 of an XATTuple to the time given.
 *
 * <br><br>Note: If any columns overlap when attempting to set a timestamp, the setTimeStamp method will
 * throw an <tt>XATTimeStampException</tt>. 
 * </p>
 * Creation date: (5/16/2003 11:43:21 PM)
 * @author: Timothy Sutherland
 * @author: tims@cs.wpi.edu
 * @author: http://www.tsutherland.com
 */
public interface XATTimeStamp {
/**
 * Returns true if all timestamps for an XATTuple are equal, false otherwise.
 * Creation date: (5/23/2003 6:58:55 PM)
 * @return boolean
 * @exception XATTimeStampException The exception description.
 */
boolean allTimeStampsEqual() throws XATTimeStampException;
/**
 * Combines two XATTimeStamps into one for the current XATTimeStamp object. This method makes the following assumptions: <br>
 *
 * <ul>
 * <li> a will be joined with b in such a manner that b's columns are after a's.
 * <li> there is no filtering of either xattuple before combining timestamps.
 * </ul>
 * <p>Example: If tuple A has columns 1-4 and tuple B has columns 1-5 combine will make timestamps
 * that will go from 1-9, copying timestamp information for each XATTuple.
 *
 * Creation date: (5/28/2003 2:27:11 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStamp
 * @param a edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStamp
 * @param b edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStamp
 * @exception edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampException The exception description.
 */
void combine(XATTimeStamp a, XATTimeStamp b) throws XATTimeStampException;
/**
 * Returns the newest timestamp contained in the XATTuple
 * Creation date: (5/22/2003 10:03:27 AM)
 * @return long -The newest timestamp
 */
long getMaxTimeStamp() throws XATTimeStampException;
/**
 * Returns the oldest timestamp contained in the XATTuple
 * Creation date: (5/22/2003 10:03:49 AM)
 * @return long
 * @exception XATTimeStampException -If no timestamp exists
 */
long getMinTimeStamp() throws XATTimeStampException;
/**
 * Returns the number of timestamps in the XATTuple.
 * Creation date: (5/23/2003 6:54:16 PM)
 * @return int
 */
int getNumTimeStamps();
/**
 * Returns the timestamp from a particual column
 * Creation date: (5/22/2003 10:02:43 AM)
 * @return long -The timestamp
 * @param column int -The column to get the timestamp for.
 * @exception XATTimeStampException -If no timestamp exists
 */
long getTimeStamp(int column) throws XATTimeStampException;
/**
 * Returns an array of all of the timestamp objects.
 * Creation date: (5/28/2003 3:40:25 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.TimeStamp[]
 * @exception edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampException The exception description.
 */
TimeStamp[] getTimeStampObjects() throws XATTimeStampException;
/**
 * Returns all timestamps for a tuple
 * Creation date: (5/22/2003 10:05:40 AM)
 * @return long[] - An array of timestamps
 */
long[] getTimeStamps() throws Exception;
/**
 * Sets a timestamp for a range of tuples using current time in milliseconds
 * since 1970. (Uses System.currentTimeMillis() method)
 * Creation date: (5/22/2003 10:02:01 AM)
 * @param startColumn int -The beginning column for the timestamp
 * @param endColumn int -The ending column for the timestamp
 * @exception XATTimeStampException -If no timestamps exist
 */
void setTimeStamp(int startColumn, int endColumn) throws XATTimeStampException;
/**
 * Sets a timestamp for a range of tuples.
 * Creation date: (5/22/2003 10:01:20 AM)
 * @param time long -The time for the timestamp
 * @param startColumn int -The beginning column for the timestamp
 * @param endColumn int -The ending column for the timestamp
 * @exception XATTimeStampException -If no timestamps exist
 */
void setTimeStamp(long time, int startColumn, int endColumn) throws XATTimeStampException;
/**
 * Sets a timestamp for the given XATTuple. Note, this just sets the XATTimeStamp object
 * using information about the XATTuple passed in its parameter. In order to set the timestamp for
 * the tuple, the XATTuple.setTimeStamp() method must still be called. The XATTuple parameter
 * is simply to gather information about the XATTuple to assign proper information.
 * Creation date: (6/2/2003 1:13:31 PM)
 * @param timestamp long
 * @param tuple edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple
 * @exception edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampException The exception description.
 */
void setTimeStamp(long timestamp, XATTuple tuple) throws XATTimeStampException;
/**
 * Sets a timestamp for the given XATTuple. Note, this just sets the XATTimeStamp object
 * using information about the XATTuple passed in its parameter. In order to set the timestamp for
 * the tuple, the XATTuple.setTimeStamp() method must still be called. The XATTuple parameter
 * is simply to gather information about the XATTuple to assign proper information.
 * Creation date: (6/2/2003 1:13:58 PM)
 * @param tuple edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple
 * @exception edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampException The exception description.
 */
void setTimeStamp(XATTuple tuple) throws XATTimeStampException;
/**
 * Returns the XATTimeStamp as a string
 * Creation date: (5/27/2003 11:04:19 AM)
 * @return java.lang.String
 */
String toString();

/**
 * @param time
 * @param i
 */
void replaceTimeStamp(long time, int startColumn, int endColumn) throws XATTimeStampException;

/**
 * @param timestamps
 */
void setTimeStamps(ArrayList timestamps);

/**
 * @param maxTime
 */
void setMaxTimeStamp(long maxTime);

/**
 * @param minTime
 */
void setMinTimeStamp(long minTime);

Object clone();
}
