package edu.wpi.cs.dsrg.xmldb.xat.common.dag;

import java.util.ArrayList;
import java.util.Iterator;
import java.io.Serializable;

/**
 * Implementation of the XATTimeStamp interface.
 * Creation date: (5/22/2003 10:16:23 AM)
 * @author: Tim Sutherland
 */
public class XATTimeStampImp implements XATTimeStamp, Serializable, Cloneable {
	private ArrayList timestamps;
	private long maxTimeStamp;
	private long minTimeStamp;
/**
 * XATTimeStampImp constructor comment.
 */
public XATTimeStampImp() {
	timestamps = new ArrayList();
	maxTimeStamp = Long.MIN_VALUE;
	minTimeStamp = Long.MAX_VALUE;
}
/**
 * Returns true if all timestamps for an XATTuple are equal, false otherwise.
 * Creation date: (5/23/2003 7:04:41 PM)
 * @return boolean
 */
public boolean allTimeStampsEqual(){

	if(maxTimeStamp == minTimeStamp) return true;
	else return false;
	
}
/**
 * Returns the newest timestamp contained in the XATTuple
 * Creation date: (5/22/2003 10:16:23 AM)
 * @return long -The newest timestamp
 */
public long getMaxTimeStamp() throws XATTimeStampException {

	if(maxTimeStamp == Long.MIN_VALUE) 
		throw new XATTimeStampException("No timestamp has been set");
	else
		return maxTimeStamp;
}
/**
 * Returns the oldest timestamp contained in the XATTuple
 * Creation date: (5/22/2003 10:16:23 AM)
 * @return long
 * @exception XATTimeStampException -If no timestamp exists
 */
public long getMinTimeStamp() throws XATTimeStampException {

	if(minTimeStamp == Long.MAX_VALUE)
		throw new XATTimeStampException("No Timetamp has been set");
	else
		return minTimeStamp;
}
/**
 * Returns the number of timestamps in the XATTuple.
 * Creation date: (5/23/2003 7:04:41 PM)
 * @return int
 */
public int getNumTimeStamps() {
	return timestamps.size();
}
/**
 * Returns the timestamp from a particual column
 * Creation date: (5/22/2003 10:16:23 AM)
 * @return long -The timestamp
 * @param column int -The column to get the timestamp for.
 * @exception XATTimeStampException -If no timestamp exists
 */
public long getTimeStamp(int column) throws XATTimeStampException {

	Iterator i = timestamps.iterator();
	TimeStamp ts;
	
	while(i.hasNext()){

		ts = (TimeStamp) i.next();
		if(ts.inRange(column))
			return ts.getTimeStamp();
	}

	throw new XATTimeStampException("There is no timestamp for the column indicated");
}
/**
 * Returns all timestamps for a tuple
 * Creation date: (5/22/2003 10:16:23 AM)
 * @return long[] - An array of timestamps
 */
public long[] getTimeStamps() throws XATTimeStampException {

	if(timestamps.isEmpty())
		throw new XATTimeStampException("There are currently no timestamps for this XATTuple");

	long[] times = new long[timestamps.size()];
	int j=0;
	Iterator i = timestamps.iterator();
	TimeStamp ts;
	
	while(i.hasNext()){

		ts = (TimeStamp) i.next();
		times[j] = ts.getTimeStamp();
		j++;
	}

	return times;
	
}
/**
 * Sets a timestamp for a range of tuples using current time.
 * Creation date: (5/22/2003 10:16:23 AM)
 * @param startColumn int -The beginning column for the timestamp
 * @param endColumn int -The ending column for the timestamp
 * @exception XATTimeStampException -If no timestamps exist
 */
public void setTimeStamp(int startColumn, int endColumn) throws XATTimeStampException {

	this.setTimeStamp(System.currentTimeMillis(),startColumn,endColumn);
	}
/**
 * Sets a timestamp for a range of tuples.
 * Creation date: (5/22/2003 10:16:23 AM)
 * @param time long -The time for the timestamp
 * @param startColumn int -The beginning column for the timestamp
 * @param endColumn int -The ending column for the timestamp
 * @exception XATTimeStampException -If no timestamps exist
 */
public void setTimeStamp(long time, int startColumn, int endColumn) throws XATTimeStampException {
	
	if(endColumn < startColumn) 
		throw new XATTimeStampException("Start column must be less than the end column");


	// Check all of the current timestamp objects to make sure that the range is not currently in use.

	Iterator i = timestamps.iterator();
	TimeStamp ts;
	
	while(i.hasNext()){

		ts = (TimeStamp) i.next();
		if(ts.inRange(startColumn))
			throw new XATTimeStampException("The column " + startColumn + " is already contained in a timestamp");
		if(ts.inRange(endColumn))
			throw new XATTimeStampException("The column " + endColumn + " is already contained in a timestamp");
				
	}
	
	
	// Set the timestamp object
	
	timestamps.add(new TimeStamp(time,startColumn,endColumn));

	// Reset the max and min if need be.

	if(time < minTimeStamp) minTimeStamp = time;
	if(time > maxTimeStamp) maxTimeStamp = time;	
	}

/**
 * Combines two XATTimeStamps into one for the current XATTimeStamp object. This method makes the following assumptions: <br>
 *
 * <ul>
 * <li> a will be joined with b in such a manner that b's columns are after a's.
 * <li> there is no filtering of either xattuple before combining timestamps.
 * </ul>
 * <p>Example: If tuple A has columns 1-4 and tuple B has columns 1-5 combine will make timestamps
 * that will go from 1-9, copying timestamp information for each XATTuple.
 *
 * Creation date: (5/28/2003 3:29:17 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStamp
 * @param a edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStamp
 * @param b edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStamp
 * @exception edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampException The exception description.
 */
public void combine(XATTimeStamp a, XATTimeStamp b) throws XATTimeStampException {
	
	if(timestamps.size() > 0) 
		throw new XATTimeStampException("A timestamp is already defined, combine can only be called on an empty timestamp");

	int aSize = a.getNumTimeStamps();
	int bSize = b.getNumTimeStamps();
	int aRange = 0;
	TimeStamp aTimeStamps[] = a.getTimeStampObjects();
	TimeStamp bTimeStamps[] = b.getTimeStampObjects();

	try{

		// keep track of the overall range of tuple a
		
	for(int i = 0; i < aSize ; i++) {

		this.setTimeStamp(aTimeStamps[i].getTimeStamp(),aTimeStamps[i].getStartColumn(),aTimeStamps[i].getEndColumn());
		aRange += aTimeStamps[i].getRange();
	}
	for(int j = 0; j < bSize ; j++) {

		this.setTimeStamp(bTimeStamps[j].getTimeStamp(),bTimeStamps[j].getStartColumn()+aRange,bTimeStamps[j].getEndColumn()+aRange);


	}

	}catch(XATTimeStampException xse){ System.out.println(xse.getMessage()); }	
		
	}

/**
 * Returns an array of all of the timestamp objects.
 * Creation date: (5/28/2003 3:40:44 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.dag.TimeStamp[]
 * @exception edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampException The exception description.
 */
public edu.wpi.cs.dsrg.xmldb.xat.common.dag.TimeStamp[] getTimeStampObjects() throws XATTimeStampException {
	TimeStamp[] ts = new TimeStamp[timestamps.size()];

	for(int i = 0; i < timestamps.size() ; i++){
		ts[i] = (TimeStamp) timestamps.get(i);	
	}

	return ts;
}

/**
 * Returns the string representation of a timestamp.
 * Creation date: (5/27/2003 11:11:11 AM)
 * @return java.lang.String
 */
public String toString() {

	String ret = new String();

	ret += "Min Timestamp: " + minTimeStamp;
	ret += " Max Timestamp: " + maxTimeStamp;

	Iterator i = timestamps.iterator();
	TimeStamp ts;
	
	while(i.hasNext()){

		ts = (TimeStamp) i.next();
		ret += "\n\t[" + ts.toString() + "]";
	}

	
	return ret;
}

/**
 * Sets a timestamp for the given XATTuple. Note, this just sets the XATTimeStamp object
 * using information about the XATTuple passed in its parameter. In order to set the timestamp for
 * the tuple, the XATTuple.setTimeStamp() method must still be called. The XATTuple parameter
 * is simply to gather information about the XATTuple to assign proper information.
 * Creation date: (6/2/2003 1:17:40 PM)
 * @param timestamp long
 * @param tuple edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple
 * @exception edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampException The exception description.
 */
public void setTimeStamp(long timestamp, XATTuple tuple) throws XATTimeStampException {


	int endColumn =  tuple.getLength()-1;
	int startColumn = 0;

	try{


	this.setTimeStamp(timestamp,startColumn,endColumn);

	} catch (XATTimeStampException xse) { System.out.println(xse.getMessage()); }
	
	
	
	
	}

/**
 * Sets a timestamp for the given XATTuple. Note, this just sets the XATTimeStamp object
 * using information about the XATTuple passed in its parameter. In order to set the timestamp for
 * the tuple, the XATTuple.setTimeStamp() method must still be called. The XATTuple parameter
 * is simply to gather information about the XATTuple to assign proper information.
 * Creation date: (6/2/2003 1:17:40 PM)
 * @param tuple edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTuple
 * @exception edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStampException The exception description.
 */
public void setTimeStamp(XATTuple tuple) throws XATTimeStampException {
	
	long currentTime = System.currentTimeMillis();
	
	try{

		this.setTimeStamp(currentTime,tuple);
		
	}catch(XATTimeStampException xse) {}
	
	
	}

/* (non-Javadoc)
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStamp#replaceTimeStamp(long, int, int)
 */
public void replaceTimeStamp(long time, int startColumn, int endColumn) throws XATTimeStampException {
	if(endColumn < startColumn) 
		throw new XATTimeStampException("Start column must be less than the end column");


	// Check all of the current timestamp objects to make sure that the range is not currently in use.

	int numTimestamps = timestamps.size();
	boolean found = false;
	for (int i=0; i<numTimestamps; i++) {
		TimeStamp ts = (TimeStamp) timestamps.get(i);
		if (ts.getStartColumn() == startColumn && ts.getEndColumn() == endColumn) {
			found = true;
			try {
				ts.setTimeStamp(time);
				break;
			} catch(Exception e) {
			}
		}
	}

	if (!found) {
		TimeStamp newTS = new TimeStamp(time, startColumn, endColumn);
		timestamps.add(newTS);
	} 
	
	// Reset the max and min if need be.
	if(time < minTimeStamp) minTimeStamp = time;
	if(time > maxTimeStamp) maxTimeStamp = time;	
}

/* (non-Javadoc)
 * @see java.lang.Object#clone()
 */
public Object clone() {
	XATTimeStampImp xattimestamp = new XATTimeStampImp();
	ArrayList newTimestamps = new ArrayList();
	int numTimestamps = this.timestamps.size();
	for (int i=0; i<numTimestamps; i++) {
		TimeStamp ts = (TimeStamp)timestamps.get(i);
		newTimestamps.add((TimeStamp) ts.clone());
	}
	xattimestamp.setTimeStamps(newTimestamps);
	xattimestamp.setMaxTimeStamp(this.maxTimeStamp);
	xattimestamp.setMinTimeStamp(this.minTimeStamp);
	return xattimestamp;
}

/* (non-Javadoc)
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStamp#setTimeStamps(java.util.ArrayList)
 */
public void setTimeStamps(ArrayList timestamps) {
	this.timestamps = timestamps;
}

/* (non-Javadoc)
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStamp#setMaxTimeStamp(long)
 */
public void setMaxTimeStamp(long maxTime) {
	this.maxTimeStamp = maxTime;
}

/* (non-Javadoc)
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATTimeStamp#setMinTimeStamp(long)
 */
public void setMinTimeStamp(long minTime) {
	this.minTimeStamp = minTime;
}
}
