package edu.wpi.cs.dsrg.xmldb.xat.common.expression;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.sqlgeneration.*;

/**
 * This class defines a terminal expression.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 */

public class TerminalExpressionImp implements TerminalExpression {

	/**
	 * The type of the terminal expression.
	 * @since 1.0
	 */
    int iType;

    /**
     * The value of the expression.
     * @since 1.0
     */
    Object value;
/**
 * The default TerminalExpressionImp constructor.
 * @since 1.0
 */
public TerminalExpressionImp() {
}
/**
 * The default TerminalExpressionImp constructor using a double value.
 * @param dValue The double value used to construct the terminal expression.
 * @since 1.0
 */

public TerminalExpressionImp(double dValue) {
    iType = DOUBLE;
    value = new Double(dValue);
}
/**
 * The default TerminalExpressionImp constructor using a column name object.
 * @param cnValue The Column Name object used to construct the terminal expression.
 * @since 1.0
 */

public TerminalExpressionImp(ColumnName cnValue) {
    iType = NAME;
    value = cnValue;
}
/**
 * The default TerminalExpressionImp constructor using a boolean object.
 * @param bValue The boolean object used to construct the terminal expression.
 * @since 1.0
 */

public TerminalExpressionImp(Boolean bValue) {
    iType = BOOL;
    value = bValue;
}
/**
 * The default TerminalExpressionImp constructor using a Double object.
 * @param dValue The Double object used to construct the terminal expression.
 * @since 1.0
 */

public TerminalExpressionImp(Double dValue) {
    iType = DOUBLE;
    value = dValue;
}
/**
 * The default TerminalExpressionImp constructor using a String object.
 * @param sValue The String object used to construct the terminal expression.
 * @since 1.0
 */
public TerminalExpressionImp(String sValue) {
    iType = STRING;
    value = sValue;
}
/**
 * The default TerminalExpressionImp constructor using a boolean value.
 * @param bValue The boolean value used to construct the terminal expression.
 * @since 1.0
 */

public TerminalExpressionImp(boolean bValue) {
    iType = BOOL;
    value = new Boolean(bValue);
}
/**
 * Evaluate the expression of the XATOperator.
 * @return The evaluate result of expression.
 * @param xopOper The operator evaluated.
 * @param iPositions The position of tuple in the XATTable.
 * @since 1.0
 */
 
public Object eval(XATOperator xopOper, int[] iPositions) {
    switch (iType) {
        case STRING :
        case DOUBLE :
            return new XATDataValueImp(getValue());
        case NAME :
            // needs to return the value in the iPosition for the table
            XATNode[] children = xopOper.getNode().getChildren();
            String sColumnName = ((ColumnName) value).getColumnName();
            String sTableName = ((ColumnName) value).getTableName();
            if (sTableName == null) {
                // look through the tables until we find a matching column

                for (int i = 0; i < children.length; i++) {
                    XATTable xtabTable = (XATTable) children[i].getOperator().getOutputDataSource();
                    try {
                        int iColumnPosition = xtabTable.getColumnPosition(sColumnName);
                        Object o = xtabTable.getValue(iColumnPosition, iPositions[i]);
                        return o;

                    } catch (XATTableColumnNoExistException e) {
                        // so what?
                    } catch (XATTableOutOfRangeException e) {
                        // should never happen
                    } catch (XATTableEmptyException e) {
                        // should never happen
                    }
                }
                // throw something...
            } else {
                // look in the specified table
                for (int i = 0; i < children.length; i++) {
                    XATTable xtabTable = (XATTable) children[i].getOperator().getOutputDataSource();
                    if (xtabTable.getName().equals(sTableName)) {
                        try {
                            int iColumnPosition = xtabTable.getColumnPosition(sColumnName);
                            Object o = xtabTable.getValue(iColumnPosition, iPositions[i]);
                            return o;
                        } catch (XATTableColumnNoExistException e) {
                            // damn.  don't throw an exception yet, though.
                        } catch (XATTableOutOfRangeException e) {
                            // should never happen
                        } catch (XATTableEmptyException e) {
                            // should never happen
                        }
                    }
                }
                // throw something...
            }
        default :
            return new XATDataValueImp();
    }
}
/**
 * get column names related with expression.
 * @return ColumnName[] The array containing all the related column names.
 * @since 1.0
 */

public ColumnName[] getRelatedColumnNames() {
	if(value instanceof ColumnName){
		return new ColumnName[]{(ColumnName)this.value};
	}else
		return new ColumnName[]{};
}
/**
 * Get the type of expression.
 * @return int The type of expression.
 * @since 1.0
 */

public int getType() {
    return iType;
}
/**
 * Get the value of this expression.
 * @return Object The value of this expression.
 * @since 1.0
 */
public Object getValue() {
    return value;
}
/**
 * Set the type of the expression.
 * @param type The number used to represent the type of expression.
 * @since 1.0
 */

public void setType(int type) {
    iType = type;
}
/**
 * Set the double value for this terminal expression.
 * @param value The double value of this expression.
 * @since 1.0
 */

public void setValue(double dValue) {
    iType = 1;
    value = new Double(dValue);
}
/**
 * Set the column name value for this terminal expression.
 * @param value The column name value of this expression.
 * @since 1.0
 */

public void setValue(ColumnName cnValue) {
    iType = 2;
    value = cnValue;
}
/**
 * Set the string value for this terminal expression.
 * @param value The string value of this expression.
 * @since 1.0
 */
 
public void setValue(String sValue) {
    iType = 0;
    value = sValue;
}
/**
 * Set the value of this expression using given boolean value.
 * @return Object The boolean value for this expression.
 * @since 1.0
 */

public void setValue(boolean bValue) {
    iType = 1;
    value = new Boolean(bValue);
}
/**
 * Convert the expression as a SQL statement.
 * @return java.lang.String The converted string.
 * @param soe The operator evaluator for SQL generation.
 * @since 1.0
 */
public String toSQL(OperatorEvaluator soe) {
    switch (iType) {
        case STRING :
            return "'" + value + "'";
        case DOUBLE :
            return value.toString();
        case NAME :
            //return the column name with double quotes bookending it
            // so that SQL doesn't confuse any $ symbols.
            return "\"" + ((ColumnName) value).toString() + "\"";

            //return soe.createString(((ColumnName)value)/*.getColumnName()*/);

    }
    return "";
}
/**
 * Output a string of this expression.
 * @return String the output result.
 * @since 1.0
 */

public String toString() {
    switch (iType) {
        case STRING :
            return "\"" + value + "\"";
        case DOUBLE :
        case NAME :
            return value.toString();
        default :
            return "Terminal(" + Integer.toString(iType) + ")";
    }
}
}
