package edu.wpi.cs.dsrg.xmldb.xat.common.treestructure;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.*;
import java.lang.reflect.Array;

import edu.wpi.cs.dsrg.xmldb.xat.common.dompattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.*;

/**
 * This Class implemented the DOMpattern Interface. 
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 */

public class DOMPatternImp implements DOMPattern {
	/**
	 * The root of the tree.
	 * @since 1.0
	 */

	protected RootNode rootNode;

	/**
	 * The lever after convert the DOMPattern to Indent String.
	 * @since 1.0
	 */
	int level;

/**
 * The default DOMPatternImp constructor.
 * @since 1.0
 */
public DOMPatternImp() {
}
/**
 * The DOMPatternImp constructor.
 * @param root The root of DOMPatterm.
 * @since 1.0
 */
public DOMPatternImp(RootNode root) {
    rootNode = root;
}
/**
 * Add a new parent node for a given DOMPatternNode node. 
 * The parent node is updated to have a link to the child node.
 * The child node is updates to have a link to the parent node.
 * @return void
 * @param parent The parent node it attached to.
 * @param child The child node attached.
 * @since 1.0
 */

public void addNode(DOMPatternNode newNode, DOMPatternNode parentNode) {
    newNode.setParent(parentNode);
    parentNode.addChild(newNode);
}
/**
 * Test to see if this DOMPattern contains a specific column name in its
 * pattern tree.
 * @return boolean True: The given DOMPattern contains the columnName. False: otherwise.
 * @param columnName The column name searching for.
 * @since 1.0
 */

public boolean containsColumn(String columnName) {
    // let's first make a list pattern out of it.
    return toListPattern(getRoot()).containsColumn(columnName);
}
/**
 * Return the columnNames associated with this DomPatternNode
 * use the toListPattern method which traverses the tree
 * but toListPattern produces a Vector with both the tagNames and columnNames
 * in it, so we need to remove the tagNames
 * Therefore, create a new Vector which just has a string representation of
 * the columnNames of the ListPattern
 * 
 * @return java.lang.String[]
 * @since 1.0
 */
public String[] getColumnNames() {
    //call toListPattern with the root
    ListPattern lp1 = toListPattern(this.getRoot());
    Vector listP = lp1.getList();
    Vector returnV = new Vector();
    String s1;
    for (int i = 0; i < listP.size(); i++) {
        if (listP.elementAt(i) instanceof ColumnNameImp) {
            //create a new String equal to the columnName of this ColumnNameImp object
            s1 = new String(((ColumnNameImp) listP.elementAt(i)).getColumnName());
            //add it to the vector to return;
            returnV.addElement(s1);
        }
    }
    return (String[]) returnV.toArray(new String[] {
    });
}
/*
 * Get the root of the tree.
 * @return RootNode The root of the DOMPattern.
 * @since 1.0
 */
public RootNode getRoot() {
    return rootNode;
}
/*
 * Searching if the given DOMPatternNode is in the tree or not.
 * @return boolean True: The node is found in the tree. False: otherwise.
 * @param node The DOMPatternNode searching for.
 * @since 1.0
 */
public boolean isInTree(DOMPatternNode node) {
    DOMPatternNode[] children = rootNode.getChildren();
    int size = Array.getLength(children);

    for (int i = 0; i < size; i++) {
        //compare to see if the nodes are equal
        String nodeClass = children[i].getClass().toString();
        String nodeTagName = children[i].getTagName();
        //first you check to see if they are of the same class
        //then you check to see if they have the same tag name
        if (nodeClass.equals(node.getClass().toString())) {
            if (nodeTagName.equals(node.getTagName())) {
                return true;
            }
        }
        isInTree(children[i]);
    }
    return false;
}
/**
 * Delete the given DOMPatternNode. 
 * It will affect number of nodes. All the children nodes of the
 * to-be-deleted node will be the children of the parent node of the to-be-delete node.
 * @return void
 * @param node the to-be-deleted child node.
 * @since 1.0
 */

public void removeNode(DOMPatternNode node) {
    //get all the children of the node
    DOMPatternNode[] children = node.getChildren();
    //get the parent of the node
    DOMPatternNode parent = node.getParent();

    parent.deleteChild(node);

    int length = Array.getLength(children);
    for (int i = 0; i < length; i++) {
        children[i].setParent(parent);
        parent.addChild(children[i]);
    }
}
/**
 * Set root node of DOMPattern.
 * @return void.
 * @param new_root The new root node.
 * @since 1.0
 */

public void setRoot(RootNode root) {
    rootNode = root;
}
/**
 * Get the indented print output of the result.
 * This method is used for the GUI.
 * @return String The indented print output of the result.
 * @since 1.0     
 */

public String toIndentString() {
    level = -1;
    return toIndentString(rootNode);
}
/**
 * Get the indented print output of the result.
 * This method is used for the GUI.
 * @return String The indented print output of the result.
 * @since 1.0     
 */

public String toIndentString(DOMPatternNode rn) {
    level++;
    // return all the children.
    String result = "";
    if (rn instanceof RootNode) {
        DOMPatternNode children[] = rn.getChildren();
        if (children.length > 0) {
            for (int i = 0; i < children.length - 1; i++) {
                result += toIndentString(children[i]) + ",\n";
            }
            result += toIndentString(children[children.length - 1]);
        }
    } else
        if (rn instanceof AttributeNode) {
            result += ((AttributeNode) rn).getTagName() + ":";
            DOMPatternNode children[] = rn.getChildren();
            if (children.length > 0) {
                for (int i = 0; i < children.length - 1; i++) {
                    result += toIndentString(children[i]) + ",";
                }
                result += toIndentString(children[children.length - 1]);
            }
        } else
            if (rn instanceof ColumnNameNode) {
                result = ((ColumnNameNode) rn).getColumnName();
            } else
                if (rn instanceof TagNode) {
                    result = "<" + ((TagNode) rn).getTagName() + ">\n";
                    DOMPatternNode children[] = rn.getChildren();
                    if (children.length > 0) {
                        for (int i = 0; i < children.length - 1; i++) {
                            result += toIndentString(children[i]) + ",\n";
                        }
                        result += toIndentString(children[children.length - 1]) + "\n";
                    }
                    result += "</" + ((TagNode) rn).getTagName() + ">";
                } else
                    if (rn instanceof TextNode) {
                        result = ((TextNode) rn).getTagValue().toString();
                    } else { // just get its string value and put it there.
                        result = rn.toString();
                    }
    level--;
    return result;
}
/**
 * Convert the tree into a list by extracting the TagName and TagValue.
 * @return ListPattern The result listPattern of given DOMPattern Node.
 * @param node The DOMPatternNode translate for.
 * @since 1.0     
 */

public ListPattern toListPattern(DOMPatternNode node) {

    DOMPatternNode[] children = node.getChildren();
    ListPattern result = new ListPatternImp();

    if (node instanceof RootNodeImp) {
        //handle the RootNode
        for (int i = 0; i < children.length; i++)
            result.add(toListPattern(children[i]));
    } else
        if (node instanceof TagNodeImp) {
            //handle Tag node

            // if the parent of this node and previous node are the same,
            // close the tag
            StringBuffer openTag = new StringBuffer();

            openTag.append("<" + node.getTagName());

            if (children.length > 0) {
                // break the open tag in multiple pieces if there is any attribute.
                result.add(openTag.toString());
                openTag = new StringBuffer();

                for (int i = 0; i < children.length; i++) {
                    if (children[i] instanceof AttributeNodeImp) {

                        openTag.append(" " + children[i].getTagName() + "=\"");

                        DOMPatternNode attValue = children[i].getChild(0);

                        if (attValue instanceof TextNode) {
                            // if that's a text node, put the value here.
                            openTag.append(((TextNode) attValue).getTagValue());
                        } else
                            if (attValue instanceof ColumnNameNode) {
                                // if that's a column name node, put special varstep here.
                                //handle Column Name node
                                //get the first element of the Navigation Step Array
                                // represendint the column Name
                                Object o = Array.get(attValue.getTagValue(), 0);

                                String stepValue = ((NavigationStep) o).getStepValue();

                                result.add(openTag.toString());
                                result.add(new ColumnNameImp(stepValue));
                                openTag = new StringBuffer();
                            }
                        openTag.append("\"");

                    }
                }
            }

            // all the attributes have been printed, how print ">"
            openTag.append(">");

            result.add(openTag.toString());

            //Continue with the rest of the tree
            for (int i = 0; i < children.length; i++) {
                if (!(children[i] instanceof AttributeNodeImp))
                    result.add(toListPattern(children[i]));
            }

            //print the closing tag
            result.add("</" + node.getTagName() + ">");
        } else
            if (node instanceof ColumnNameNodeImp) {
                //handle Column Name node
                //get the first element of the Navigation Step Array
                // represendint the column Name
                Object o = Array.get(node.getTagValue(), 0);

                String stepValue = ((NavigationStep) o).getStepValue();

                result.add(new ColumnNameImp(stepValue));
            }

    //handle Text Node
    else
        if (node instanceof TextNodeImp) {

            //format: "text"

            result.add("\"" + (String) node.getTagValue() + "\"");

        }

    return result;

}
/**
 * Convert the tree into a String.
 * @return String
 * @since 1.0
 */
public String toString() {
    ListPattern lp = this.toListPattern(this.getRoot());
    return lp.toString();
}
}
