package edu.wpi.cs.dsrg.xmldb.xat.common.xatnode;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.lang.reflect.Array;

import edu.wpi.cs.dsrg.utils.DeepCopyMaker;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.*;

/**
 * The class defines the node of XATTree.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0  
 */

public class XATNodeImp implements XATNode, Cloneable {

	/**
	 * The operator connected with this XATNode.
	 * @since 1.0
	 */
    protected XATOperator operator;

	/**
	 * The parent nodes of this XATNode.
	 * @since 1.0
	 */
    protected ArrayList parent;

	/**
	 * The child nodes of this XATNode.
	 * @since 1.0
	 */
    protected List children;

	/**
	 * The required schema of this XATNode.
	 * @since 1.0
	 */
    protected Set schema = null;

/**
 * The default constructor of XATNodeImp.
 * @since 1.0
 */
public XATNodeImp() {
    operator = null;

    parent = null;

    children = new ArrayList();
    parent = new ArrayList();
}
/**
 * Add one child to this node. This should be called only by the XATTree.addNode().
 * Remember the children are order unsensitive.
 * @return int The number of children.
 * @param new_child The new child node added.
 * @since 1.0
 */
public int addChild(XATNode new_child) {
	if((new_child != null) && !children.contains(new_child))
	    children.add(new_child);
    return children.size();
}
/**
 * Add one child to this node at specified position. 
 * This method should be called only by the XATTree.addNode().
 * The children are order unsensitive.
 * @return The number of children.
 * @param new_child The new child node.
 * @param position The position of new added node.
 * @since 1.0
 */
public int addChild(XATNode new_child, int position) {
    children.add(position, new_child);
    return children.size();
}
/**
 * Clear the schema of this XATNode.
 * @return Set The old schema of this XATNode.
 * @since 1.0
 */
public Set clearSchema() {
	Set oldSchema = schema;
	schema = null;
    return oldSchema;
}
/**
 * Deep clone this XATNode.
 * @return Object The cloned result.
 * @since 1.0
 */
public Object clone() {
    XATNode result = new XATNodeImp();
    try {
        result.setOperator((XATOperator) DeepCopyMaker.makeDeepCopy(operator));
        result.setSchema((Set) DeepCopyMaker.makeDeepCopy(schema));
    } catch (Exception e) {
        System.out.println("Deep clone failed!");
        e.printStackTrace(System.err);
    }
    result.getOperator().setNode(result);
    return result;
}
/**
 * Delete a child from this node.
 * @return int The number of children.
 * @param child The to-be-deleted child node.
 * @since 1.0
 */
public int deleteChild(XATNode child) {
    children.remove(child);
    return children.size();
}
/**
 * Get a specific child node of this node by their position.
 * @return XATNode The child node got.
 * @param position The position of the child node.
 * @throws XATNodeDoesNotExistException
 * @since 1.0
 */
public XATNode getChild(int position) throws XATNodeDoesNotExistException {
    try {
        return (XATNode) children.get(position);
    } catch (ArrayIndexOutOfBoundsException e) {
        throw new XATNodeDoesNotExistException();
    }
}
/** 
 * Get all the children nodes of this node.
 * @return XATNode[] The array of all the children.
 * @since 1.0
 */
public XATNode[] getChildren() {
    /* hideous, hideous kludge... java is too pedantic */
    return (XATNode[]) children.toArray(new XATNode[] {
    });
}
/**
 * Returns how many children this node has
 * @return int The number of children.
 * @since 1.0
 */
public int getChildrenSize() {
	return children.size();
}
/**
 * Return the index of a node.
 * @return int The index number of given node.
 * @param node The XATNode computing index for.
 * @deprecated user getIndexOfChild() instead
 * @since 1.0
 */

public int getIndexOf(XATNode node) {
	for(int i = 0; i < children.size(); i++){
		if(children.get(i) == node)
			return i;
	}
	return -1;
}
/** 
 * Get the operator in this Node.
 * @return XATOperator The operator of this node.
 * @since 1.0
 */
public XATOperator getOperator() {
    return operator;
}
/** 
 * Get the parent node of this node.
 * @return XATNode The parent of this node.
 * @since 1.0
 */
public XATNode getParent() {
	return (parent.size() > 0) ? (XATNode)parent.get(0) : null;	    
}
/**
 * Get an Iterator of all of the parents.  
 * The XATNode only has one parent, thus this iterator will only contain one element. 
 * However, subclasses may have multiple parents, thus nothing is lost by placing this method here.
 * @return Iterator Including all the parents.
 * @since 1.0
 */
public java.util.Iterator getParents() {
	return parent.iterator();
}
/**
 * Get the schema of this XATNode.
 * @return Set The schema of this XATNode.
 * @since 1.0
 */
public Set getSchema() {
    return schema;
}
/**
 * Get a position for a specific child node.
 * If not found, return -1.
 * @return int The position of this child.
 * @param child The reference of child node searching for.
 * @since 1.0
 */
public int locateChild(XATNode child){
	return children.indexOf(child);
}
/**
 * Set a specific child to be a new child node.
 * @return XATNode The original child node.
 * @param position The position of the child node.
 * @param new_child The new child node.
 * @since 1.0
 */
public XATNode setChild(int position, XATNode new_child)
    throws XATNodeDoesNotExistException {
    try {
        XATNode tmp = null;
        if (children.size() > position) {
            tmp = (XATNode) children.get(position);
            children.set(position, new_child);
        }
        else {
            children.add(new_child);
        }
        return tmp;
    }
    catch (NullPointerException e) {
        throw new XATNodeDoesNotExistException();
    }
}
/**
 * Set all the children nodes of this node.
 * @return XATNode[] The array containing all the old children.
 * @param new_children The new children nodes.
 * @since 1.0
 */
public XATNode[] setChildren(XATNode[] new_children) {
    List tmp = children;
    int i;
    int length = Array.getLength(new_children);
    children = new ArrayList(length);
    for (i = 0; i < length; i++) {
        children.add(new_children[i]);
    }
    return (XATNode[]) children.toArray(new XATNode[] {
    });
}
/**
 * Set the operator in this Node.
 * @return XATOperator The old operator in this node.
 * @param new_operator The new operator in this node.
 * @since 1.0
 */
public XATOperator setOperator(XATOperator new_operator) {
    XATOperator tmp = operator;
    operator = new_operator;
    return tmp;
}
/**
 * Set the parent node of this node.
 * @return XATNode The original parent of this node.
 * @param new_parent The new parent node of this node.
 * @since 1.0
 */
public XATNode setParent(XATNode new_parent) {

    XATNode tmp = getParent();
    if (parent.size() > 0) {        
        parent.set(0, new_parent);
    } else {
        parent.add(new_parent);
    }
    return tmp;

}
/**
 * Set the schema for current node.
 * @return Set The old schema of this node.
 * @param new_schema The new schema for this node.
 * @since 1.0
 */
public Set setSchema(Set new_schema) {

    if (new_schema == null)
        return schema;

    Set tmp = schema;
    schema = new_schema;
    // also make the output xat table in this schema.
    XATTable output = new XATTableImp();
    Object[] names = schema.toArray();
    try {
        for (int i = 0; i < names.length; i++)
            output.appendColumn((String) names[i]);
    } catch (XATTableTupleNoMatchException nme) {
        nme.printStackTrace(System.err);
    }

    if (this.getOperator() != null)
        this.getOperator().setOutput(output);

    return tmp;
}
/**
 * Print current node information
 * @return String The String representation fot this node.
 * @since 1.0
 */
public String toString() {
    try {
        return this.hashCode() + ":" +  operator.toString();
    }
    catch (NullPointerException e) {
        return "NOP";
    }

}
/* (non-Javadoc)
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode#getChildrenList()
 */
public List getChildrenList() {
	// TODO Auto-generated method stub
	return null;
}
/* (non-Javadoc)
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode#setChildren(java.util.List)
 */
public void setChildrenList(List new_children) {
	children = new_children;
	
}
}
