package edu.wpi.cs.dsrg.xmldb.xat.component.distribution.connection;
import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.ConnectException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.state.OperatorState;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.Packet;
import edu.wpi.cs.dsrg.xmldb.xat.component.distribution.packet.PacketPool;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.DistributedExecutionController;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;

/**
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 8, 2003
 * 
 * The connection listner is responsible for running on the Raindrop Client
 * and listens for connection requests from the Distribution Manger. These
 * requests are then processed by the listener and the query plan is updated
 * to reflect these changes. There are currently 3 types of connections:
 * 
 *  - Stream to Machine
 *  - Machine to Machine
 *  - Machine to Application
 * 
 */
public class ConnectionListener extends Thread implements Runnable {

	/**Stores the reference to the port number*/
	private int port = -1;
	private Socket incoming;
	private ConnectionTable inConnections;
	private ConnectionTable outConnections;
	private String clientname;
	private boolean keepGoing;
	private ArrayList connectionHandlers;
	private DistributedExecutionController dExec;
	
	/**
	 * Creates a new Package receiver for the execution controller.
	 */
	public ConnectionListener(
		int port,
		String clientname,
		ConnectionTable in,
		ConnectionTable out,
		DistributedExecutionController exec)
		throws IOException {
		super();
		this.port = port;
		this.inConnections = in;
		this.outConnections = out;
		this.clientname = clientname;
		keepGoing = true;
		connectionHandlers = new ArrayList(10);
		this.dExec = exec;
	}

	/* Runs the Connection Listener
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		try {
			ServerSocket socket = new ServerSocket(this.port);
			System.out.println(
				"Starting up Connection Listener on port: " + port);
			int i = 1;
			while (keepGoing) {
				Socket inc = socket.accept();
				ConnectionHandler newHandler =
					new ConnectionHandler(
						inc,
						inConnections,
						outConnections,
						i, dExec);
				newHandler.setName("ConnectionHandler" + i);
				newHandler.setDaemon(true);
				newHandler.start();
				connectionHandlers.add(newHandler);
				i++;
			}

		} catch (IOException ioe) {
			System.out.println(
				"Could not start the Connection Listener on port "
					+ this.port
					+ ioe.getMessage());
		}
	}

	public void requestStop() {
		System.out.println("Trying to interrupt the threads");

		Iterator i = connectionHandlers.iterator();

		while (i.hasNext()) {
			ConnectionHandler c = (ConnectionHandler) i.next();
			if (c != null) {
				c.requestStop();
			}
		}

		keepGoing = false;
		interrupt();

	}
}

/**
 * 
 * @author Timothy M. Sutherland
 * @author tims@cs.wpi.edu
 * @author http://www.tsutherland.com
 * @version 1.0
 * 
 * Dec 8, 2003
 * 
 * The PacketReceiverHandler class is responsible for
 * maintaining a connection between the local execution engine
 * and a remote engine. The PacketReceiver will spawn one of these threads
 * whenever a Packet from a new location is sent.
 * 
 */
class ConnectionHandler extends Thread {

	private Socket incoming;
	private PacketPool pool;
	private int handlerNumber;
	private boolean keepGoing;
	private ConnectionTable inConnections;
	private ConnectionTable outConnections;
	private StatisticsGatherer stats;
	private DistributedExecutionController dExec;
	
	public ConnectionHandler(
		Socket i,
		ConnectionTable in,
		ConnectionTable out,
		int handlerNumber, DistributedExecutionController exec) {
		this.incoming = i;
		this.handlerNumber = handlerNumber;
		this.keepGoing = true;
		this.inConnections = in;
		this.outConnections = out;
		stats = StatisticsGatherer.getStatisticsGather();
		this.dExec = exec;
		
	}

	public synchronized void run() {
	


		try {

			ObjectInputStream in =
				new ObjectInputStream(incoming.getInputStream());
			XATStreamOperator operator = null;
			String stream;
			String operatorNumber;

			boolean enable = true;
			while (enable) {

				//Receive the new Packets

				Object o = in.readObject();
				Packet cp = (Packet) o;
				Connection c = (Connection) cp.getObject();
				//System.out.println(cp.toString());

				// Manage the connection and add it to the hashtable 
				// whether its incoming or outgoing

				//*** STREAM TO MACHINE CONNECTION
				if (c instanceof ReceiveStream) {
					stream = c.getStreamName();
					operatorNumber = c.getToOperator();

					Iterator j = stats.getAllKnownOperators();
					for (; j.hasNext();) {
						XATStreamOperator temp = ((XATStreamOperator) j.next());
						if (temp
							.getProperty("OBJECT_ID")
							.equals(operatorNumber)) {
							operator = temp;
							break;
						}
					}
	
					OperatorQueue oq = new OperatorQueue(operator,c.getToQueueNumber());

					// Add the operator to the table if it bridges a network
					// connection. This will apply to Stream->Machine, 
					// Machine -> Application and Machine->Remote Machine
					if (operator != null) {
						inConnections.addConnection(stream, oq);
						System.out.println(inConnections);
						operator.setActive(true);
					}
				} //endif
				//*** LOCAL MACHINE CONNECTION
				else if (c instanceof Activate) {
					// nothing to do here but active the operator(s)
					// the queues are already connected

					operatorNumber = c.getFromOperator();

					Iterator j = stats.getAllKnownOperators();
					for (; j.hasNext();) {
						XATStreamOperator temp = ((XATStreamOperator) j.next());
						if (temp
							.getProperty("OBJECT_ID")
							.equals(operatorNumber)) {
							operator = temp;
							break;
						}
					}
					//Activate the operator
					if (!operator.isActive()) {
						operator.setActive(true);
					}
				} //endif
				// *** SEND DATA
				else if (c instanceof SendData) {

					//First get the operator to connect to the application
					operatorNumber = c.getFromOperator();

					Iterator j = stats.getAllKnownOperators();
					for (; j.hasNext();) {
						XATStreamOperator temp = ((XATStreamOperator) j.next());
						if (temp
							.getProperty("OBJECT_ID")
							.equals(operatorNumber)) {
							operator = temp;
							break;
						}
					}

					// Now create the socket
					Socket s =
						new Socket(
							c.getToIPAddress(),
							c.getToPort());
					ObjectOutputStream outputStream =
						new ObjectOutputStream(s.getOutputStream());

					outConnections.addConnection(operator, outputStream);
					System.out.println(outConnections);

				} //endif
				//*** RECEIVING FROM MACHINE CONNECTION
				else if (c instanceof ReceiveData) {
					
					String operatorFrom = c.getFromOperator();
					String operatorTo = c.getToOperator();
					
					Iterator j = stats.getAllKnownOperators();
					for (; j.hasNext();) {
						XATStreamOperator temp = ((XATStreamOperator) j.next());
						if (temp
							.getProperty("OBJECT_ID")
							.equals(operatorTo)) {
							operator = temp;
							break;
						}
					}

					OperatorQueue oq = new OperatorQueue(operator,c.getToQueueNumber());
					// Add the operator to the table if it bridges a network
					// connection. This will apply to Stream->Machine, 
					// Machine -> Application and Machine->Remote Machine
					if (operator != null) {
						inConnections.addConnection(operatorFrom, oq);
						System.out.println(inConnections);

					}
				} //endif
				else if(c instanceof LightWeightStats){
					LightWeightStats lws = (LightWeightStats) c;
					dExec.setLightweightStats(lws.getStats());
				}//endif
				else if (c instanceof DeActivate) {
					// nothing to do here but active the operator(s)
					// the queues are already connected

					operatorNumber = c.getFromOperator();

					Iterator j = stats.getAllKnownOperators();
					for (; j.hasNext();) {
						XATStreamOperator temp = ((XATStreamOperator) j.next());
						if (temp
							.getProperty("OBJECT_ID")
							.equals(operatorNumber)) {
							operator = temp;
							break;
						}
					}

					//De-activate the operator
					if (operator.isActive()) {
						operator.setActive(false);
					}
					
				} //endif
				else if (c instanceof StopSend) {

					//First get the operator to connect to the application
					operatorNumber = c.getFromOperator();

					Iterator j = stats.getAllKnownOperators();
					for (; j.hasNext();) {
						XATStreamOperator temp = ((XATStreamOperator) j.next());
						if (temp
							.getProperty("OBJECT_ID")
							.equals(operatorNumber)) {
							operator = temp;
							break;
						}
					}

					if(operator != null){
						outConnections.removeConnection(operator);
						System.out.println(outConnections);
					}

				} //endif
				else if (c instanceof SendState) {
					OperatorState[] opstate;
					
					//First get the operator to connect to the application
					operatorNumber = c.getFromOperator();

					// Now, where to send it. 
					String ipAddress = c.getToIPAddress();
					int port = c.getToPort();
					
					Iterator j = stats.getAllKnownOperators();
					for (; j.hasNext();) {
						XATStreamOperator temp = ((XATStreamOperator) j.next());
						if (temp
							.getProperty("OBJECT_ID")
							.equals(operatorNumber)) {
							operator = temp;
							break;
						}
					}

					if(operator.hasState()){
						opstate = operator.getState();
						Connection d = new State(opstate);
						Packet p = new Packet("state",d);
						Socket s = new Socket(ipAddress,port);
						ObjectOutputStream os = new ObjectOutputStream(s.getOutputStream());
						os.writeObject(p);
						os.close();
						s.close();
					
					}
				} //endif
				else if (c instanceof State) {

					//First get the operator to connect to the application
					operatorNumber = c.getToOperator();
					State tempState = (State) c;
					Iterator j = stats.getAllKnownOperators();
					for (; j.hasNext();) {
						XATStreamOperator temp = ((XATStreamOperator) j.next());
						if (temp
							.getProperty("OBJECT_ID")
							.equals(operatorNumber)) {
							operator = temp;
							break;
						}
					}

					if(operator != null){
						operator.setState(tempState.getState());
					}

				} //endif
				else if (c instanceof ConnectionSeperator){
					
				}
				else{
					System.out.println("The Connection: " + c + " is not supported by the Connection Listener.");
				}
			}
		} catch (ConnectException ce) {
			System.out.println("==========ConnectException=========");
			ce.printStackTrace();
			
			System.out.println("===================================");
		} 
		catch (EOFException eofe) {
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public void requestStop() {
		keepGoing = false;
	}
}
