package edu.wpi.cs.dsrg.xmldb.xat.component.gui;

import java.awt.Toolkit;
import java.awt.Dimension;
import java.awt.Point;
import javax.swing.JFrame;

import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.SystemObject;

/** This <code>CAPE_GUI</code> runs on query machines.
 * Unlike the "<code>DISTRIBUTION_MANAGER</code>" <code>CAPE_GUI</code>, it does not
 * display labels on the operators to show which machine that specific operator is running on.
 * An overview of <code>CAPE_GUI</code> features:
 * <ul>
 * <li><code>BasicGraphModelMaker</code> - makes the <code>diva.graph.basic.BasicGraphModel</code>s
 * for each of the <code>CAPE_GUI</code>'s loaded trees.
 * <li><code>FramesManager</code> - manages all of the open frames
 * and the making of new frames to display.
 * <li><code>GraphManager</code> - manages open graphs and the drawing of the
 * graphs to the tabbed pane.
 * <li><code>OptionsStorage</code> - stores all of the user's selected/deselected
 * options.
 * <li><code>PanelMaker</code> - makes all of the <code>CAPE_GUI</code>'s panels.
 * <li><code>StatisticStorage</code> - stores all statistics for the tree(s), queue(s), 
 * operator(s) and system.
 * <li><code>TableMaker</code> - handles making all of the tables.
 * <li><code>TreeStorage</code> - stores all of the operator(s) and queue(s) of all loaded trees.
 * </ul>
 * 
 * @author Maylene Waltz
 */
public class CAPE_GUIQueryEngine extends CAPE_GUI implements Runnable{
	
	/** The <code>CAPE_GUI</code>'s name is "QUERY_ENGINE".
	 * This flags this <code>CAPE_GUI</code> to NOT include machine names on the operators
	 * to identify which machine that specific operator is running on. */
	private String _GUIName = new String("QUERY_ENGINE");
	
	/** Constructor initializes the <code>CAPE_GUI</code>.
	 * Note that no trees have been loaded at this point. */
	public CAPE_GUIQueryEngine(){
		initGUI();
	}
	
	/** Constructor loads the given <code>trees</code>, sets the
	 * <code>CAPE_GUI</code>'s client name to the given <code>clientName</code>
	 * and initializes the <code>CAPE_GUI</code>.
	 * @param trees the trees to be loaded into the <code>CAPE_GUI</code>
	 * @param clientName the name of the client query engine machine. For example, "Machine 1". */
	public CAPE_GUIQueryEngine(XATTree[] trees, String clientName){
		this._clientName = clientName;
		this._XATTrees = trees;
		initGUI();
	}
	
	/** Constructor loads the given <code>trees</code>, sets the
	 * <code>CAPE_GUI</code>'s client name to the given <code>clientName</code>,
	 * sets the <code>CAPE_GUI</code>'s system object to the given <code>systemObject</code>
	 * and initializes the <code>CAPE_GUI</code>.
	 * @param trees the trees to be loaded into the <code>CAPE_GUI</code>
	 * @param clientName the name of the client query engine machine. For example, "Machine 1"
	 * @param systemObject the object used to gather system statistics */
	public CAPE_GUIQueryEngine(XATTree[] trees, String clientName, SystemObject systemObject){
		this._XATTrees = trees;
		this._clientName = clientName;
		this._systemObject = systemObject;
		initGUI();
	}
	
	/** Returns the name of the <code>CAPE_GUI</code>. For example, if this <code>CAPE_GUI</code>
	* runs on the distribution manager, then the name would be "<code>DISTRIBUTION_MANAGER</code>".
	* The <code>CAPE_GUI</code> named "<code>DISTRIBUTION_MANAGER</code>" is flagged for the <code>BasicGraphModelMaker</code>
	* to add labels to the operators identifying which machine that specific operator is running on.
	* <code>CAPE_GUI</code>s running on quey machines are named
	* "<code>QUERY_ENGINE</code>" and lack this machine identifier.
	* @return the name of the <code>CAPE_GUI</code> */
	public String getName(){
		return _GUIName;
	}
	
	/** Initializes the <code>CAPE_GUI</code>. */
	public void init(){
		//Set the look and feel
		JFrame.setDefaultLookAndFeelDecorated(false);
		
		//Create and setup window
		JFrame frame = new JFrame("CAPE: " + this._clientName);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		
		//Create and setup content pane
		CAPE_GUIQueryEngine gui = this;
		gui.setOpaque(true);
		frame.setJMenuBar(gui.getJMenuBar());
		frame.setContentPane(gui);
		
		//Display window
		frame.pack();
		Toolkit tk = Toolkit.getDefaultToolkit();
		frame.setSize(new Dimension(new Double(tk.getScreenSize().getWidth()*0.7).intValue(), 
			new Double(tk.getScreenSize().getHeight()*0.8).intValue()));
		frame.setLocation(new Point(new Double(tk.getScreenSize().getWidth()*0.15).intValue(),
			new Double(tk.getScreenSize().getHeight()*0.1).intValue()));
		frame.setVisible(true);
		frame.toFront();
		
		//Draw the trees
		getSplitPane().setDividerLocation(0.60);
		try{
			for(int i=0; i<getXATTreeCount(); i++){
				getGraphManager().drawModel(getModelMaker().makeModel(getXATTree(i)),getXATTree(i));
			}
		}catch(Exception e){System.err.println("No trees have been loaded.");}
	
		getStatisticStorage().initialize();
		makeUpdateQueuesTimer().start();
		makeUpdateNodesTimer().start();
		makeUpdateStatisticTableTimer();
	}
	
	/** Main method. */
	public static void main(String[] args){
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
			CAPE_GUIQueryEngine gui = new CAPE_GUIQueryEngine();
			
			public void run() {
				gui.init();
			}
		 });
	}
	
	/** Run method. */
	public void run(){
		javax.swing.SwingUtilities.invokeLater(new Runnable() {
			 public void run() {
				 init();
			 }
		 });	
	}
}