package edu.wpi.cs.dsrg.xmldb.xat.component.gui.listeners;

import diva.graph.GraphController;
import diva.graph.basic.BasicGraphModel;
import diva.canvas.interactor.AbstractInteractor;
import diva.canvas.event.LayerEvent;
import diva.canvas.Figure;
import diva.canvas.FigureLayer;
import diva.canvas.event.EventLayer;

import javax.swing.JScrollPane;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils.JTabbedPaneCAPE;

/** The <code>StatsInteractor</code> handles user interaction with 
 * nodes and queues. When the user clicks a node or queue
 * a statistic table opens for the node or queue, respectively.
 * 
 * @author Maylene Waltz
 */
public class StatsInteractor extends AbstractInteractor{
	
	/** The event layer */
	private EventLayer _eventLayer;

	/** The figure layer */
	private FigureLayer _figureLayer;

	/** The graph controller that manages this interactor.
	 * Every graph has its own graph controller. */
	private GraphController _controller;
    
	/** Create a new <code>StatsInteractor</code> and give it a pointer
	 * to its <code>controller</code> so it can find other useful objects */
	public StatsInteractor(GraphController controller){
		_controller = controller;
		setEventLayer(_controller.getGraphPane().getBackgroundEventLayer());
		setFigureLayer(_controller.getGraphPane().getForegroundLayer());
	}
	
	/** Get the layer that drag events are listened on */
	public EventLayer getEventLayer () {
		return _eventLayer;
	}

	/** Get the layer that figures are selected on */
	public FigureLayer getFigureLayer () {
		return _figureLayer;
	}
	
	/** Handles the mouse pressed event. */
	public void mousePressed(LayerEvent e){
		if(!isEnabled()) return;
		
		// Consume the event
		if (isConsuming()){
			e.consume();
		}
	}	
	
	/** Handles the mouse release event. 
	 * A statistic table is opened for the node or queue
	 * that was clicked. */
	public void mouseReleased(LayerEvent e){
		if(!isEnabled()) return;
		
		//Get the figure that the event occured on
		Figure figure = e.getFigureSource();
			
		//Get the semantic object of the figure and
		//focus to the tab with the statistcs or make a new tab
		if(figure != null){
			openStats(_controller.getGraphModel().getSemanticObject(figure.getUserObject()));
		}
			
		// Consume the event
		if (isConsuming()){
			e.consume();
		}
	}	
	
	/** If the statistics table is already open for the given <code>semanticObject</code>,
	 * then focus to that tab. Otherwise, make a new tab and create
	 * the statistics table for the given <code>semanticObject</code>.
	 * @param semanticObject the semantic object that was clicked */
	public void openStats(Object semanticObject){
		//get the statistic tabbed pane
		JTabbedPaneCAPE statsTabbedPane = (JTabbedPaneCAPE)_controller.getStatsTabbedPane();
		
		//get the graph model in order to retreive the "TREE_ID"
		BasicGraphModel model = (BasicGraphModel)_controller.getGraphModel();
		
		String title = new String("");
		try{
			//if a node was right-clicked
			if(semanticObject instanceof XATNode){
				DagNode nodePeer = (DagNode)semanticObject;
				//Construct the tab title to compare to the open tabs
				title = new String("T" + model.getProperty(model.getRoot(), "TREE_ID") + "_O" 
					+ nodePeer.getOperator().getProperty("OBJECT_ID"));
					semanticObject = ((XATNode)semanticObject).getOperator();
			}//else if a queue was right-clicked
			else if(semanticObject instanceof XATQueue){
				XATQueue queuePeer = (XATQueue)semanticObject;
				//Construct the tab title to compare to the open tabs
				//if this queue is an input stream then make the title:
				//QueueName+"_O"+OBJECT_ID+"_Q"+QueueNumber  Example: Nile_O2_Q0
				if(queuePeer.getQueueName().length() > 0){
					title = new String(queuePeer.getQueueName()+"_O"+
							((XATStreamOperator)queuePeer.getQueueOwner().get(0)).getProperty("OBJECT_ID")
							+"_Q"+queuePeer.getQueueNumber());
				}
				else{
					title = new String("T" + model.getProperty(model.getRoot(), "TREE_ID")
						+ "_O"+((XATStreamOperator)((XATQueue)queuePeer).getQueueOwner().get(0)).getProperty("OBJECT_ID")
						+ "_Q" + queuePeer.getQueueNumber());
				}
			}
		}catch(Exception e){}

		boolean makeNewTab = true;
		//if the tab is already open, then focus to that tab
		for(int i=0; i<statsTabbedPane.getTabCount(); i++){
			if(statsTabbedPane.getTitleAt(i).equals(title)){
				statsTabbedPane.setSelectedIndex(i);
				makeNewTab = false;
				break;
			}
		}
		if(makeNewTab){
			JScrollPane tablePane = _controller.getGUI().getTableMaker().makeTable(semanticObject);
			statsTabbedPane.addTab(title, semanticObject, tablePane, "STATS");
			statsTabbedPane.setSelectedIndex(statsTabbedPane.getTabCount()-1);
			_controller.getGUI().getUpdateStatisticTableTimer().start();
		}		
	}
	
	/**
	 * Set the layer that drag events are listened on
	 */
	public void setEventLayer(EventLayer l) {
		if (_eventLayer != null) {
			_eventLayer.removeLayerListener(this);
		}
		_eventLayer = l;
		_eventLayer.addLayerListener(this);
	}

	/**
	 * Set the layer that figures are selected on
	 */
	public void setFigureLayer(FigureLayer l) {
		_figureLayer = l;
	}
}