package edu.wpi.cs.dsrg.xmldb.xat.component.gui.storage;

import java.util.HashMap;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Enumeration;
import java.sql.Timestamp;

import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;

/** The <code>StatisticStorage</code> class manages all of the statistics. 
 * The first <code>HashMap</code>'s 'key' is the <code>XATQueryObject</code>
 * representing a:
 * <ul>
 * <li>Tree
 * <li>Queue
 * <li>Operator
 * <li>System
 * </ul>
 * and the 'value' is another <code>HashMap</code>. This second
 * <code>HashMap</code>'s 'key' is the statistic name and 'value'
 * is an <code>ArrayList</code> of statistic values.
 * 
 * 
 * @author Maylene Waltz
 */
public class StatisticStorage{
	
	/** The CAPE_GUI. */
	private CAPE_GUI _gui;
	
	/** <code>HashMap</code> of statistics where the 'key' is the 
	 * <code>XATQueryObject</code> and the 'value' is another <code>Hashmap</code>.
	 * This second <code>HashMap</code>'s 'key' is the statistic name
	 * and the 'value' is an <code>ArrayList</code> of statistic values. */
	private HashMap _stats;
	
	/** Initializes the <code>StatisticStorage HashMap</code>. */
	public StatisticStorage(CAPE_GUI gui){
		_stats = new HashMap();
		_gui = gui;
	}
	
	/** Returns an <code>ArrayList</code> of statistic values for the given
	 * <code>key</code> and <code>statistic</code>.
	 * @param key - edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
	 * @param statistic - java.lang.String
	 * @return ArrayList - An <code>ArrayList</code> of statistic values. */
	public Object getStatistic(Object key, String statistic){
		ArrayList values = new ArrayList();
		try{
			values = (ArrayList)((HashMap)_stats.get(key)).get(statistic);
		}catch(Exception e){System.err.println("StatisticStorage: Error retrieving statistic from storage.");}
		return values;
	}
	
	/** Initializes the <code>StatisticStorage HashMap</code> keys to 
	 * all of the trees, queues and operators and values to <code>HashMap</code>s.
	 * The second <code>HashMap</code>'s keys are the statistic names and values
	 * are empty <code>ArrayList</code>s to hold the statistic values.
	 * @param tree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree */
	public void initialize(){
		try{
			//put all trees into storage with a HashMap of tree statistics
			for(Iterator trees = _gui.getTreeStorage().getAllTrees(); trees.hasNext();){
				XATTree tree = (XATTree)trees.next();
				HashMap treeStats = new HashMap();
				//put each tree statistc into this tree's HashMap
				for(Enumeration stats = tree.getStatistics(); stats.hasMoreElements();){
					treeStats.put(stats.nextElement(), new ArrayList());
				}
				_stats.put(tree, treeStats);
			}
			//put all queues into storage with a HashMap of queue statistics
			for(Iterator queues = _gui.getTreeStorage().getAllQueues(); queues.hasNext();){
				XATQueue queue = (XATQueue)queues.next();
				HashMap queueStats = new HashMap();
				//put each queue statistc into this queue's HashMap
				for(Enumeration stats = queue.getStatistics(); stats.hasMoreElements();){
					queueStats.put(stats.nextElement(), new ArrayList());
				}
				_stats.put(queue, queueStats);
			}
			//put all operators into storage with a HashMap of operator statistics
			for(Iterator ops = _gui.getTreeStorage().getAllOperators(); ops.hasNext();){
				XATStreamOperator op = (XATStreamOperator)ops.next();
				HashMap opStats = new HashMap();
				//put each operator statistc into this operator's HashMap
				for(Enumeration stats = op.getStatistics(); stats.hasMoreElements();){
					opStats.put(stats.nextElement(), new ArrayList());
				}
				_stats.put(op, opStats);
			}
			//put system object into storage
			HashMap sysStats = new HashMap();
			//put each system statistic into this system's HashMap
			for(Enumeration stats = _gui.getSystemObject().getStatistics(); stats.hasMoreElements();){
				sysStats.put(stats.nextElement(), new ArrayList());
			}
			_stats.put(_gui.getSystemObject(), sysStats);
		}catch(Exception e){System.err.println("StatisticStorage: Error initializing.");}
	}
	
	/** Puts the given statistic <code>value</code> in the <code>ArrayList</code> corresponding
	 * to the given <code>queryObject</code> and <code>statistic</code>.
	 * The current system time is paired with the statistic value.
	 * @param queryObject - edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject
	 * @param statistic - java.lang.String 
	 * @param value - java.lang.String */
	private void setStatisticValue(XATQueryObject queryObject, String statistic, String value){
		try{
			Timestamp ts = new Timestamp(System.currentTimeMillis());
			((ArrayList)((HashMap)_stats.get(queryObject)).get(statistic)).add(new StatisticValue(value, ts));
		}catch(Exception e){System.err.println("StatisticStorage: Error setting statistic value.");}
	}
	
	/** Updates all of the statistic values for the given <code>queryObject</code>.
	 * @param queryObject - edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject */
	public void update(Object queryObject){
		StatisticsGatherer myG = StatisticsGatherer.getStatisticsGather();
		try{
			//if this is a node, need to get operator for statistics
			if(queryObject instanceof XATNode){
				queryObject = ((XATNode)queryObject).getOperator();
			}
			//for each statistic that this queryObject has
			for(Iterator stats = ((HashMap)_stats.get(queryObject)).keySet().iterator(); stats.hasNext();){
				String stat = (String)stats.next();
				setStatisticValue((XATQueryObject)queryObject, stat, myG.getValue(stat, (XATQueryObject)queryObject));
			}
		}catch(Exception e){System.err.println("StatisticStorage: Error updating statistic(s).");}
	}
}