package edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils;

import java.util.ArrayList;
import java.util.Vector;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNodeImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATMemoryQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATMemoryQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATPersistentQueueImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;

import diva.canvas.toolbox.LabelFigure;
import diva.graph.basic.BasicGraphModel;
import diva.graph.modular.Node;
import diva.graph.modular.Edge;
import diva.graph.modular.Graph;

/** The <code>BasicGraphModelMaker</code> class makes the <code>diva.graph.basic.BasicGraphModel</code>
 * for the given <code>XATTree tree</code>. It constructs additional nodes and edges to represent the
 * input and output streams of the query plan. All of the tree's operators and queues are stored in 
 * <code>TreeStorage</code>.
 * 
 * @author Maylene Waltz
 */
public class BasicGraphModelMaker{
	
	/** The CAPE_GUI */
	private CAPE_GUI _gui;	
	
	/** Constructor associates the given <code>gui</code> with this model maker. */
	public BasicGraphModelMaker(CAPE_GUI gui){
		_gui = gui;
	}
	
	/** Constructs the <code>diva.graph.basic.BasicGraphModel</code> for the given <code>tree</code>.
	 * All of the <code>tree</code> operators and queues are stored in <code>TreeStorage</code>. 
	 * If the <code>CAPE_GUI</code> is a "DISTRIBUTION_MANAGER" <code>CAPE_GUI</code>, then labels are added to the nodes to 
	 * display which machine the node is running on.
	 * @param tree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
	 * @return the graph model of the given <code>tree</code> */
	public BasicGraphModel makeModel(XATTree tree){
		//create the graph model for the given tree
		BasicGraphModel model = new BasicGraphModel();
		
		//all of the tree's operators
		ArrayList treeOps = new ArrayList();
		
		//all of the tree's queues
		ArrayList treeQs = new ArrayList();
		
		XATNode[] leaves = tree.getLeaves();
		if(leaves.length > 0){ //if the tree has nodes display them
			Vector resultSet = new Vector();
			try{
				//Do breath first traverse
				Set curSet = new HashSet();//contains nodes on current visiting level
				Set nextSet = new HashSet();//contains nodes on next visiting level

				//traversal begins from the leaves(s)
				for(int i=0; i<leaves.length; i++){
					curSet.add(leaves[i]);
					do{ //for each node in current level
						for(Iterator it = curSet.iterator(); it.hasNext();){
							DagNode curNode = (DagNode)it.next();
						
							//visit this node
							resultSet.add(curNode);
							
							if(curNode.getChildrenSize() == 0){ //if the current node is a leaf
								//create visual leaf node
								Node leaf = model.createNode(curNode);
								
								//add this leaf to the graph's roots
								model.addRootNode(leaf);
								
								//add this node/operator to the ArrayList of operators
								treeOps.add(curNode.getOperator());
								
								//set the visual leaf node to the DagNode it represents
								curNode.setVisualObject(leaf);
								
								//set the leaf node's label
								LabelFigure nodeLabel = new LabelFigure(String.valueOf(curNode.getOperator().getProperty("OBJECT_ID")));
								leaf.setLabelFigure(nodeLabel);
								leaf.setToolTipText(((XATStreamOperator)curNode.getOperator()).toToolTipString());
								
								if(_gui.getName().equals("DISTRIBUTION_MANAGER")){
									LabelFigure machineLabel = new LabelFigure("M");
									leaf.setLabelFigureMachineID(machineLabel);
								}
								
								//create node(s) to represent input stream(s)
								//If the current input queue count is zero, its because we havent
								//connected the operator to streams yet.
								XATStreamOperator curOp = (XATStreamOperator) curNode.getOperator();
								
								if (curOp.getInputQueueCount() == 0) {
									Iterator iter = curNode.getStreamNames();
									StatisticsGatherer stats = StatisticsGatherer.getStatisticsGather();
									while(iter.hasNext()){
										String currStream = (String) iter.next();
										XATQueue newQ = new XATMemoryQueueImp(curOp,currStream);
										//XATQueue newQ = new XATQueueImp(curOp,currStream);
										//XATQueue newQ;
										if(System.getProperty("DATA_MODEL").endsWith("XATMemoryQueueImp")) {
											newQ = new XATMemoryQueueImp(curOp,currStream);
										}
										else if (System.getProperty("DATA_MODEL").endsWith("XATPersistentQueueImp")) {
											newQ = new XATPersistentQueueImp(curOp,currStream);
										}
										else {
											System.out.println("!!! Error: Unknow type of DATA_MODEL !!!");
											System.exit(-1);
											return null;
										}
										curOp.addInputQueue(newQ);
										stats.addNewQueryObject(newQ);
									}
								}
								
								for(Iterator ie = ((XATStreamOperator)curNode.getOperator()).getInputQueues(); ie.hasNext();){
									XATQueue inputQ = (XATQueue)ie.next();
									
									//add the queue to the tree's ArrayList of queues
									treeQs.add(inputQ);
									
									//create node for the input queue to originate from
									DagNode leafInputDagNode = new DagNodeImp();
									Node leafInputNode = model.createNode(leafInputDagNode);
									leafInputDagNode.setVisualObject(leafInputNode);
									LabelFigure inputLabel = new LabelFigure(inputQ.getQueueName());
									leafInputNode.setLabelFigure(inputLabel);
									leafInputNode.setToolTipText("Input Stream");
									model.addNode(this, leafInputNode, (Graph)model.getRoot());
									
									//create an edge signifying the input queue
									Edge edge = model.createEdge(inputQ);
									LabelFigure queueLabel = new LabelFigure(String.valueOf(inputQ.getQueueSize()));
									queueLabel.setSize(10);
									edge.setLabelFigure(queueLabel);
									//@TODO Set Queue Tool Tip Text to what??
									edge.setToolTipText(inputQ.toString());
									model.connectEdge(this, edge, leafInputNode, leaf);
								}
		
								//add the visual leaf node to the graph
								model.addNode(this, leaf, (Graph)model.getRoot());							
							} else {
								//create visual current node
								Node currentNode = model.createNode(curNode);
								
								//add this node/operator to the ArrayList of operators
								treeOps.add(curNode.getOperator());
								
								//set the visual current node to the DagNode it represents
								curNode.setVisualObject(currentNode);
								
								//set the node's label
								LabelFigure nodeLabel = new LabelFigure(String.valueOf(curNode.getOperator().getProperty("OBJECT_ID")));
								currentNode.setLabelFigure(nodeLabel);
								currentNode.setToolTipText(((XATStreamOperator)curNode.getOperator()).toToolTipString());
								
								//display the machine the node is/will be running on
								if(_gui.getName().equals("DISTRIBUTION_MANAGER")){
									LabelFigure machineLabel = new LabelFigure("M");
									currentNode.setLabelFigureMachineID(machineLabel);
								}
			
								//add the visual current node to the graph
								model.addNode(this, currentNode, (Graph)model.getRoot());
								
								//if the current node is a root node, then create node(s) to represent output stream(s)
								if(curNode.getParentSize() == 0){
									//for each output queue make a node and connecting edge with queue size
									for(Iterator ie = ((XATStreamOperator)curNode.getOperator()).getOutputQueues(); ie.hasNext();){
										XATQueue outputQ = (XATQueue)ie.next();
										
										//add the queue to the tree's ArrayList of queues
										treeQs.add(outputQ);
										
										DagNode rootOutputDagNode = new DagNodeImp();
										Node rootOutputNode = model.createNode(rootOutputDagNode);
										rootOutputDagNode.setVisualObject(rootOutputNode);
										LabelFigure outputLabel = new LabelFigure("Output");
										rootOutputNode.setLabelFigure(outputLabel);
										rootOutputNode.setToolTipText("Output Stream");
										model.addNode(this, rootOutputNode, (Graph)model.getRoot());
									
										//create an edge signifying the output queue
										Edge edge = model.createEdge(outputQ);
										LabelFigure queueLabel = new LabelFigure(String.valueOf(outputQ.getQueueSize()));
										edge.setLabelFigure(queueLabel);
										//@TODO Set Queue Tool Tip Text to what??
										edge.setToolTipText(outputQ.toString());
										model.connectEdge(this, edge, currentNode, rootOutputNode);
									}
								}
								
								//get the current node's children
								XATNode[] children = curNode.getChildren();
								for(int k=0; k<children.length; k++){
									DagNode dagChild = (DagNode)children[k];
									//for each input queue of the current node check if it
									//connects to the current child
									for(Iterator inputQs = ((XATStreamOperator)curNode.getOperator()).getInputQueues(); inputQs.hasNext();){
										XATQueue inputQ = (XATQueue)inputQs.next();
										//if this queue connects this child to the current node
										if(inputQ.isOwner(dagChild.getOperator())){
											//add the queue to the tree's ArrayList of queues
											treeQs.add(inputQ);
											//then create the connecting edge
											Edge edge = model.createEdge(inputQ);
											//set the edge's label
											LabelFigure edgeLabel = new LabelFigure(String.valueOf(inputQ.getQueueSize()));
											edge.setLabelFigure(edgeLabel);
											//@TODO Set Queue Tool Tip Text to what??
											edge.setToolTipText(inputQ.toString());
											//add the edge to the graph
											model.connectEdge(this, edge, dagChild.getVisualObject(), currentNode);
										}
									}		
								}//end for
								
								//add input stream nodes and edges if this node connects to input stream(s)
								if(curNode.getStreamSize() > 0){
									for(Iterator ie = ((XATStreamOperator)curNode.getOperator()).getInputQueues(); ie.hasNext();){
										XATQueue inputQ = (XATQueue)ie.next();
										String name = inputQ.getQueueName();
										
										//add the queue to the tree's ArrayList of queues
										treeQs.add(inputQ);
									
										if(name.length() > 0){
											//create node for the input queue to originate from
											DagNode leafInputDagNode = new DagNodeImp();
											Node leafInputNode = model.createNode(leafInputDagNode);
											leafInputDagNode.setVisualObject(leafInputNode);
											LabelFigure inputLabel = new LabelFigure(inputQ.getQueueName());
											leafInputNode.setLabelFigure(inputLabel);
											model.addNode(this, leafInputNode, (Graph)model.getRoot());
										
											//create an edge signifying the input queue
											Edge edge = model.createEdge(inputQ);
											LabelFigure queueLabel = new LabelFigure(String.valueOf(inputQ.getQueueSize()));
											edge.setLabelFigure(queueLabel);
											edge.setLabelFigure(queueLabel);
											//@TODO Set Queue Tool Tip Text to what??
											edge.setToolTipText(inputQ.toString());
											model.connectEdge(this, edge, leafInputNode, currentNode);
										}
									}
								}
							}
							
							//if there are more parents to visit
							if (curNode.getParentsSize() > 0) {
								for(Iterator p = curNode.getParents(); p.hasNext();){
								//push each parent in next level
									DagNode curParent = (DagNode)p.next();
									//flag == true means all parents are visited
									boolean flag = true;
									XATNode[] parents = curParent.getChildren();
									for(int j=0; j<parents.length; j++){
										DagNode curPrev = (DagNode)parents[j];
										if (!resultSet.contains(curPrev)) {
											flag = false;
											break;
										}//end if
									}//end for
									if (flag) //flag == true means all parents are visited
										nextSet.add(curParent);
								}//end for: push each parent				
							}//end if: push next level
						}//for
					curSet = nextSet;
					nextSet = new HashSet();
					} while (!curSet.isEmpty());
					}//end for: roots iterator
				} catch (Exception e) {
					e.printStackTrace();
				}//end catch
			}//end if
		model.setProperty(model.getRoot(), "TREE_ID", tree.getProperty("TREE_ID"));
		_gui.getTreeStorage().setOperators(tree, treeOps);
		_gui.getTreeStorage().setQueues(tree, treeQs);
		return model;
	}
		
	
}