package edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils;

import ptolemy.plot.PlotLive;

import java.util.Iterator;
import java.util.ArrayList;

import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.SystemObject;

/** StatsPlotLive extends PlotLive from the Ptplot 5.3 - Java Plotter package from 
 * the Ptplot team at UC Berkeley, EECS department. StatsPlotLive plots user-selected
 * statistics for each user-selected tree, queue, operator, or system and updates the plot
 * every user-specified milliseconds.
 * 
 * @author Maylene Waltz
 */ 
public class StatsPlotLiveTest extends PlotLive {
	
	/** The update frequency in milliseconds. This specifies how often to add a new point
	 * to the plot that represents the latest statistic value for each data set. */
	private long _freq;
	
	/** The position on the x-axis to plot the data point. */
	private double x_pos;
	
	/** Vector of Strings containing the statistics to plot. */
	private ArrayList _stats;
	
	/** Vector of <code>XATQueryObject</code>s to plot statistics for. */
	private ArrayList _queryObjects;
	
	/** The <code>SystemObject</code> to plot statistics for. */
	private SystemObject _systemObject;
	
	/** X-axis range minimum */
	final private int X_MIN = 0;
	
	/** X-axis range maximum */
	final private int X_MAX = 100;
	
	/** Y-axis range minimum */
	final private int Y_MIN = 0;
	
	/** Y-axis range maximum */
	final private int Y_MAX = 100;
	
	/** Plot graph width */
	final private int PLOT_WIDTH = 600;
	
	/** Plot graph height */
	final private int PLOT_HEIGHT = 300;
	
	/** Plot title */
	final private String TITLE = "CAPE Plot";
	
	/** X-axis label */
	final private String X_LABEL = "Time";
	
	/** Y-axis label */
	final private String Y_LABEL = "Statistics";
	
	/** Constructor that initializes the graph x and y range, mark style and
	 * shows the buttons, such as print and resize plot. */
	public StatsPlotLiveTest(){
		setYRange(0, 1);
		setXRange(0, 1);
		setMarksStyle("dots");
		setButtons(true);
	}
	
	/** Constructor that initializes the graph x and y range, mark style, 
	 * shows the buttons, such as print and resize plot, and sets the update frequency to 
	 * the given <code>freq</code>. */
	public StatsPlotLiveTest(int freq){
		setYRange(0, 1);
		setXRange(0, 1);
		setMarksStyle("dots");
		setButtons(true);
		_freq = freq;
	}
	
	/** Constructor that initializes the graph and legend to plot the given
	 * <code>stats</code> and <code>queryObject</code> and sets the update
	 * frequency to the given <code>freq</code>.
	 * @param stats the statistic(s) to be plotted
	 * @param queryObjects the query objects to plot statistic(s) for
	 * @param freq how often (in milliseconds) to add a point to the plot */
	public StatsPlotLiveTest(Object stats, Object queryObjects, String freq){
		if(stats instanceof ArrayList)_stats = (ArrayList)stats;
		if(queryObjects instanceof ArrayList) _queryObjects = (ArrayList)queryObjects;
		else if(queryObjects instanceof SystemObject){
			_queryObjects = new ArrayList();
			_queryObjects.add(queryObjects);
		}
		_freq = Long.parseLong(freq);
		initLegend();
		initGraph();
	}
	
	/** Makes a legend of all the statistics. */
	private void initLegend(){
		int count = 0;
		//Constructs a key for the legend in the following format: 
		//XATQueryObject:Statistic for each queryObject
		for(Iterator e = _queryObjects.iterator(); e.hasNext();){
			XATQueryObject qo = ((XATQueryObject)e.next());
			//for each statistic
			for(Iterator e2 = _stats.iterator(); e2.hasNext();){
				//Make some replacements for shorter legend names
				String name = e2.next().toString();
				name = name.replaceAll("NUMBER", "#");
				
				String key = new String("");
				if(qo instanceof XATTree){
					key += "T"+qo.getProperty("TREE_ID")+":";
				}else if(qo instanceof XATQueue){
					key += "Q"+((XATQueue)qo).getQueueNumber()+":";
				}else if(qo instanceof XATStreamOperator){
					key += "O"+qo.getProperty("OBJECT_ID")+":";
				}else if(qo instanceof SystemObject){
					key += "S:";
				}
				key += name;
				addLegend(count, key);
				count++;
			}
		}
	}
	
	/** Initializes the plot by setting:
	 * <ul>
	 * <li>the plot title
	 * <li>the size of the plot (width and height)
	 * <li>x and y minimum and maximum range values
	 * <li>x and y axis labels
	 * <li>whether or not to display buttons (i.e. print button) and
	 * <li>the mark style (i.e. "dots").
	 * </ul> */
	private void initGraph(){
		//@TODO make a loading... popup while graph initializes
		setTitle(TITLE);
		setSize(PLOT_WIDTH, PLOT_HEIGHT);
		setXRange(X_MIN, X_MAX);
		setYRange(Y_MIN, Y_MAX);
		setXLabel(X_LABEL);
		setYLabel(Y_LABEL);
		setButtons(true); //display print, resize, etc. buttons
		setMarksStyle("dots");
	}
	
	/** Adds a new point to the plot every user-specified millisecond. */
	public synchronized void addPoints() {
		StatisticsGatherer myG = StatisticsGatherer.getStatisticsGather();
		int dataset = 0;
		//for each queryObject
		for(Iterator e = _queryObjects.iterator(); e.hasNext();){
			XATQueryObject qo = (XATQueryObject)e.next();
			//for each statistic
			for(Iterator e2 = _stats.iterator(); e2.hasNext();){
				addPoint(dataset, x_pos, Double.parseDouble(myG.getValue((String)e2.next(), qo)), true);
				//e2.next();
				//addPoint(dataset, x_pos, Math.random(), true);
				dataset++;
			}
		}
		x_pos += _freq;
		fillPlot();
		try {
			Thread.sleep(_freq);
		} catch (InterruptedException e) {}
	}

}