package edu.wpi.cs.dsrg.xmldb.xat.component.gui.utils;

import java.util.ArrayList;
import javax.swing.JCheckBox;
import javax.swing.table.AbstractTableModel;
import java.io.File;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.XATQueue;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.SystemObject;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.CAPE_GUI;
import edu.wpi.cs.dsrg.xmldb.xat.component.gui.storage.*;

/** The table model for statistic tables. Extends <code>javax.swing.table.AbstractTableModel</code>
 * to add a constructor with a <code>CAPE_GUI</code> parameter. 
 * The <code>CAPE_GUI</code> parameter is required to retrieve the <code>StatisticStorage</code>
 * object for the <code>update()</code> table method. The <code>TableMaker</code> class uses 
 * this table model to make tables and the update function provides a table model with updated statistic values.
 * 
 * @author Maylene Waltz
 */
public class StatsTableModel extends AbstractTableModel{
	
	/** The table's column headers. */
	private String[] columnHeaders;
	
	/** The table's data. */
	private Object[][] data;
	
	/** The <code>CAPE_GUI</code>. */
	private CAPE_GUI _gui;
	
	/** Constructor associates the given <code>gui</code> with this table model. 
	 * @param gui the <code>CAPE_GUI</code> to associate with this table model */
	public StatsTableModel(CAPE_GUI gui){
		_gui = gui;
	}
	
	/** Returns the number of columns in the table. 
	 * @return the number of columns in the table */
	public int getColumnCount(){
		return columnHeaders.length;
	}
	
	/** Returns the table's column headers.
	 * @return the table's column headers */
	public String[] getColumnHeaders(){
		return columnHeaders;
	}
	
	/** Returns the number of rows in the table.
	 * @return the number of rows in the table */
	public int getRowCount(){
		return data.length;
	}
	
	/** Returns the <code>Object</code> located at the given <code>row</code>
	 * and <code>col</code>.
	 * @param row the row with the desired <code>Object</code>
	 * @param col the column with the desired <code>Object</code>
	 * @return the <code>Object</code> at the given row and column */
	public Object getValueAt(int row, int col){
		return data[row][col];
	}
	
	/** Updates the table's data. There is a row for each of the
	 * user's selected statistic(s) and a column for all available statistic values.
	 * The most recent statistic value is the left most column progressing to
	 * the right with earlier statistic values.
	 * @param queryObject the query object which we are updating the statistic table for
	 * @return the table's data */
	public Object[][] update(Object queryObject){
		ArrayList statBoxes = new ArrayList();//holds the queryObject's options
		try{
			if(queryObject instanceof XATTree){
				_gui.getPanelMaker().makePanel("SELECT_TREE_STATS");
				statBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("SELECT_TREE_STATS");
			}else if(queryObject instanceof XATQueue){
				_gui.getPanelMaker().makePanel("SELECT_QUEUE_STATS");
				statBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("SELECT_QUEUE_STATS");
			}else if(queryObject instanceof XATStreamOperator || queryObject instanceof XATNode){
				_gui.getPanelMaker().makePanel("SELECT_OPERATOR_STATS");
				statBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("SELECT_OPERATOR_STATS");
			}else if(queryObject instanceof SystemObject){
				_gui.getPanelMaker().makePanel("SELECT_SYSTEM_STATS");
				statBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("SELECT_SYSTEM_STATS");
			}else if (queryObject instanceof File) {
				_gui.getPanelMaker().makePanel("SELECT_FILE_STATS");
				//statBoxes = (ArrayList)_gui.getOptionsStorage().getOptions("SELECT_FILE_STATS");
			}else System.err.println("Unable to update the statistics for the given XATQueryObject.");
		}catch(Exception e){System.err.println("StatsTableModel: Error retrieving selected statistics from OptionsStorgage.");}
		
		ArrayList selectedStats = new ArrayList(); //holds the user's selected statistics
		ArrayList statValues = new ArrayList(); //holds all the statistic values
		
		//for every available statistic
		for(int i=0; i<statBoxes.size(); i++){
			//if the user wants to see this statistic in the stats table
			//then add to selectedStats ArrayList
			JCheckBox statBox = (JCheckBox)statBoxes.get(i);
			if(statBox.isSelected()){
				selectedStats.add(statBox.getText());
			}
		}
		selectedStats.trimToSize();
		
		//if there ARE selected statistics then make data[][]
		if(selectedStats.size() > 0){
			//tell StatisticStorage to update values for this queryObject
			_gui.getStatisticStorage().update(queryObject);
			
			//the # of columns = # of statistic values + 1(statistic name)
			int numColumns = ((ArrayList)_gui.getStatisticStorage()
				.getStatistic(queryObject, (String)selectedStats.get(0))).size()+1;
			
			//make the data[][] row size = the # of selected stats and 
			//the column size = to the # of statistic values
			//if only one column for the statistic name then add another column
			//and column header to add '0.0' (which occurs below).
			if(numColumns == 1){
				columnHeaders = new String[numColumns+1];
				data = new Object[selectedStats.size()][numColumns+1];
			}else{
				columnHeaders = new String[numColumns];
				data = new Object[selectedStats.size()][numColumns];
			}
			columnHeaders[0] = "Statistic";
			
			//for each row in the table populate with statistic values
			for(int i=0; i<data.length; i++){
				//set the first row value to the name of the statistic
				data[i][0] = selectedStats.get(i);
				int colCount = 0;//keeps track of the column to insert the current statistic value
			
				//get all statistic values for this query object
				statValues = (ArrayList)_gui.getStatisticStorage()
					.getStatistic(queryObject, (String)data[i][0]);
				statValues.trimToSize();
			
				//if there are no statistic values, then put '0.0'
				if(statValues.size() == 0){
					data[i][1] = new Double(0.0);
					columnHeaders[1] = "00:00:00";
				//otherwise, populate the row with all value(s)
				}else{
					colCount = statValues.size();
					for(int j=0; j<statValues.size(); j++){
						StatisticValue pair = (StatisticValue)statValues.get(j);
						data[i][colCount] = pair.getValue();
						columnHeaders[colCount] = pair.getTime();
						colCount--;
					}
				}
			}
		}
		//if the user has no selected statistics then display an empty table
		else{
			columnHeaders = new String[1];
			columnHeaders[0] = "No Statistics Selected.";
			data = new Object[1][1];
			data[0][0] = "";
		}
		return data;
	}
}