package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Random;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.LightWeightQueryObject;
import edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.ConfigFileParser;
/**
 * This will pick the next algorithm to run with probability equal
 * to the algorithm's score. 
 * @author: Brad Pielech
 */
public class ProbabilisticGreedyAdapter extends XATExecutionAdapter {
	protected java.util.Random rand;
	/**
		This will help to determine how much to decay the scores for each algorithm
		based on how many times other algorithms have been chosen since this one was run.  By
		doing it this way instead of by time, we do not have to worry about how often the scores
		are recalculated.  
	
	 key: scheduling algorithm's category
	 value: the number of times that new algorithms have been chosen since this
	 		algorithm was last run.  
	*/
	private Hashtable strategyRunCount;
	//this is equal to the SCORE_DELAY_FACTOR property and this value is set through that interface
	private double decayFactor;
/**
 * ProbabilisticGreedyAdapter constructor comment.
 * @param parser edu.wpi.cs.dsrg.xmldb.xat.component.executioncontroller.ConfigFileParser
 * @param tree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 */
public ProbabilisticGreedyAdapter(ConfigFileParser parser, edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree tree) {
	super(parser, tree);
	rand = new Random();
	//strategyRunCount = new Hashtable();
}
/**
 * After each time the next algorithm is chosen, we wish to decay the other algorithm's scores
 * because they are based on older data.  We have to be careful not to decay the scores too
 * quickly, otherwise one algorithm could easily dominate the scheduling
 *
 * This algorithm will iterate through all of the scores and decay all of the scores for
 * algorithms that were just not run.  
 */
private void decayScores() {
	
	for(Enumeration e = scores.keys(); e.hasMoreElements();){
		Object nextCategory = e.nextElement();
		//if we are not looking at the current category
		if(nextCategory != currentCategory){
			double score = ((Double)scores.get(nextCategory)).doubleValue();
			double timesSinceRun = ((Double)strategyRunCount.get(nextCategory)).doubleValue();
			//decay the score exponentially 
			score = score * Math.pow(decayFactor, timesSinceRun);
			scores.put(nextCategory, new Double(score));
		}
	}
	
	
	
	}
/**
 * Periodically the execution controller will request the Adapter to
 * change the algorithm b/c the previous one was not working well (ex: starvation).
 * Therefore, this method is similar to getNextScheduler, but getNext could result
 * in the same scheduler being used as the previous.  This method will guarentee a different
 * algorithm <p>
 * This can create a problem if there is only 1 algorithm available.  Therefore, it throws
 * an IllegalStateException if a switch is not possible. 
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATScheduler
 * @exception IllegalStateException if there is only 1 scheduler and a change is not possible
 */
public XATScheduler forceChange() throws java.lang.IllegalStateException {
	return getNextScheduler(true);
}
/**
 * Returns the index of the Scheduling algorithm that should be used next.
 * This method updates the currentCategory class variable and then
 * return the next scheduler using getCurrentScheduler().  <p>
 * This getNextScheduler will sum the scores of the schedulers and create a random double
 * between 0 and that count. Next to figure out which algorithm was selected, the scores will
 * again be summed and the score that places this 2nd count over the random variable will be chosen.  
 * @return XATScheduler the next algorithm to use.  
 */
public XATScheduler getNextScheduler() {
	return this.getNextScheduler(false);
}
/**
 * This is generalized method for forceChange and getNextScheduler.  The boolean
 * variable controls whether the next algorithm must be different than the previous
 * @return XATScheduler the next algorithm to use.  
 */
private XATScheduler getNextScheduler(boolean force) {
    if ((scores.size() == 0) && force == true) {
        throw new IllegalStateException("The Adapter cannot switch because there is only 1 algorithm to choose from");
    }    
    
    /** 
     * if we still havent run every scheduler, then do so by calling super.getNext
     * when there are non left, then just continue with the regular logic
    */
    if (isLeftToRun()) {
	    if(debug){
		    System.out.println("Still have some strategies left to try");
	    }
        return super.getNextScheduler();
    } else {
	    //only update scores if we have run each algorithm 
		updateScores();
		//now decay those just-updated-scores
		decayScores();
        double count = 0;
        //sum the scores
        for (Iterator i = scores.values().iterator(); i.hasNext();) {
            double next = ((Double) i.next()).doubleValue();
            if (next > 0) {
                count += next;
            }
        }
        Object old = currentCategory;
        while (true) {
            //get the random variable between 0 and count
            double nextValue = rand.nextDouble() * count;
            count = 0;

            //now walk over the scores again and see which one makes the sum greater than our target
            for (Enumeration e = scores.keys(); e.hasMoreElements();) {
                Object nextCat = e.nextElement();
                double next = ((Double) scores.get(nextCat)).doubleValue();
                count += next;
                if (count >= nextValue) {
                    /* 
                     * if we are supposed to change and the new and old are the same
                     * then loop again
                     */
                    if (force && (nextCat == old)) {
                        break;
                    } else {
                        currentCategory = nextCat;
                        break;
                    }
                }
            }
            /** 
             * if we are not forcing a change or the we changed algorithms, then return
             * otherwise keep looping
             */
            if ((!force) || (currentCategory != old)) {
	            //now update the run counts so we can decay scores next time 
	            updateRunCounts();
                return getCurrentScheduler();
            }
        }
    }
}
/** 
 * Initializes the strategyRunCount and decay factor variables.  First calls the super.init()
 * to ensure the other class variables are set
 * @throws Exception any number of exceptions can come from setting properties or parsing XML.  They
 * are just passed to somebody else to handle
 */
protected void init() throws Exception{
		super.init();
		strategyRunCount = new Hashtable();
		for (Enumeration e = getSchedulers(); e.hasMoreElements();) {
			//each algorithm starts off with a zero run count
			//have to use Double's here b/c Math.pow is used in decay
			//and that only works for doubles
			strategyRunCount.put(e.nextElement(), new Double(0));
		}
		decayFactor = Double.parseDouble(this.getProperty("SCORE_DECAY_FACTOR"));
}
	/**
    * called by the constructor.  this will set the default properties
    * that the ALL query objects will support. Each subclass should override
    * this method and add their own unique properties, but should also call
    * super.setDefaultProperties() so that the common properties may be set.
    */   
	protected void setDefaultProperties(){
		super.setDefaultProperties();
		/** 
	 		this will determine how rapidly the scores will decay for the algorithms that were not chosen
	 		a value of 1 means the score will never decay, smaller values will cause it to decay faster 
		*/
		addProperty("SCORE_DECAY_FACTOR", "0.95");			
	}
/**
 * After each time the next algorithm is chosen, we wish to record
 * how long algorithms have gone without running.  Right now, this is just
 * based on how many times in a row that it has lost the lottery.  In the future,
 * this could be changed to factor time into the equation. <p>
 *
 * This will also reset the currentCategory's score to 0.  
 */
private void updateRunCounts() {
	
	for(Enumeration e = strategyRunCount.keys(); e.hasMoreElements();){
		Object nextCategory = e.nextElement();
		//if we are not looking at the current category
		if(nextCategory != currentCategory){
			double count = ((Double)scores.get(nextCategory)).doubleValue() + 1;
			strategyRunCount.put(nextCategory, new Double(count));
		} else if(nextCategory == currentCategory){
			//reset the current category 
			strategyRunCount.put(nextCategory, new Double(0));	
		}
	}
	
	
	
	}
	/* (non-Javadoc)
	 * @see edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATQueryObject#getLightWeightQueryObject()
	 */
	public LightWeightQueryObject getLightWeightQueryObject() {
		return new LightWeightQueryObject("GreedyAdapter",1,this.getClass().getName(), true);
	}

}
