package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;

import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.common.dag.DagNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsReference;
/**
 * This scheduler will try to push tuples as far through the tree as possible.
 * It will start at one of the leaf operators and move up in the tree until it gets to the
 * root -or- it reaches an operator that did not produce any results -or- does not
 * have any work to do.  In that case, it will start again with a different root operator.  
 */
public class PushTuplesThroughScheduler extends XATScheduler {
	private XATNode[] nodes;
	private int index;
	private XATStreamOperator lastOperator;
	/** 
	 * this table stores the indicies of the parents that we have visited previously
	 * This way, when we encounter an operator with multiple parents, we can remember
	 * which branch we went to last time so we can go to a different one. 
	 */
	private java.util.Hashtable parentTable;
	private int nodesCount;
/**
 * PushTuplesThroughScheduler constructor comment.
 */
public PushTuplesThroughScheduler() {
	super();
	init();
}
/**
 * Insert the method's description here.
 * Creation date: (2/22/03 1:51:50 PM)
 */
private void init() {
	if(getTree() != null){		
		nodes = getTree().getLeaves();
		nodesCount = getTree().convertToArray().length;
	}		
	index = 0;
	parentTable = new Hashtable();
	
	}
/**
 * make the next scheduling decision.   
 * If the operator is chosen but has workload of 0, then
 * we go back to the bottom of the tree to start the process again. 
 * This reduces the chances of starvation, that is, when the scheduler decides 
 * to run an operator with no tuples in its
 * input buffer.   
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerDecision
 * @throws XATSchedulerException when the leaf operators do not have any work
 * to do.
 */
public XATSchedulerDecision makeNextDecision() throws XATSchedulerException {
    XATSchedulerDecision decision = new XATSchedulerDecision();
    int totalTuples = 0;    
    int operatorsChecked = 0;
    while (true) {
        XATStreamOperator operator = null;
        if (lastOperator != null) {
            /** look at how many tuples this operator outputted last time */
            double lastOutput =
                Double.parseDouble(StatisticsGatherer.getStatisticsGather().getValue("NUMBER_OF_TUPLES_OUTPUTTED", lastOperator, StatisticsReference.LAST));
            /** if it did not output anything, we will start from the bottom again */    
            if (lastOutput > 0) {
                try {
                    int parentIndex = 0;
                    /** deal with issues related to having multiple parents */
                    DagNode dagNode = (DagNode) lastOperator.getNode();
                    if (dagNode.getParentsSize() > 1) {
                        if (parentTable.containsKey(dagNode)) {
                            int oldParentIndex = ((Integer) parentTable.get(dagNode)).intValue();
                            oldParentIndex++;
                            parentIndex = (oldParentIndex >= dagNode.getParentsSize()) ? 0 : oldParentIndex;
                        }
                        parentTable.put(dagNode, new Integer(parentIndex));
                    }
                    DagNode node = null;
                    try {
	                    node = dagNode.getParent(parentIndex);	
                    } catch (IndexOutOfBoundsException e1) {
                    	//there is no parent, so keep node = null
                    }
                    
                    if(node != null){
	                    operator = (XATStreamOperator) node.getOperator();
                    }
                } catch (ClassCastException e) {
	                XATNode node = lastOperator.getNode().getParent();
	                
                    if(node != null){
	                    operator = (XATStreamOperator) node.getOperator();
                    }                    
                }
               
            }
        } else {
            operator = (XATStreamOperator) nodes[index].getOperator();
            index++;
            if (index >= nodes.length) {
                index = 0;
            }            
            operatorsChecked++;
        }
        lastOperator = operator;
        /** 
         * operator is still null is if the last operator's parent was null
         * -or- the last operator did not output anything last time it ran.  
         * then we have to start from the bottom again.  
         */
        if (operator == null) {
            //start again from the bottom
        } 
        else {
            int workload = calculateWorkLoad(operator);
            if (workload > 0 && operator.isActive()) {
                decision.setNextOperator(operator);
                decision.setNumberOfTuples(workload);
                //System.out.println("Selected: " + operator.toString());
                return decision;
            } 
            else {
                if (index >= nodes.length) {
             	   index = 0;
            	}
            }
            
            if (operatorsChecked == nodesCount) {
                throw new XATSchedulerException("Looked at all leaves and there is no work left to do");
            }
        }
    }
}
/**
 * Insert the method's description here.
 * Creation date: (2/22/03 1:46:59 PM)
 * @param newTree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 */
public void setTree(edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree newTree) {
	super.setTree(newTree);
	nodes = getTree().getLeaves();
	nodesCount = getTree().convertToArray().length;
	lastOperator = null;
}
/**
 * Insert the method's description here.
 * Creation date: (2/26/2003 6:22:23 PM)
 * @return java.lang.String
 */
public String toString() {
	String rs = new String();

	String sName = this.getClass().getName();
	//find the last dot.
	int index = 0;
	for(index = sName.length()-1; index >=0; index--)
		if(sName.charAt(index) == '.')
			break;
	rs = sName.substring(index + 1) + " " + hashCode(); // + "  data:";	
    return rs;
}
}
