package edu.wpi.cs.dsrg.xmldb.xat.component.scheduler;
import java.util.ArrayList;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATPropertiesImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer;
/**
 * This is the ABC for all Schedulers and provides 1 method that all Schedulers must implement
 * <tt>makeNextDecision</tt> and one method that Schedulers <i>may</i> implement <tt> calculateWorkLoad</tt>.
 * <tt>makeNextDecision</tt> is responsible for determining which operator should be run next.  The different
 * schedulers can do this anyway they choose and the query plan and statistics gatherer is available for assistance. <p>
 * <tt>calculateWorkLoad</tt> determines how many tuples an operator should dequeue.  The default implementation
 * will look at the WORKLOAD_RATIO property and dequeue NUMBER_OF_TUPLES_IN_INPUT_QUEUES X WORKLOAD_RATIO.  If
 * the scheduler requires dequeuing based on time stamps then this should be overwritten as such. <p>
 * The scheduler(s) to use should be specified in the configuration file.  
 * @author: Brad Pielech
 */
public abstract class XATScheduler extends XATPropertiesImp  {
	private XATTree[] tree;
	private edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsGatherer stats;
/**
 * XATScheduler constructor comment.
 */
public XATScheduler() {
	super();
	stats = StatisticsGatherer.getStatisticsGather();	
}
/**
 * Insert the method's description here.
 * Creation date: (1/28/03 3:20:53 PM)
 * @return int
 * @param operator edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.XATStreamOperator
 */
protected int calculateWorkLoad(XATStreamOperator operator) {
	/** 
         * now figure out how many tuples to dequeue
         * iterate through all input queues and see how many tuples
         * are available for this operator
         * If the totalTuples is 0, then it may be 0 because we havent updated the NUMBER_OF_TUPLES_IN_INPUT_QUEUES
         * property lately (b/c the StreamController doesnt update this).
         * So, try to update the property, check the value, and either break out or proceed as normal
         */
        
        int totalTuples = Integer.parseInt(stats.getValue("NUMBER_OF_TUPLES_IN_INPUT_QUEUES", operator));                
        if(totalTuples == 0){
			stats.updateValue("NUMBER_OF_TUPLES_IN_INPUT_QUEUES", operator);
			totalTuples = Integer.parseInt(stats.getValue("NUMBER_OF_TUPLES_IN_INPUT_QUEUES", operator));
        } 
         
        double ratio = Double.parseDouble(this.getProperty("WORKLOAD_RATIO"));
        int threshold = Integer.parseInt(this.getProperty("WORKLOAD_THRESHOLD"));
        int tuplesToDequeue = (totalTuples * ratio > threshold) ? new Double(totalTuples * ratio).intValue() : (totalTuples);		
 		return tuplesToDequeue; 
}
/**
 * Insert the method's description here.
 * Creation date: (2/22/03 1:46:59 PM)
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 */
public edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree getTree() {
	return tree[0];
}

public edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree getTree(int location) {
	return tree[location];
}

public edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree[] getTrees() {
	return tree;
}
/**
 * This is called by the Execution Controller.  This method decides
 * which operator to run next and how many tuples that operator should dequeue.
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.scheduler.XATSchedulerDecision
 */
public abstract XATSchedulerDecision makeNextDecision() throws XATSchedulerException;
	/**
    * called by the constructor.  this will set the default properties
    * that the ALL query objects will support. Each subclass should override
    * this method and add their own unique properties, but should also call
    * super.setDefaultProperties() so that the common properties may be set.
    */   
	protected void setDefaultProperties(){
		super.setDefaultProperties();
		/** this is the percentage (between 0 and 1) of tuples that the scheduler will dequeue from the input queues*/
		addProperty("WORKLOAD_RATIO", "0.5");
		/**
		* Let A = NUMBER_OF_TUPLES_IN_INPUT_QUEUES X WORKLOAD_RATIO
		* if A > WORKLOAD_THRESHOLD
		*  then the operator will dequeue A tuples.
		* else
		* the operator will dequeue NUMBER_OF_TUPLES_IN_INPUT_QUEUES
		* Setting this value higher or lower will alter the behavior of the execution.
		* Setting this very high could decrease performance, but setting it too low
		* could slow performance because an operator will not be fully utilized.
		* Small changes will not have any affect
		*/
		addProperty("WORKLOAD_THRESHOLD", "10");
		/** 
		 * this is the ratio by which priorities will be decreased 
		 * if the highest priority operator does not have any work to do
		 * This is really only for GreedyPriorityScheduler
		*/
		//addProperty("PRIORITY_DECREASE_FACTOR", ".1");
		/** 
		 * if the ratio between the operator that has run the most times
		 * and an average amount of times is less than this ratio, then that operator's priority
		 * will be decreased.  This should help to prevent one operator from being
		 * run too often
		 */
		//addProperty("STARVATION_RATIO", ".3");
		/** 
		 * How often should the ratio be checked. This is only relevent
		 * in starvation-prone algorithms
		 */
		//addProperty("STARVATION_CHECK_FREQUENCY", "20000");	

		
	}
	
/**
* sets the given property to the new value.  This method is responsible
* for validating the given property.  An exception is thrown if
* the value is not correct.
* Validation should be done in the following manner: 
* First the method should check if the object supports the given property by calling
* <tt>isValidPropertyName(String)<tt>.  If it is not a valid property name, then an exception should
* be thrown. <p>
*
* Once it is known that the object supports the property, the value must be validated.  Each implementor of
* XATStatistics has a set of properties that are common to all of its types and some that are specific to types.  
* For example, all XATOperators (which implement XATStatistics) have certain properties while an individual XATOperator
* has specific properties. The indivual types should only validate their own specific types.  If the caller is trying
* to set a common type, it should be validated by the super class by calling <tt>super.setProperty(name, value)<tt>.  This
* will reduce duplicate code. <p>
*
* When it is determined the name / value pair is correct, the pair is updated using <tt>setNewPropertyValue()<tt> <p>
* <tt>XATStatisticsImp<tt> provides default validation by checking to see if the property's value is
* >=0.  propertyValues that are Strings will just be set.  If the property requires more strict constraints than this, 
* it is up to the subclass that created the property to validate it (as described above).
* @param name the name of the property to set
* @param value the new value of the property
* @throws Exception if the given property cannot be found or
* if the value is not valid for the given property.
*/

public void setProperty(String name, String value) throws Exception {

    if (isValidPropertyName(name)) {
        if (name == "WORKLOAD_RATIO") {
            try {
                double selectivity = Double.parseDouble(value);
                if ((selectivity >= 0) && (selectivity <= 1)) {
                    setNewPropertyValue(name, value);
                } else {
                    System.out.println("Trying to set WORKLOAD_RATIO to: " + selectivity);
                    throw new NumberFormatException(
                        "The property named "
                            + name
                            + " requires a double value between 0 and 1 (inclusive)");
                }
            } catch (NumberFormatException e) {
                throw new NumberFormatException(
                    "The property named "
                        + name
                        + " requires a double value between 0 and 1 (inclusive)");
            }
        } else {
            super.setProperty(name, value);
        }
    } else {

        throw new Exception("The object does not support the property named " + name);
    }
}
/**
 * Insert the method's description here.
 * Creation date: (2/22/03 1:46:59 PM)
 * @param newTree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree
 */
public void setTree(edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree newTree) {
	ArrayList trees = new ArrayList();
	trees.add(newTree);
	tree = (XATTree[])trees.toArray(new XATTree[] {});
	//tree[0] = newTree;
}
public void setTrees(edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree[] newTree) {
	tree = newTree;
}
}
