package edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////
import java.util.Hashtable;

import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.InheritableVisitor;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;

/**
 * This class to do the schema cleanup.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0  
 */

public class SchemaCleanupVisitor extends InheritableVisitor {
    /**
      * The hash table to keep the map between XATNode and its corresponding SchemaCleanUpNode.
      * @since 1.0     
      */
    Hashtable hash;
	/**
     * A visitor used to compute the Produce, Consumed columns for each XATNode.
     * @since 1.0
     */
    PCMVisitor pcmvisitor;
	/**
     * A visitor used to compute the minimun schema.
     * @since 1.0
     */    
    MinSchemaVisitor rscvisitor;

/**
 * The default SchemaCleanupVisitor constructor.
 * @since 1.0
 * @see PCMVisitor
 * @see MinSchemaVisitor
 */
public SchemaCleanupVisitor() {
    hash = new Hashtable();
    pcmvisitor = new PCMVisitor(hash);
    rscvisitor = new MinSchemaVisitor(hash);
}
/**
 * The SchemaCleanupVisitor constructor.
 * @param  hash_init The hash table to keep the map between 
 * XATNode and its corresponding SchemaCleanUpNode.
 * @since 1.0
 * @see PCMVisitor
 * @see MinSchemaVisitor
 */
public SchemaCleanupVisitor(Hashtable hash_init) {
    hash = hash_init;
    pcmvisitor = new PCMVisitor(hash);
    rscvisitor = new MinSchemaVisitor(hash);
}
/**
 * Compute the schema, without cutting the tree.
 * @return void
 * @param tree The given XATTree.
 * @since 1.0
 * @see traverseTopDown(XATNode, InheritableVisitor) traverseTopDown
 */

public void computeSchema(XATTree tree) {
    XATNode root = tree.getRoot();
    traverseTopDown(root, pcmvisitor);
    traverseTopDown(root, rscvisitor);
}
/**
 * Compute the schema, without cutting the tree.
 * @return void
 * @param root The root node of given XATTree.
 * @since 1.0
 * @see traverseTopDown(XATNode, InheritableVisitor) traverseTopDown
 */
public void computeSchema(XATNode root) {
    // keep the parent of this node.
    XATNode parent = root.getParent();
    // make the root of this node null, so that, it will not cofuse the pcm and rsc visitors.
    root.setParent(null);
	
    traverseTopDown(root, pcmvisitor);
    traverseTopDown(root, rscvisitor);

    // restore the original parent.
    root.setParent(parent);
}
public Object dispatch() {
    return null;
}
public Object dispatch(Object o) {
    return o;
}
/**
 * Get the value of the propertity "DEBUG_SchemaCleanUp".
 * @return boolean True: With debug information. False: No debug information.
 * @since 1.0
 */
public static boolean isDebug() {
	return System.getProperty("DEBUG_SchemaCleanUp")!=null;
}
/**
 * Recursively traverse the tree from top down to do the schema cleanup.
 * @return void
 * @param node The current visited XATNode.
 * @param v    The visitor used to do cleanup function.
 * @since 1.0
 */

private void traverseTopDown(XATNode node, InheritableVisitor v) {
	if((!(hash.containsKey(node))&&(v instanceof PCMVisitor))|| (v instanceof MinSchemaVisitor)){
		v.visit(node);
		XATNode[] children = node.getChildren();
		for (int i = 0; i < children.length; i++) {
			traverseTopDown(children[i], v);
		}
	}
}
/**
 * Visit a XATTree to clean up schema.
 * @return Object
 * @param tree The XATTree current cleaned.
 * @since 1.0
 */

public Object visit(XATTree tree) {
    XATNode root = tree.getRoot();
    traverseTopDown(root, pcmvisitor);
    traverseTopDown(root, rscvisitor);

    return null;
}
/**
 * Visit a XATNode to clean up schema.
 * @return Object
 * @param root The XATNode current visited.
 * @since 1.0
 */
public Object visit(XATNode root) {
    // keep the parent of this node.
    XATNode parent = root.getParent();
    // make the root of this node null, so that, it will not cofuse the pcm and rsc visitors.
    root.setParent(null);
    traverseTopDown(root, pcmvisitor);
    traverseTopDown(root, rscvisitor);
    // restore the original parent.
    root.setParent(parent);

    return null;
}
}
