package edu.wpi.cs.dsrg.xmldb.xat.component.sqlgeneration;

import edu.wpi.cs.dsrg.xmldb.xat.common.exception.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.utils.*;
/**
 * <pre>
 * IQueryGenerator is the primary class in Query Generation
 * It is responsible for travering the XAT and finding operators to evaluate
 * </pre>
 * @author: Brad Pielech
 */
public class QueryGenerator {
	/**the XAT to generate a query from */
    protected XATTree queryTree;
   	/**full generation*/
    public final static int FULL = 1;
    /**minimal generation */
    public final static int MINIMAL = 0;
    /**binding BindingTable*/
    protected static BindingTable variableTable; 
    /**instance of the OperatorEvaluation */
    protected OperatorEvaluator opEval;
    /**the incremental statement to use */
    protected IncrementalQuery qStmt;
    /**the incremental statement to use */
    protected static boolean debug = false;
    /**the type of query to generate */
	protected int queryType;
    /**
     * Start the generator.
     */
    public QueryGenerator(BindingTable varTable) {
        variableTable = varTable;
        debug = false;
              
    }
    /**
     * Start the generator.
     */
    public QueryGenerator(BindingTable varTable, boolean deb) {
        variableTable = varTable;
        debug = deb;
              
    }
/**
 * adds a new SQL operator as a child of root parameter
 * @param root the parent of the SQLOperator that will be created
 */
protected void addSQLOperator(XATOperator root) {

    try {
        String sqlText = qStmt.getStatement();
        XATNode newNode = new XATNodeImp();
        if(queryType == QueryTypes.SQL){
	        SQLStmt s1 = new SQLStmtImp(newNode, sqlText);
            newNode.setOperator(s1);
        }else if(queryType == QueryTypes.XQUERY){
			XQueryStmt s1 = new XQueryStmtImp(newNode, sqlText);
            newNode.setOperator(s1);
        }

        root.getNode().addChild(newNode);
        newNode.setParent(root.getNode());
        opEval.setSourceVisited(false);
    } catch (SQLNoStatementExists e1) {
	    //the SQL statement was blank so dont modify the tree
		return ;
    }

}
    /**
     * find an operator, using post order breadth first search
     * call operator evaluator, then if take the resulting SQLFragment
     * and add it to the SQLIncrementalStatement if the operator was evaluated
     * 
     * @param treeToGenerateSQL XATTree
     * @param heuristic int
     * @param queryTypeIn the type of query to generate
     */
    public void generateQuery(XATTree treeToGenerateSQL, int heuristic, int queryTypeIn) throws UnknownQueryType{
        /****
        ***  I dont do anything with the heuristic parameter besides pass it around
        */
        queryTree = treeToGenerateSQL;
        queryType = queryTypeIn;
        //initialize the class variables depending on queryType 
        initialize(heuristic, queryType);
        
        
        XATOperator firstRoot = queryTree.getRoot().getOperator();
        //traverse tree does all the work
        traverseTree(firstRoot);
		return ;
    }
    /**
     * Set the variable table
     * 
     * @param newVariableTable BindingTable
     */
    public BindingTable getVariableTable() {
        return variableTable;
    }
 /*
 * Initialize qStmt to the appropriate type of Query according to the queryType parameter
 * @param queryType int
 */
public void initialize(int heuristic, int queryType) throws UnknownQueryType{
	
	       switch (queryType) {
        	case QueryTypes.SQL: {
        		qStmt = new IncrementalSQL(getVariableTable());
        		opEval = new SQLOperatorEvaluator(heuristic, variableTable);
        		break;
        	}
        	case QueryTypes.XQUERY: {
				qStmt = new IncrementalXQuery();
				opEval = new XQueryOperatorEvaluator(heuristic, variableTable);
				break;
        	}
        	default: {
        		throw new UnknownQueryType("Please enter a valid query type");
        	}
        }
	}
/**
* <pre>
* processOperator will call evaluateOperator on the param operator
* then take the QueryResult object that was returned and will add the
* sql fragment to the incremental SQL statement if the operator was evaluated
* the toProcess operator will then be deleted from the tree
* </pre>
* @return boolean true if the operator was evaluated, false otherwise
*/
protected boolean processOperator(XATOperator toProcess) {
    QueryResult result = opEval.evaluateOperator(toProcess);
    if (result.isEvaluated()) {
        QueryFragment[] resultFrags = result.getFragment();
        for (int i = 0; i < resultFrags.length; i++) {
            qStmt.addClause(resultFrags[i]);
        }

        try {
            queryTree.deleteNode(toProcess.getNode());

        } catch (XATTreeNodeHasNoParentException e) {
            e.printStackTrace();
            return false;
        }
    }
    return result.iscommitQuery();
}
    /**
     * Set the variable table
     * 
     * @param newVariableTable BindingTable
     */
    public void setVariableTable(BindingTable newVariableTable) {
        variableTable = newVariableTable;
    }
/**
* <pre>
*  Iterates over a tree starting at the root parameter
*  traverses to the right child, then left child, then decides whether to evalutate this operator
*  Once an operator's children have been examined, the operator is then processed and the
*  result is added to sqlsmt. An operator is deleted after it is processed
*  If a Tagger or Combine operator is found, the incremental query is added to the 
* tree in addSQLOperator
* </pre>
*  sqlStmt is reset in case there are other operators
*  @param root the root of the tree to traverse.   
*/
protected void traverseTree(XATOperator root) {

    try {
        //go the the right first if possible, if no right child, then go to the exception handler
        XATOperator rightChild = (XATOperator) root.getNode().getChild(1).getOperator();
        traverseTree(rightChild);
        //try the left child for this operator
        XATOperator leftChild = (XATOperator) root.getNode().getChild(0).getOperator();
        traverseTree(leftChild);

    } catch (XATNodeDoesNotExistException e1) {
        /*
        * XATNodeDoesNotExistException is thrown when there 
        * is no child for an operator.  So, this exception this means there is no right
        * child to traverse at this part of the tree
        * so we just try the left, if there is no left child, then process the operator
        */
        try {
            XATOperator leftChild = (XATOperator) root.getNode().getChild(0).getOperator();
            traverseTree(leftChild);
        } catch (XATNodeDoesNotExistException e2) {
        }
    }
    /*
    * this is the default case, after the left and right sides have been processed
    * check for a Tagger or an Combine, if root is one, then add the SQLOperator to
    * the tree. The sqlStmt is created again in case there are more statements to add
    */
    //    if ((root instanceof Tagger) || (root instanceof Combine)) {
    //the sql operator will only be added if there is a from clause in the SQL statement
    //        addSQLOperator(root);

    //process the operator, if the result says to commit the query, then add the SQL to the tree and initialize
    if (processOperator(root)) {
	    if(debug)System.out.println("QueryGenerator.traverseTree():: processOperator(" + root + ") sql created!");
	    addSQLOperator(root);
	    if(debug)System.out.println("QueryGenerator.traverseTree():: qStmt is:\n" + qStmt);
        qStmt.initialize();

    }else{
	    if(debug)System.out.println("QueryGenerator.traverseTree():: processOperator(" + root + ") no sql yet!");
    }

    return;
}
}
