package edu.wpi.cs.dsrg.xmldb.xat.component.xatdaggeneration;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import xacute.quilt.*;
import java.util.Vector;
import java.util.Hashtable;

import edu.wpi.cs.dsrg.utils.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dompattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.expression.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.operatorfactory.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dag.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.xatrewrite.common.*;

/**
 * This class visits the objects of the Kweelt parser and puts
 * the query in the correct algebra node.
 * <p>
 * To understand this code, one should know how the Kweelt parser
 * works and what our algebra is.  It would also be helpful to 
 * know what a visitor pattern is because that is what we use to
 * traverse through the query.
 * <p>
 * There are different types of the return value from the visit() function.
 * For visit(QuiltQuery) returns a Vector of XATTrees.
 * For visit(*Op*) returns a Vector of Expression + XATDag.
 * For visit(LocationStep) returns a Vector of Step + XATDag.
 * For visit(Value*) returns a Primitive Type (*).
 * For visit(*) returns a XATDag.
 *
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 * @see xacute.quilt
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern
 */

public class DagParserTraverse extends InheritableVisitor {
    /**
    * The number of nodes in the tree, used for output column name assignment
    *
    * @since 1.0
    */
    private long numNodes;
    /**
    * Keep track of numNodes to have it global unique.
    *
    * @since 1.0
    */
    private SequenceNumber seq;
    /**
    * Current level in the parse tree. It is used for the indent method
    *
    * @since 1.0
    */
    private int level;
    /**
    * The hashtable contains all the XATs for funDefs and also for the Body.
    *
    * @since 1.0
    */
    private Hashtable xatTrees;
    /**
    * Variable for the context of the update statements.
    *
    * @since 1.0
    */
    private String updateContext;
    /**
    * Varaible to keep the context of visit(CurrentNode) method.
    * It will only be set in the visit(Nav) method,
    * It will only be used in the visit(CurrentNode) to generate a VarStep.
    *
    * @since 1.0
    */
    private String currentNodeColumnName;

    /**
    * Current context for the context-insensitive xpath.
    *
    * @since 1.0
    */
    private String xPathContext;

    /**
    * Used to create operators.
    *
    * @since 1.0
    */
    protected OperatorBuilder builder;

    /**
    * Internal structure to return by function findEntryPointAndCreateHeadTree().
    *
    * @since 1.0
    */
    class EntryTree {
        // the head tree.
        XATDag tree;
        // the entry point.
        NavigationStep entry;

        public NavigationStep getEntry() {
            return entry;
        }

        public XATDag getTree() {
            return tree;
        }

        public EntryTree(NavigationStep entry, XATDag tree) {
            this.tree = tree;
            this.entry = entry;
        }
    }
/**
 * Class Default Constructor
 *
 * @since 1.0
 */
public DagParserTraverse() {
    this(false);
}
/**
 * Class constructor specifying debug mode
 *
 * @since 1.0 
 */
public DagParserTraverse(boolean debugMain) {
    level = 0;
    // assume each tree has no more than 1000 nodes.
    seq = new SequenceNumber("counter", 1000);
    // make it an unique number.
    numNodes = seq.get();
    setDebug(debugMain);
}
/**
 * Put Expost operator on top of NON-FLWR XQuery.
 * 
 * @param tree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATDag
 * @since 1.0
 */
void addExpose(XATDag tree) {
    // if the main query is not FLWU, we put the expose operator,
    // otherwise, for FLWU, we don't need that.
    
    // add the expose operator.
    DagNode root = (DagNode)tree.getRoot();
    // make a new expose operator use the output column of the root.
    DagNode expNode = new DagNodeImp();
    Expose exp = builder.newExpose(expNode);
    exp.setColumnName(root.getOperator().getOutputColumnName());
    exp.setOutputColumnName(root.getOperator().getOutputColumnName());
    // make the expose as the root and make the root as the child of the exp.
    tree.setRoot(expNode);
    tree.addNode(root, expNode);
}
/**
 * Generate GroupBy operator using specified context
 * <p>
 * This method will put the specified Operator into the inner tree of
 * the groupby operator.
 *
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.DagNode
 * @param op operator to be specified.
 * @param context collumn names to be grouped by on.
 * @since 1.0
 */
XATOperator createGBOp(XATOperator op, String context) {

    GroupBy gbOp = builder.newGroupBy(new DagNodeImp());
    gbOp.setOutputColumnName(op.getOutputColumnName());

    // set the groupBy's input columns.
	gbOp.setColumnNames(new String[]{context});

	// create the subtree of the groupby operator.
    XATDag subTree = new XATDagImp();

    // disconnect the op's node from it's children.
    DagNode opNode = (DagNode)op.getNode().clone();
    opNode.setChildren(new DagNodeImp[0]);

    // add the op's node to the subtree.
    subTree.setRoot(opNode);

    gbOp.setXATTree(subTree);

    return gbOp;
}
/**
* To create a select operator out of a Kweelt Nav.
* It will be used for:
* FilterExpression of LocationStep, eg /book[author]
*
* @return XATDag for that expression.
* @param predicate take a Kweelt Operator object.
* @param outputColName for the output columnname of 
*         dummy rename operator on the top of the select operator.
* @since 1.0
*/
XATDag createSelect(Nav predicate, String outputColName) {
    // create the subtree.
    XATDag subtree = new XATDagImp();

    //Create select condition like $col != ""
    BinCOMPExpression compExp = new BinCOMPExpressionImp();
    //set the type to whatever type of comparison it is
    compExp.setType(xacute.common.Constants.NEQ);
    // set the left operand.
    compExp.setLeft(handleExpression(visit(predicate), subtree));
    String emptyString = new XATDataValueImp().toString();
    compExp.setRight(handleExpression(emptyString, subtree));

    // create the select operator.
    DagNode selectNode = new DagNodeImp();
    Select selectOp = builder.newSelect(selectNode, compExp);
    selectOp.setOutputColumnName(outputColName);
    // set the select node as the root of the subtree.
    subtree.insertRoot(selectNode);

    // insert a dummy rename node to keep track of the output column name.
    DagNode nameNode = new DagNodeImp();
    NameColumn nameOp = builder.newNameColumn(nameNode);
    nameOp.setOriginalColumnName(outputColName);
    nameOp.setNewColumnName(outputColName);
    nameOp.setOutputColumnName(outputColName);
    subtree.insertRoot(nameNode);
    return subtree;
}
/**
* To create a select operator out of a Kweelt Operator.
* It will be used for:
* FilterExpression of LocationStep and Where Clause
*
* @return a XATDag for that expression.
* @param predicate take a Kweelt Operator object.
* @param outputColName for the output columnname of dummy
*        rename operator on the top of the select operator.
* @since 1.0
*/
XATDag createSelect(Operator predicate, String outputColName) {
    // create the subtree.
    XATDag subtree = new XATDagImp();
    // get the expression from the visited result.
    Vector temp = (Vector) visit(predicate);
    Expression exp = (Expression) temp.elementAt(0);
    // create the select operator.
    DagNode selectNode = new DagNodeImp();
    Select selectOp = builder.newSelect(selectNode, exp);
    selectOp.setOutputColumnName(outputColName);
    // set the select node as the root of the subtree.
    subtree.setRoot(selectNode);
    // insert a dummy rename node to keep track of the output column name.
    DagNode nameNode = new DagNodeImp();
    NameColumn nameOp = builder.newNameColumn(nameNode);
    nameOp.setOriginalColumnName(outputColName);
    nameOp.setNewColumnName(outputColName);
    nameOp.setOutputColumnName(outputColName);
    subtree.insertRoot(nameNode);
    // if there exists the sub-tree with the expression, put it under.
    if (temp.size() > 1) {
	    XATDag predicateTree = (XATDag) temp.elementAt(1);
        subtree = mergeTreeWithSources(subtree,predicateTree);
    }
    return subtree;
}
/**
* To create a select operator out of a Kweelt expression.
*
* @return a XATDag for that expression.
* @param kweeltExp Kweelt expression object.
* @param columnName for the output of dummy
*        rename operator on the top of the select operator.
* @since 1.0
*/
XATDag createSelect(QuiltExpression kweeltExp, String columnName) {
	// if that's an operator, use different createSelect function.
	if(kweeltExp instanceof Operator){
		return createSelect((Operator)kweeltExp, columnName);
	}
	// if that's an Nav, use different createSelect function.
	if(kweeltExp instanceof Nav){
		return createSelect((Nav)kweeltExp, columnName);
	}
    // create the subtree.
    XATDag subtree = new XATDagImp();
    // Set the context for the context insensitive XPath
    String oldContext = this.xPathContext; // remember the old context.
    this.xPathContext = columnName;
    // get XAT tree from the visited result.
    XATDag tempTree = (XATDag) visit(kweeltExp);
    this.xPathContext = oldContext; // restore the old context.
    // create the expression where to check whether the last column is true or false.
    BinCOMPExpression exp = new BinCOMPExpressionImp(new TerminalExpressionImp(new ColumnNameImp(tempTree.getRoot().getOperator().getOutputColumnName())), new TerminalExpressionImp(true));
    exp.setType(BinCOMPExpression.EQ);
    // create the select operator.
    DagNode selectNode = new DagNodeImp();
    Select selectOp = builder.newSelect(selectNode, exp);
    selectOp.setOutputColumnName(columnName);
    // set the select node as the root of the subtree.
    subtree.setRoot(selectNode);
     // insert a dummy rename node to keep track of the output column name.
    DagNode nameNode = new DagNodeImp();
    NameColumn nameOp = builder.newNameColumn(nameNode);
    nameOp.setOriginalColumnName(columnName);
    nameOp.setNewColumnName(columnName);
    nameOp.setOutputColumnName(columnName);    
    subtree.insertRoot(nameNode);
   
    subtree = mergeTreeWithSources(subtree,tempTree);
    
    return subtree;
}
public Object dispatch() {
    if (isDebug())
        System.out.println(indent() + "NULL");
    return null;
}
public Object dispatch(Object obj) {
    System.out.println(
        indent() + "Dispatch!! " + obj.getClass().toString() + " is not handled yet!");
    return null;
}
	public Object dispatch(Object ob1, Object ob2){
		System.err.println("WARNING: " + ob1.getClass().getName() + ":(" + ob1.toString() +  ") with such parameter " +  ob2.getClass().getName() + ":(" + ob2.toString() + ") doesn't support!");
		return null;
	}
/**
 * Convert an Expression Vector into an XAT Tree.
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATDag
 * @param expVector Vector to be converted.
 * @since 1.0
 */
public XATDag expressionVectorToXAT(Vector expVector) {

	XATDag result = null;
	
	// If that's an arithmatic expression, we use function() operator to capture it.
	Expression exp = (Expression)((Vector)expVector).elementAt(0);
	XATDag expTree = (XATDag)((Vector)expVector).elementAt(1);
	// the parameter of the function will be this arithmatic expression.
	DagNode funNode = new DagNodeImp();
	FunArith funOp = new FunArith(funNode);
	// set the parameter of the funOp.
	funOp.setParams(new Object[]{exp});

	// set the output column name for this funOp.
	this.numNodes++;
	funOp.setOutputColumnName("$col" + numNodes);

	// make a new returnClause XATtree with this FunNode as the root node.
	result = new XATDagImp();
	result.setRoot(funNode);

	// link the rest of the XATtree for that expression underneath the returnTree.
	result = mergeTreeWithSources(result,expTree);
	
	return result;
}
/**
 * Find an entry point and construct the header tree for a given Head object.
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.VarStep
 * @param head java.lang.Object
 * @since 1.0
 */
EntryTree findEntryPointAndCreateHeaderTree(Object head) {

	NavigationStep entry = null;
	XATDag result = new XATDagImp();
	
    if (head instanceof XATDagImp) { 
	    // if that's a XATDag
        // it's going to contain the source operator, and should be put at the bottom.
        result = (XATDag) head;
        // it's going to contain the source operator, and should be put at the bottom.
        entry = new VarStepImp(result.getRoot().getOperator().getOutputColumnName());
    } else if (head instanceof VarStepImp) {
        //if that's a VarStep, then, that's the entry step.
        entry = (VarStep) head;
    } else if (head instanceof Vector) { 
	    //if that's an arithmatic expression.
        result = this.expressionVectorToXAT((Vector) head);
        entry = new VarStepImp(result.getRoot().getOperator().getOutputColumnName());
    }
    
    return new EntryTree(entry, result);
}
/**
 * Returns the xat. It is used by other components / classes to retrieve the tree.
 *
 * @return XAT Tree
 * @since 1.0
 */
public XATDag getTree() {
    return (XATDag) this.xatTrees.get("main");
}
/**
 * Returned the XAT for the user defined functions.
 *
 * @return XAT Tree
 * @param functionName function to get the XAT tree.
 * @since 1.0
 */
public XATDag getTree(String functionName) {
    if (functionName == null)
        return new XATDagImp();
    return (XATDag) this.xatTrees.get(functionName);
}
/**
* It check the result from a visit(), and depends on its type.
* It will merge the proper trees together and also return the expression.
*
* @return Expression
* @param result java.Object
* @param subtree XAT Tree to be visited.
* @since 1.0
*/
private Expression handleExpression(Object result, XATDag subtree) {
    Expression exp = null;
    if (result instanceof XATDag) { // it's a xattree.
        XATDag temp = (XATDag) result;
        // need to create a TerminalExpression.
        ColumnName col =
            new ColumnNameImp(temp.getRoot().getOperator().getOutputColumnName());
        exp = new TerminalExpressionImp(col);
        // keep the subtree.
        subtree = mergeTreeWithSources(subtree, temp);
    } else if (result instanceof Vector) { // it's another expression
        // get the expression part.
        exp = (Expression) ((Vector) result).elementAt(0);
        // keep the subtree.
        subtree =
            mergeTreeWithSources(subtree, (XATDag) ((Vector) result).elementAt(1));
    } else if (result instanceof Double) { // terminal expression double.
        exp = new TerminalExpressionImp((Double) result);
    } else if (result instanceof String) { // terminal expression string.
        exp = new TerminalExpressionImp((String) result);
    }
    return exp;
}
/**
 * Generate the XAT Tree for Where Clause
 *
 * @return XAT Tree
 * @param flwr XQurey's FLWR Expression
 * @param returnClause XAT Tree already generated for Return Clause
 * @param returnCol Returned column name
 * @since 1.0
 */
public XATDag handleWhereClause(
    FLWRExpression flwr,
    XATDag returnClause,
    String returnCol) {
    XATDag whereClause = null;
    try {
        // visit the expressions in where clause and create the subtree.
        whereClause = createSelect(flwr.getWhere(), returnCol);
    } catch (ClassCastException e) {
        System.out.println("Problem with Operator (kweelt) handling in FLWR");
        e.printStackTrace();
    }
    // Added to the top of the return clause.
    // check if there is a source operator in the where clause.
	whereClause = mergeTreeWithSources(whereClause, returnClause);

	return whereClause;

}
/**
 * Check if a XATDag has a for node.
 *
 * @return boolean
 * @param tree The XAT Tree
 * @since 1.0
 */
boolean hasFOR(XATDag tree) {
    return hasFOR((DagNode)tree.getRoot());
}
/**
 * Find the for operator in XAT Tree recurrsivly(DFS)
 *
 * @return boolean
 * @param root Root of the XAT Tree
 * @since 1.0
 */
boolean hasFOR(DagNode root) {
    if (root == null)
        return false;
    if (root.getOperator() instanceof FOR)
        return true;
    DagNode[] children = (DagNode[])root.getChildren();
    for (int i = 0; i < children.length; i++) {
        if (hasFOR(children[i]))
            return true;
    }
    return false;
}
/**
* Check if a XATDag has a source node.
*
* @return boolean
* @param tree The XAT Tree
* @since 1.0
*/
boolean hasSource(XATDag tree) {
    if (tree == null)
        return false;

    // just check if any leaf is a source node.
    DagNode[] leaves = (DagNode[])tree.getLeaves();
    for (int i = 0; i < leaves.length; i++) {
        if (leaves[i].getOperator() instanceof Source) {
            return true;
        }
    }
    return false;
}
/**
 * Print out the indent space.
 *
 * @since 1.0
 */
String indent() {
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < level - 1; i++)
        sb.append("  ");
    return sb.toString();

}
/**
 * Get the value of the propertity 'DEBUG_ParserTraverse'.
 * @return boolean True: With debug information. False: No debug information.
 *
 * @since 1.0
 */
public boolean isDebug() {
	return System.getProperty("DEBUG_ParserTraverse")!=null;
}
/**
 * Merge two trees together and aware of source operators.
 * <p>
 * If there is a source operator, then the lower part are copied many
 * times with product operator.
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATDag
 * @param top edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATDag
 * @param bottom edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATDag
 * @since 1.0
 */
protected XATDag mergeTreeWithSources(XATDag top, XATDag bottom) {
	// check the validation of the two parameters.
	if(top == null){
		return bottom;
	}else if(bottom == null){
		return top;
	}else if(bottom.getRoot() == null){
		return top;
	}

	// merge the two trees.
	if (!hasSource(top))
        top.addXATTree(bottom);
    else { //it has source operators.
        // need to create product for each source and the return clause.
        DagNode[] leaves = (DagNode[])top.getLeaves();
        for (int i = 0; i < leaves.length; i++) {
			// clone the bottom clause.
   			XATDag clone = (XATDag)bottom.clone();
            if (leaves[i].getOperator() instanceof Source) {

	            if(leaves[i].getParent().getOperator() instanceof CartesianProduct){
		            // if the parent already has a cartesian product,
		            // we don't need to add another one again.
		            continue;
	            }
                // make a new product operator.
                DagNode productNode = new DagNodeImp();
                CartesianProduct productOp = new CartesianProductImp(productNode);

                // replace leave node with product node.
                top.replaceNode(leaves[i], productNode);

                // add the leave as child of productNode
                // add the clone as another child of the productNode
                top.addNode(leaves[i], productNode);
                top.addNode(clone.getRoot(), productNode);

            }else{
	            // put the clone as the child of the leave.
	            top.addNode(clone.getRoot(), leaves[i]);
            }
        }
    }
    return top;
}
/**
 * FOR each predicate tree, we need to traverse the tree and add a groupby 
 * operator for each function which is context sensitive, and also
 * the groupby is on the output column before the navigate operator, which is
 * the entry point of the navigate operator.
 * 
 * @param predicateTree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATDag
 * @param context java.lang.String
 * @since 1.0
 */
private void preprocessPredicateTree(XATDag predicateTree, String context) {
	// traverse the tree.
	// if the tree contains a function operator.
	// if the function operator is a context sensitive operator.
	// make a groupby operator.
	// The algorithm is like decorrelation.
	if(predicateTree == null)
		return;
	if(context == null){
		System.err.println("WARNING: ParserTraverse::preprocessPredicateTree() In XPath Generation, for the context of the predicates. We see empty Context!");
		return;
	}
	preprocessPredicateTree((DagNode)predicateTree.getRoot(), context);
}
/**
 * For each operator if it is a context sensitive function,
 * create a groupby operator based on the context.
 * 
 * @param predicateTree edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATDag
 * @param context java.lang.String
 * @since 1.0
 */
private void preprocessPredicateTree(DagNode root, String context) {
    if (root == null)
        return;
    if (context == null) {
        System.err.println(
            "WARNING: ParserTraverse::preprocessPredicateTree() In XPath Generation, for the context of the predicates. We see empty Context!");
        return;
    }
    // test the operator of this xatnode.
    XATOperator op = root.getOperator();
    // HERE: We assume all the functions are context sensitive.
    boolean contextSensitive = false;
    if (op instanceof Function) {
        contextSensitive = true;
    } else if (op instanceof GroupBy) {
        contextSensitive = true;
    } else if (op instanceof Combine) {
        contextSensitive = true;
    }
    if (contextSensitive) {
        // make a groupby operator.
        XATOperator gbOp = createGBOp(op, context);
        gbOp.setNode(root);
        root.setOperator(gbOp);
    }
    // handle the children of this xatnode.
    DagNode[] children = (DagNode[])root.getChildren();
    if (children == null)
        return;
    for (int iChild = 0; iChild < children.length; iChild++) {
        preprocessPredicateTree(children[iChild], context);
    }
}
/**
 * Create XAT Tree for Attibutes.
 * 
 * @return java.util.Vector 1) A Vector of AttributeNodes with their values or columnnames. 
                            2) A XATDag.
 * @param atts xacute.quilt.AttributeExpression[]
 * @since 1.0
 */
Vector processAtts(AttributeExpression[] atts) {
	// prepare the result vector.
	Vector result = new Vector();
	// retrive all the quilt expression and put those into an expression.
	QuiltExpression[] content = new QuiltExpression[atts.length];
	for(int i = 0; i < atts.length; i++){
		content[i]= atts[i].getE();
	}
	// reuse the process content to get the results back.
	Vector contentResult = this.processContent(content);

	// get the list of column names or values.
	Vector cols = (Vector)contentResult.elementAt(0);

	// get the XATDag back.
	XATDag tree = (XATDag)contentResult.elementAt(1);

	// prepare the AttributeNode List.
	Vector attNodeList = new Vector();

	// for each atts, create the AttributeNode, and then, add the columnnameNode or textNode underneath each attributeNode.
	// we assume the number of cols are the same as the number of atts.
	for(int i = 0; i < atts.length; i++){
		// create the AttributeNode.
		AttributeNode attNode = new AttributeNodeImp(atts[i].getName());
		// create the ColumnNameNode or TestNode and added to the AttributeNode.
	    if(cols.elementAt(i) instanceof ColumnNameImp){
		    ColumnNameNodeImp tempNode = new ColumnNameNodeImp(((ColumnName)cols.elementAt(i)).getColumnName());
			attNode.addChild(tempNode);
			tempNode.setParent(attNode);
	    }else{
		    TextNodeImp tempNode = new TextNodeImp(cols.elementAt(i).toString());
		    attNode.addChild(tempNode);
		    tempNode.setParent(attNode);
		}
		// add the AttributeNode to the AttributeNode List (vector).
		attNodeList.addElement(attNode);
	}

	// add the AttributeNode List to the result.
	result.addElement(attNodeList);

	// add the XATDag to the result.
	result.addElement(tree);

	// return the result.
	return result;
}
/**
 * Create XAT Tree for different types of contents.
 * 
 * @return java.util.Vector: 1) Vector of Column names in the contents and also the constants. 
                             2) XATDag.
 * @param contents xacute.quilt.QuiltExpression[]
 * @since 1.0
 */
Vector processContent(QuiltExpression[] contents) {
	/*
	Here is the algorithm.
	The content could be different types, a terminal, e.g., string or double, or an XATDag.
	No matter whether there is a source node in the subtree or not, the subtree should put
	under the merge operator.
	*/
	// prepare the result.
	Vector result = new Vector();
	Vector Cols = new Vector();

	// create the Merge node.
	DagNode mergeNode = new DagNodeImp();
	Merge mergeOp = builder.newMerge(mergeNode);

	// create the Merge Tree.
	XATDag mergeXAT = new XATDagImp();
	mergeXAT.setRoot(mergeNode);

	// create the upper Tree.
	XATDag xat= new XATDagImp();

	
	// iterate through the contents.
	for(int i = 0; i < contents.length; i++){
		Object temp = visit(contents[i]);
		if(temp instanceof XATDagImp){
			String outputCol = ((XATDag)temp).getRoot().getOperator().getOutputColumnName();
			Cols.addElement(new ColumnNameImp(outputCol));
			// put it under the merge operator.
			if(((XATDag)temp).getRoot() != null){
				mergeNode.addChild(((XATDag)temp).getRoot());
				((XATDag)temp).getRoot().setParent(mergeNode);
			}
			
		}else if(temp instanceof String){
			Cols.addElement(temp);
		}else if(temp instanceof Double){
			Cols.addElement(temp);
		}
	}

	if(mergeNode.getChildren().length == 1){
		// if there only one child of the merge operator, delete it.
		try{
			mergeXAT.setRoot(mergeNode.getChild(0));
			xat = mergeTreeWithSources(xat,mergeXAT);
		}catch(XATNodeDoesNotExistException e){
			System.err.println("Impossible to have this Error in normal cases.");
		}
	}else if(mergeNode.getChildren().length > 1){
		// if more than one, the keep the merge node.
		xat = mergeTreeWithSources(xat,mergeXAT);
	}// else, no child, so do nothing to the upper.

	result.addElement(Cols);
	result.addElement(xat);
		
	return result;
}
/**
 * Set to show debugging message.
 * 
 * @param debug boolean
 * @since 1.0
 */
public void setDebug(boolean debug) {
    if (debug)
        System.setProperty("DEBUG_ParserTraverse", "OK");
}
/**
 * Convert a intermedia result into an XATDag.
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATDag
 * @param obj java.lang.Object
 * @since 1.0
 */
XATDag toXATTree(Object visitResult) {
	XATDag result = null;
	if(visitResult instanceof XATDag)
	 	result = (XATDag)visitResult;
	else if (visitResult instanceof VarStepImp){
		// make a tree with only one node, which is a dummy rename operator.
		result = new XATDagImp();
		DagNode root = new DagNodeImp();
		NameColumn ren = builder.newNameColumn(root);
		ren.setOriginalColumnName(((VarStepImp)visitResult).getStepValue());
		ren.setNewColumnName(((VarStepImp)visitResult).getStepValue());
		ren.setOutputColumnName(((VarStepImp)visitResult).getStepValue());
		result.setRoot(root);
	}else if (visitResult instanceof Vector){
		result = this.expressionVectorToXAT((Vector)visitResult);
	}else if (visitResult instanceof String){
		// if it is a string, make it a Arith Function with only a string.
		result = new XATDagImp();
		DagNode root = new DagNodeImp();
		FunArith fun = builder.newFunArith(root);
		fun.setParams(new Object[]{new TerminalExpressionImp((String)visitResult)});
		this.numNodes++;
		fun.setOutputColumnName("$col" + this.numNodes);

		result.setRoot(root);
	}else if (visitResult instanceof ValueString){
		// if it is a value string, make it a Arith Function with only a string.
		result = new XATDagImp();
		DagNode root = new DagNodeImp();
		FunArith fun = builder.newFunArith(root);
		fun.setParams(new Object[]{new TerminalExpressionImp(((ValueString)visitResult).getValue())});
		this.numNodes++;
		fun.setOutputColumnName("$col" + this.numNodes);

		result.setRoot(root);
	}else{
		System.err.println("ERROR: toXATTree has unhandled Visit Result.");
	}
	return result;
}
/**
 * Create XAT Tree for Binary AND Expression.
 * It will return a Vector of 1) expression and 2) subtree of required operators.
 * 
 * @return Vector of BinANDExpression and subXATTree.
 * @param andOp xacute.quilt.BinOpAND
 * @since 1.0
 * @see BinOpAND
 */
public Object visit(BinOpAND andOp) {
    level++;
    if (isDebug())
        System.out.println(indent() + "BinOpAND");

    //prepare the Subtree.
    XATDag subtree = new XATDagImp();

    //initialize a new BinAndExpression
    BinANDExpression andExp = new BinANDExpressionImp();

    // keep the root of this AND expression.
    BinANDExpression rootExp = andExp;

    // handle the result, split the expression and also the subtree parts.
    andExp.setLeft(handleExpression(visit(andOp.getE()), subtree));

    // There could be multiple operands. Set the operands with new AND expressions.
    QuiltExpression[] operands = andOp.getOperands();
    for (int i = 0; i < operands.length - 1; i++) {
        BinANDExpression innerExp = new BinANDExpressionImp();
        andExp.setRight(innerExp);
        innerExp.setLeft(handleExpression(visit(operands[i]), subtree));
        andExp = innerExp;
    }

    // set the last operand as the right one.
    // Note: andExp might updated in the above for loop.
    andExp.setRight(
        handleExpression(visit(operands[operands.length - 1]), subtree));

    // prepare the result vector.
    Vector result = new Vector();
    result.addElement(rootExp);
    result.addElement(subtree);

    level--;
    return result;
}
/**
 * Create XAT Tree for Binary Arithmatic Expression.
 * This method is visited an arithmatic expression encountered in a query
 * 
 * @return Vector of BinArithExpression + XATTRee
 * @param arith xacute.quilt.BinOpArith
 * @since 1.0
 * @see BinOpArith
 */
public Object visit(BinOpArith arith) {
    level++;
    if (isDebug())
        System.out.println(indent() + "BinOpArith");

    BinArithExpression arithExp = new BinArithExpressionImp();
    BinArithExpression arithRoot = arithExp;

    // create the subtree.
    XATDag subtree = new XATDagImp();

    //set the type to whatever type of arithmatic it is
    arithExp.setType(arith.getOp());

    // set the left operand.
    arithExp.setLeft(handleExpression(visit(arith.getE()), subtree));

    // BinOpArith can have more than one right children.
    QuiltExpression[] operands = arith.getOperands();
    if (operands.length > 1) {
        for (int i = 0; i < operands.length - 1; i++) {
            // create a new arithExp
            BinArithExpression arithExpRight = new BinArithExpressionImp();
            arithExpRight.setType(arith.getOp());
            // set the right child the expression.
            arithExp.setRight(arithExpRight);
            // set the left child of the right expression the current operand.
            arithExpRight.setLeft(handleExpression(visit(operands[i]), subtree));
            // update the arithExp to arithExpRight
            arithExp = arithExpRight;
        }
    }
    // set the last child.
    arithExp.setRight(
        handleExpression(visit(operands[operands.length - 1]), subtree));

    // create the result vector.
    Vector result = new Vector();
    result.addElement(arithRoot);
    result.addElement(subtree);

    level--;
    return result;
}
/**
 * This method is visited when WHERE is encountered in a query
 * It will creates a select node.
 * 
 * @return Vector of BinCOMPExpression + XATTRee
 * @param comp xacute.quilt.BinOpCOMP
 * @since 1.0
 * @see BinOpCOMP
 */
public Object visit(BinOpCOMP comp) {
    level++;
    if (isDebug())
        System.out.println(indent() + "BinOpCOMP");

    BinCOMPExpression compExp = new BinCOMPExpressionImp();

    // create the subtree.
    XATDag subtree = new XATDagImp();

    //set the type to whatever type of comparison it is
    compExp.setType(comp.getOp());

    // set the left operand.
    compExp.setLeft(handleExpression(visit(comp.getE()), subtree));

    // BinOpComp should only has one right child.
    QuiltExpression[] operands = comp.getOperands();
    if (operands.length != 1) {
        System.err.println("ERROR: BinOpComp has more than two operands!");
    }
    compExp.setRight(handleExpression(visit(operands[0]), subtree));

    // create the result vector.
    Vector result = new Vector();
    result.addElement(compExp);
    result.addElement(subtree);

    level--;
    return result;
}
/**
 * Create XAT Tree for Binary OR Expression.
 * This method is visited an OR expression encountered in a query
 * 
 * @return Vector of BinORExpression + XATDag
 * @param orOp xacute.quilt.BinOpOR
 * @since 1.0
 * @see BinOpOR
 */
public Object visit(BinOpOR orOp) {
    level++;
    if (isDebug())
        System.out.println(indent() + "BinOpOR");

    // create the subtree.
    XATDag subtree = new XATDagImp();

    //initialize a new BinORExpression
    BinORExpression orExp = new BinORExpressionImp();

    BinORExpression rootExp = orExp;

    // set the left operand.
    orExp.setLeft(handleExpression(visit(orOp.getE()), subtree));

    // like the AND, it can have more than one operands.
    QuiltExpression[] operands = orOp.getOperands();

    for (int i = 0; i < operands.length - 1; i++) {
        BinORExpression innerExp = new BinORExpressionImp();
        orExp.setRight(innerExp);
        innerExp.setLeft(handleExpression(visit(operands[i]), subtree));
        orExp = innerExp;
    }

    // set the last operands.
    orExp.setRight(handleExpression(visit(operands[operands.length - 1]), subtree));

    // create the result vector.
    Vector result = new Vector();
    result.addElement(rootExp);
    result.addElement(subtree);

    level--;
    return result;
}
/**
 * For the union, two operators will be created.
 * First, is the merge operator.
 * Second, is the xml union operaotr.
 * The xml union operator will take the two subtrees's output columns,
 * and then, generate a new output column.
 * The merge operator will make two input tables as one table.
 * Also, according to Kweelt's specification, it only take two sources.
 * 
 * @return XATDag
 * @param union xacute.quilt.BinOpUnion
 * @since 1.0
 * @see BinOpUnion
 */
public Object visit(BinOpUnion union) {
	level++;
	if(isDebug()) System.out.println(indent() + "BinOpUnion");

	// create the result.
	XATDag result = new XATDagImp();

	// create the XML Union operator.
	DagNode xuNode = new DagNodeImp();
	XMLUnion xu = builder.newXMLUnion(xuNode);

	// create the Merge Operator.
	DagNode mNode = new DagNodeImp();
	Merge merge = builder.newMerge(mNode);

	// then, get the first expression.
	QuiltExpression first = union.getE();

	// then, get the rest of the expresion.
	QuiltExpression[] rest = union.getOperands();

	// Here we only consider two operands, because Kweelt assumes that.
	// visit the left operand, and it's subtree.
	XATDag leftXAT = (XATDag)visit(first);

	// visit the right operand, and it's subtree.
	XATDag rightXAT = (XATDag)visit(rest[0]);

	// make the two column names of the top of the two subtrees as the parameters in the XML Union Operator.
	xu.addColumnName(leftXAT.getRoot().getOperator().getOutputColumnName());
	xu.addColumnName(rightXAT.getRoot().getOperator().getOutputColumnName());

	// create the new output column name for the XML Union Operator.
	this.numNodes ++;
	xu.setOutputColumnName("$col" + this.numNodes);

	// add two children to the Merge operator node.
	mNode.addChild(leftXAT.getRoot());
	leftXAT.getRoot().setParent(mNode);
	mNode.addChild(rightXAT.getRoot());
	rightXAT.getRoot().setParent(mNode);

	// add the XML Union to the result.
	result.appendXATNode(xuNode);

	// add the Merge Tree to the result.
	XATDag mergeXAT = new XATDagImp();
	mergeXAT.setRoot(mNode);
	result = mergeTreeWithSources(result,mergeXAT);

	// return the result.
	return result;
}
/**
 * Create XAT Tree for Current Node.
 * 
 * @return VarStepImp.
 * @param curN xacute.quilt.CurrentNode
 * @since 1.0
 * @see VarStepImp
 */
public Object visit(CurrentNode curN) {
	return new VarStepImp(this.currentNodeColumnName);
}
/**
 * Create XAT Tree for Delete Update Operator.
 * Called when a Delete Update is encountered.
 * 
 * @return XATDag of deleteNode.
 * @param del DeleteUpdate
 * @since 1.0
 * @see DeleteUpdate
 */
public Object visit(DeleteUpdate del) {

    XATDag result = new XATDagImp();

    DagNode deleteNode = new DagNodeImp();
    Delete deleteOp = new DeleteImp(deleteNode, updateContext, del.getvarnameT());
    deleteNode.setOperator(deleteOp);

    result.setRoot(deleteNode);

    return result;
}
/**
 * Visited when a RETURN clause is encountered
 * 
 * @return XATDag of Tagger.
 * @param ec the element construct expression to be visited.
 * @since 1.0
 * @see ElementConstruct
 */
public Object visit(ElementConstruct ec) {

    level++;
    if (isDebug())
        System.out.println(indent() + "ElementConstruct");

    // prepare the result.
    XATDag result = new XATDagImp();

    // create the pattern of the element construct.
    RootNode root = new RootNodeImp();
    DOMPattern pattern = new DOMPatternImp(root);

    /* Create a new tag node */
    DagNode newTagNode = new DagNodeImp();
    Tagger newTag = builder.newTagger(newTagNode, pattern);

    //set the output columnName for the tagger node
    numNodes++;
    newTag.setOutputColumnName("$col" + numNodes);

    newTagNode.setOperator(newTag);

    // Add it into the XATDag.
    result.setRoot(newTagNode);

    // Get the tagger Name and make it as a tag node in the pattern.
    TagNode tag = new TagNodeImp(ec.getName().getTagName());
    root.addChild(tag);
    tag.setParent(root);

    // get the attributes list.
    AttributeExpression[] atts = ec.getAtts();

    // process the attributes list. Consider to reuse the processContent.
    XATDag attTree = null;
    if (atts.length != 0) {
        Vector attXAT = this.processAtts(atts);
        // get the AttributeNodes list, please notice every AttributeNode has a ColumnNameNode underneath it if there is a XATDag generated for that AttributeNode.
        Vector attNodes = (Vector) attXAT.elementAt(0);
        // get the XATDag
        attTree = (XATDag) attXAT.elementAt(1);
        // Add the AttributeNodes to the Pattern.
        for (int i = 0; i < attNodes.size(); i++) {
            AttributeNode attNode = (AttributeNode) attNodes.elementAt(i);
            tag.addChild(attNode);
            attNode.setParent(tag);
        }
        // Add the XATDag to the ElementConstruct's Tree.
        //result = mergeTreeWithSources(result,attTree);
    }

    // get the content.
    QuiltExpression[] content = ec.getContent();

    // process the content.
    Vector con = this.processContent(content);
    // get the cols list.
    Vector cols = (Vector) con.elementAt(0);
    // get the XATDag
    XATDag conTree = (XATDag) con.elementAt(1);

    // add the column names into the tagger.
    for (int i = 0; i < cols.size(); i++) {
        if (cols.elementAt(i) instanceof ColumnNameImp) {
            ColumnNameNodeImp tempNode =
                new ColumnNameNodeImp(((ColumnName) cols.elementAt(i)).getColumnName());
            tag.addChild(tempNode);
            tempNode.setParent(tag);
        } else {
            TextNodeImp tempNode = new TextNodeImp(cols.elementAt(i).toString());
            tag.addChild(tempNode);
            tempNode.setParent(tag);
        }
    }

    //To deal with "tager attr1=$a, $b.../tag" bug
    //If the tree already has an aggregate or (GB/agg)
    //because this kind of agg on $a will remove duplicates of $b and etc. 
    //So if the atts.length!=0 then we need to generate merge operator.
    //similar with processContent()
    //Song Wang
    XATDag mergeTree = conTree;
    if (atts.length != 0) {
        // create the Merge node.
        DagNode mergeNode = new DagNodeImp();
        Merge mergeOp = builder.newMerge(mergeNode);

        // create the Merge Tree.
        XATDag mergeXAT = new XATDagImp();
        mergeXAT.setRoot(mergeNode);

        // create the upper Tree.
        mergeTree = new XATDagImp();

        // put conTree under the merge operator.
        if (conTree.getRoot() != null) {
            mergeNode.addChild(conTree.getRoot());
            conTree.getRoot().setParent(mergeNode);
        }
        // put attTree under the merge operator.
        if (attTree.getRoot() != null) {
            mergeNode.addChild(attTree.getRoot());
            attTree.getRoot().setParent(mergeNode);
        }

        mergeTree = mergeTreeWithSources(mergeTree, mergeXAT);

    }

    // update the result.
    result = mergeTreeWithSources(result, mergeTree);

    //put the tag node in the binding hash table
    ListPattern lp = new ListPatternImp();
    if ((newTag.getPattern() instanceof DOMPattern)) {
        lp =
            ((DOMPattern) newTag.getPattern()).toListPattern(
                ((DOMPattern) newTag.getPattern()).getRoot());
    } else {
        lp = (ListPattern) newTag.getPattern();
    }

    VarStep v1 = new VarStepImp(newTag.getOutputColumnName());

    level--;
    return result;
}
/**
 * Process the Filter Expression.
 * 
 * @return XATDag
 * @param filterExpr xacute.quilt.FilterExpression
 * @since 1.0
 * @see FilterExpression
 */
public Object visit(FilterExpression filterExpr) {
    level++;
    if (isDebug())
        System.out.println(indent() + "FilterExpression");

    //visit the filter expression
    Object visitResult = visit(filterExpr.getE());
    XATDag result = null;
    if (visitResult instanceof XATDag)
        result = (XATDag) visitResult;
    else if (visitResult instanceof VarStepImp) {
        // make a tree with only one node, which is a dummy rename operator.
        result = new XATDagImp();
        DagNode root = new DagNodeImp();
        NameColumn ren = builder.newNameColumn(root);
        ren.setOriginalColumnName(((VarStepImp) visitResult).getStepValue());
        ren.setNewColumnName(((VarStepImp) visitResult).getStepValue());
        ren.setOutputColumnName(((VarStepImp) visitResult).getStepValue());
        result.setRoot(root);
    } else {
        System.err.println("ERROR: FilterExpression has unhandled Expression.");
    }
    /*Because filter expression will filter each element in this evaluation result, which will be
    a collection, a nav-unnest will be created to apply the filter for each element in this collection.*/

    // add a navunnest on top of this result tree.
    DagNode unnestNode = new DagNodeImp();
    NavUnnest unnestOp = new NavUnnestImp(unnestNode);
    unnestNode.setOperator(unnestOp);

    String outputColumnName = result.getRoot().getOperator().getOutputColumnName();
    unnestOp.setEntryPoint(
        new NavigationStep[] { new VarStepImp(outputColumnName)});
    this.numNodes++;
    unnestOp.setOutputColumnName("$col" + this.numNodes);

    result.insertRoot(unnestNode);

    //now visit the filters
    //put the filters on top of the node.
    QuiltExpression[] filters = filterExpr.getFilters();
    for (int i = 0; i < filters.length; i++) {
        XATDag filterTree =
            createSelect(filters[i], result.getRoot().getOperator().getOutputColumnName());
        filterTree = mergeTreeWithSources(filterTree, result);
        result = filterTree;
    }

    //NOTE: For where clause. we DONOT need to put an aggregate upon select
    //      But for filter. we DO need to put an aggregate upon select
    //      In order to get rid of empty selection problem.
    //So add another aggregate node
    //Song Wang 
    //create an Combine node
    DagNode aggNode = new DagNodeImp();
    Combine aggregate = builder.newAggregate(aggNode);
    aggregate.setOutputColumnName(
        result.getRoot().getOperator().getOutputColumnName());
    aggNode.setOperator(aggregate);
    aggregate.setNode(aggNode);

    // add the aggregate node.
    XATDag aggTree = new XATDagImp();
    aggTree.setRoot(aggNode);
    aggTree = mergeTreeWithSources(aggTree, result);
    result = aggTree;

    level--;
    return result;
}
/**
 * Called if a FLWR (For-Let-Where-Return) expression is encountered
 * (OR when a FOR is found in the query)
 * <p>
 * The tree should be like:
 * Combine -> (Sort) -> For bindings -> (Where) -> Return Clause -> (Let bindings)
 * | result            |               | returnClause                               |
 * 
 * @return XATDag
 * @param flwr FLWR expression.
 * @since 1.0
 */
public Object visit(FLWRExpression flwr) {
    level++;
    if (isDebug())
        System.out.println(indent() + "FLWR");

    // prepare the result.
    XATDag result = new XATDagImp();

    //create an Combine node
    DagNode aggNode = new DagNodeImp();
    Combine aggregate = builder.newAggregate(aggNode);
    aggNode.setOperator(aggregate);

    // add the aggregate node.
    result.setRoot(aggNode);

    // visit the return clause.
    Object returnVisit = visit(flwr.getE());
    XATDag returnClause = null;
    if (returnVisit instanceof XATDag) // an single element construct clause.
        returnClause = (XATDag) returnVisit;
    else if (returnVisit instanceof Vector) { // an arithmatic expression.
        returnClause = this.expressionVectorToXAT((Vector) returnVisit);
    } else {
        System.err.println(
            "ERROR: Unknown type of returnVisit in ParserTraverse.visit(FLWRExpression)");
    }

    // keep return clause' output column name, used in the Combine operator.
    String returnCol = returnClause.getRoot().getOperator().getOutputColumnName();

    // make a block identification for the return clause.
    DagNode blockNode = new DagNodeImp();
    NameColumn blockOp = new NameColumnImp(blockNode, returnCol, returnCol);
    // set the block name is return.
    blockOp.setBlockDesc("return");

    // put this node as the root for the returnClause.
    returnClause.insertRoot(blockNode);

    /*
    * Note: The block node is used for decor to put the LOJ below that node.
    * 6/27/2003
    */

    // aggregate should have the returnClause's columname as the output columnname.
    aggregate.setOutputColumnName(returnCol);

    // The where and sort clause should be on the top of the returnClause.
    // visit the where clause.
    if (flwr.getWhere() != null) {
        returnClause = handleWhereClause(flwr, returnClause, returnCol);
    }

    // visit the sort clause. 
    // sort clause should below the aggregate operator.
    SortClause[] sort = flwr.getSort();
    if (sort.length >= 1) {
        // create the Sort Operator.
        DagNode orderNode = new DagNodeImp();
        OrderBy orderOp = builder.newOrderBy(orderNode);
        orderOp.setNode(orderNode);
        // create the orderby tree.
        XATDag orderTree = new XATDagImp();
        orderTree.setRoot(orderNode);
        // put the list of the column names in the OrderBy.
        for (int i = 0; i < sort.length; i++) {
            // the sort clause should return a list of XATDag.
            XATDag sortTree = (XATDag) visit(sort[i]);
            // add the column name.
            String columnname = sortTree.getRoot().getOperator().getOutputColumnName();
            orderOp.addColumnName(columnname);
            // append the xatTree.
            orderTree = mergeTreeWithSources(orderTree, sortTree);
        }
        // put the Sort Operator below the aggregate, which is the end of the resultTree.
        result = mergeTreeWithSources(result, orderTree);
    }

    // visit the bindings. 
    // get all the bindings. For each forbinding a FOR node will be created.
    Binding[] bindings = flwr.getBindings();

    visitBindings(bindings, result, returnClause);

    level--;
    return result;
}
/** 
 * Difference from the FLWR, it doesn't have the Combine Operator.
 * Called when a FLWU(Update) Expression in encountered
 *
 * @return XATDag.
 * @param flwu FLWU expression.
 * @since 1.0
 */
public Object visit(FLWUExpression flwu) {
    level++;
    if (isDebug())
        System.out.println(indent() + "FLWU");

    // prepare the result.
    XATDag result = new XATDagImp();

    // get the update context.
    this.updateContext = flwu.getVarname();

    // prepare the update clause.
    XATDag updateClause = new XATDagImp();

    // get the sequence of updates.
    QuiltExpression[] updates = flwu.getUpdates();

    // visit each update clause, and merged back to the updateClause XAT subtree.
    for (int i = 0; i < updates.length; i++)
        updateClause = mergeTreeWithSources(updateClause, (XATDag) visit(updates[i]));

    // The where clause should be below of the updateClause.
    // visit the where clause.
    if (flwu.getWhere() != null) {
        XATDag whereClause = null;
        try {
            // visit the expressions in where clause and create the subtree.
            whereClause =
                createSelect(
                    (Operator) flwu.getWhere(),
                    updateClause.getRoot().getOperator().getOutputColumnName());
        } catch (ClassCastException e) {
            System.out.println("Problem with Operator (kweelt) handling in FLWU");
            e.printStackTrace();
        }
        // Added below the update clause.
        updateClause = mergeTreeWithSources(updateClause, whereClause);
    }

    // visit the bindings. 
    // get all the bindings. For each forbinding a FOR node will be created.
    Binding[] bindings = flwu.getBindings();

    visitBindings(bindings, result, updateClause);

    level--;
    return result;
}
/**
 * Order of generated forbinding: NavUnnest->(Distinct)->Subtree
 * Contains the data for the FOR bindings
 *
 * @return XATDag.
 * @param b ForBinding
 * @since 1.0
 */
public Object visit(ForBinding b) {
    level++;
    if (isDebug())
        System.out.println(indent() + "ForBinding");

    //remember the for binding var name.
    String var = "$" + b.getName();

    //prepare a forbinding tree.
    XATDag bindingTree = new XATDagImp();

    // get the binding subtree.
    XATDag subtree = (XATDag) visit(b.getQE());

    // create a navigate unnesting operator.
    DagNode navNode = new DagNodeImp();
    Navigate navOp = new NavUnnestImp(navNode);
    navOp.setEntryPoint(
        new NavigationStep[] {
             new VarStepImp(subtree.getRoot().getOperator().getOutputColumnName())});
    navOp.setBinding(var);
    navOp.setOutputColumnName(var);
    bindingTree.setRoot(navNode);

    // add to the binding hashtable.
    NavigationStep v2[] = navOp.getEntryPoint();

    //if distinct, then create a distinct operator.
    if (b.getDistinct()) {
        String distinctVar = subtree.getRoot().getOperator().getOutputColumnName();
        // create the node.
        DagNode distinctNode = new DagNodeImp();
        numNodes++;
        Distinct distinctOp = new DistinctImp(distinctNode);
        // set the distinct columns.
        distinctOp.setInputColumnName(distinctVar);
        // set the output column name.
        distinctOp.setOutputColumnName(distinctVar);
        // add the node to the bindingTree.
        bindingTree.appendXATNode(distinctNode);
    }

    // connect with the binding tree. 
    bindingTree = mergeTreeWithSources(bindingTree, subtree);

    level--;
    return bindingTree;
}
/**
 * The superclass of all the function calls
 *
 * @return XATDag
 * @param funPrim xacute.quilt.FunCall
 * @since 1.0
 */
public Object visit(xacute.quilt.FunCall funPrim) {
    level++;
    if (isDebug())
        System.out.println(indent() + "FunCall");

    // prepare the result XAT.
    XATDag result = new XATDagImp();

    // create the function Node.
    DagNode funcNode = new DagNodeImp();
    Function funcOp = builder.newFunction(funcNode);

    if (funPrim.getName().equals("document")) {
        // we will create a source operator.
        DagNode sourceNode = new DagNodeImp();
        Source sourceOp = builder.newSource(sourceNode);
        // get the document name for the source operator.
        String docName = (String) visit(funPrim.getArgs()[0]);
        sourceOp.setName(docName);
        // set the output column name for this operator.
        this.numNodes++;
        sourceOp.setOutputColumnName("s" + this.numNodes);

        NodeStep ns1 = new NodeStepImp("SOURCE");
        NavigationStep[] n1 = new NavigationStepImp[1];
        n1[0] = ns1;
        // add the source op to the result.
        result.setRoot(sourceNode);
        level--;
        // return the source operator.
        return result;
    }

    // otherwise, it's other complex functions.
    // get all the parameters.
    QuiltExpression[] args = funPrim.getArgs();

    // Vector for remember each argument's column name.
    Vector argCols = new Vector();
    // visit each arguments, the tree created should be added below.
    for (int i = 0; i < args.length; i++) {
        Object visitResult = visit(args[i]);
        if (visitResult instanceof XATDag) {
            XATDag argTree = (XATDag) visitResult;
            // get the output column, should be put into the FunctionOp as the arguments.
            argCols.addElement(
                new ColumnNameImp(argTree.getRoot().getOperator().getOutputColumnName()));
            // connect the argument Tree to the result.
            result = mergeTreeWithSources(result, argTree);
        } else if (visitResult instanceof Vector) {
            //it is an expression vector composed of expression + XATDag
            // make the expression vector into a XATDag.
            argCols.addElement(expressionVectorToXAT((Vector) visitResult));
        } else { // it might be a valuestring, or other non XATDag object.
            argCols.addElement(visitResult);
        }
    }

    // set the name of the function.
    funcOp.setName(funPrim.getName());
    // set the parameters of the function.
    funcOp.setParams(argCols.toArray());
    // set the output column name.
    numNodes++;
    funcOp.setOutputColumnName("$col" + numNodes);

    // add the funcOp back to the result at the root node.
    if (result.getRoot() != null) {
        funcNode.addChild(result.getRoot());
        result.getRoot().setParent(funcNode);
        result.setRoot(funcNode);
    } else {
        result.setRoot(funcNode);
    }

    //add function to the binding hash table
    NodeStep ns1 = new NodeStepImp("FUNCTION_" + funcOp.getName());
    NavigationStep[] n1 = new NavigationStepImp[1];
    n1[0] = ns1;

    level--;
    return result;
}
/**
 * Visit defined functions using visit(FunCall)
 * 
 * @return java.lang.Object
 * @param fundef xacute.quilt.FunDef
 * @since 1.0
 */
public Object visit(xacute.quilt.FunDef fundef) {
	level++;
 	if(isDebug()) System.out.println(indent() + "FunDef");
	
	XATDag result = (XATDag)toXATTree(visit(fundef.getBody()));
	level--;

	return result;
}
/**
 * Visit user defined functions
 * 
 * @return java.lang.Object
 * @param param xacute.quilt.FunUserDefined
 * @since 1.0
 */
public Object visit(xacute.quilt.FunUserDefined funUD) {
    level++;
    if (isDebug())
        System.out.println(indent() + "FunUserDefined");

    // prepare the result XAT.
    XATDag result = new XATDagImp();

    // create the function Node.
    DagNode funcNode = new DagNodeImp();
    edu
        .wpi
        .cs
        .dsrg
        .xmldb
        .xat
        .common
        .operator
        .specialoperator
        .function
        .FunUserDefined funcOp =
        new edu
            .wpi
            .cs
            .dsrg
            .xmldb
            .xat
            .common
            .operator
            .specialoperator
            .function
            .FunUserDefined(funcNode);

    // otherwise, it's other complex functions.
    // get all the parameters.
    QuiltExpression[] args = funUD.getArgs();

    // Vector for remember each argument's column name.
    Vector argCols = new Vector();
    // visit each arguments, the tree created should be added below.
    for (int i = 0; i < args.length; i++) {
        Object visitResult = visit(args[i]);
        if (visitResult instanceof XATDag) {
            XATDag argTree = (XATDag) visitResult;
            // get the output column, should be put into the FunctionOp as the arguments.
            argCols.addElement(argTree.getRoot().getOperator().getOutputColumnName());
            // connect the argument Tree to the result.
            result = mergeTreeWithSources(result, argTree);
        } else { // it might be a valuestring, or other non XATDag object.
            argCols.addElement(visitResult);
        }
    }

    // set the name of the function.
    funcOp.setName(funUD.getName());
    // set the parameters of the function.
    funcOp.setParams(argCols.toArray());
    // set the output column name.
    numNodes++;
    funcOp.setOutputColumnName("$col" + numNodes);

    // add the funcOp back to the result at the root node.
    if (result.getRoot() != null) {
        funcNode.addChild(result.getRoot());
        result.getRoot().setParent(funcNode);
        result.setRoot(funcNode);
    } else {
        result.setRoot(funcNode);
    }

    //add function to the binding hash table
    NodeStep ns1 = new NodeStepImp("FUNCTION_" + funcOp.getName());
    NavigationStep[] n1 = new NavigationStepImp[1];
    n1[0] = ns1;

    level--;
    return result;
}
/**
 * Create XAT Tree for IF THEN ELSE structure.
 * IfThenElse will have a condition, e.g., like the where clause,
 * Then, it has two branches. Left for the then, and Right for the else.
 * 
 * @return java.lang.XATDag
 * @param ite xacute.quilt.IfThenElse
 * @since 1.0
 */
public Object visit(IfThenElse ite) {
    level++;
    if (isDebug())
        System.out.println(indent() + "IfThenElse");

    // create a result, which is a XATDag.
    XATDag result = new XATDagImp();

    // create a IF-THEN-ELSE operator and its node.
    DagNode ifNode = new DagNodeImp();
    IfExpr ifop = new IfExprImp(ifNode);

    // make it the root of the result.
    result.setRoot(ifNode);

    // get the condition.
    // The condition should be returned as a subtree.
    // The expression is test if the final result is TRUE. For example, an unempty collection, an arithmatic value, etc.
    XATDag exprXAT = this.toXATTree(visit(ite.getIfE()));
    // get the output column name.
    String colName = exprXAT.getRoot().getOperator().getOutputColumnName();
    // set the condition.
    ifop.setIfExpr(new TerminalExpressionImp(new ColumnNameImp(colName)));
    // add the condition Tree below.
    result = mergeTreeWithSources(result, exprXAT);

    // make a new output column, which is shared by IfOp, and its two subtrees.
    this.numNodes++;
    String outColumnName = "$col" + this.numNodes;

    // set if's output column name.
    ifop.setOutputColumnName(outColumnName);

    // get the then expression.
    XATDag thenXAT = (XATDag) toXATTree(visit(ite.getThenE()));
    // make a new rename operator to rename to output column into the if's output columnname.
    DagNode renNode = new DagNodeImp();
    NameColumn renOp = new NameColumnImp(renNode);
    renOp.setOriginalColumnName(
        thenXAT.getRoot().getOperator().getOutputColumnName());
    renOp.setNewColumnName(outColumnName);
    renOp.setOutputColumnName(outColumnName);
    // update the then XAT tree to have this new root.
    thenXAT.insertRoot(renNode);
    // set the then Expression.
    ifop.setThenXAT(thenXAT);

    // get the else expression.
    XATDag elseXAT = (XATDag) toXATTree(visit(ite.getElseE()));
    // make a new rename operator to rename to output column into the if's output columnname.
    renNode = new DagNodeImp();
    renOp = new NameColumnImp(renNode);
    renOp.setOriginalColumnName(
        elseXAT.getRoot().getOperator().getOutputColumnName());
    renOp.setNewColumnName(outColumnName);
    // update the else XAT tree to have this new root.
    elseXAT.insertRoot(renNode);
    // set the else Expression.
    ifop.setElseXAT(elseXAT);

    level--;
    return result;
}
/**
 * Called when an Insert Update is encountered
 *
 * @return XATDag
 * @param InsertUpdate ins
 * @since 1.0
 */
public Object visit(InsertUpdate ins) {

    level++;
    if (isDebug())
        System.out.println(indent() + "Insert");

    // prepare the result XAT.
    XATDag result = new XATDagImp();

    // create the InsertUpdate Op and Node.
    DagNode insertNode = new DagNodeImp();
    Insert insertOp = null;

    // set the insert Order.
    int orderPredicate = Insert.UNKNOWN;

    // get the order information from the InsertUpdate 
    if (ins.getBefore() == true)
        orderPredicate = Insert.BEFORE;
    else if (ins.getBefore() == false)
        orderPredicate = Insert.AFTER;

    if (ins.getvarnameT1() == null) // if this is not an ordered INSERT
        if (ins.getContent() == null) // ie a variable/update function appears
            {
            // if there is no update function
            if (ins.getUpdateFunction() == null)
                insertOp = new InsertImp(insertNode, updateContext, ins.getvarnameT());
            else {
                String params[] = { "one", "two" };
                int updateFunction = 99;

                params[0] = ins.getUpdateFunction().getstringT();
                params[1] = ins.getUpdateFunction().getContent().getStringContent();

                if (ins.getUpdateFunction().getnewitemT().equals("new_attribute"))
                    updateFunction = Insert.NEW_ATTRIBUTE;
                else if (ins.getUpdateFunction().getnewitemT().equals("new_ref"))
                    updateFunction = Insert.NEW_REF;

                insertOp = new InsertImp(insertNode, updateContext, updateFunction, params);
            }
        } else {
            insertOp = new InsertImp(insertNode, updateContext);
            // get the  output column name from the content
            XATDag subtree = (XATDag) visit(ins.getContent().getContent());
            insertOp.setVariable(subtree.getRoot().getOperator().getOutputColumnName());
            // make the subtree a branch of the insertOp.
            insertNode.addChild(subtree.getRoot());
            subtree.getRoot().setParent(insertNode);
        }

    else // if this is an ordered INSERT
        {
        if (ins.getContent() == null) // ie a variable is to be inserted
            insertOp =
                new InsertImp(
                    insertNode,
                    updateContext,
                    ins.getvarnameT(),
                    ins.getvarnameT1(),
                    orderPredicate);
        else {
            insertOp =
                new InsertImp(insertNode, updateContext, ins.getvarnameT1(), orderPredicate);
            // get the output column name from the content    
            // get the  output column name from the content
            XATDag subtree = (XATDag) visit(ins.getContent().getContent());
            insertOp.setVariable(subtree.getRoot().getOperator().getOutputColumnName());
            // make the subtree a branch of the insertOp.
            insertNode.addChild(subtree.getRoot());
            subtree.getRoot().setParent(insertNode);
        }
    }

    insertNode.setOperator(insertOp);

    // add the insertNode to the result XAt.
    result.setRoot(insertNode);

    level--;
    return result;
}
/**
 * The let binding will add a rename operator to show the binding.
 * 
 * @return XATDag.
 * @param let xacute.quilt.LetBinding
 * @since 1.0
 */
public Object visit(LetBinding let) {
    level++;
    if (isDebug())
        System.out.println(indent() + "LetBinding");

    //get the value of the let binding
    String var = "$" + let.getName();

    //visit the navigate node structure
    XATDag result = (XATDag) visit(let.getQE());

    // get the output columnname of the subtree.
    String oldName = result.getRoot().getOperator().getOutputColumnName();

    //create rename operator.
    DagNode renameNode = new DagNodeImp();
    NameColumn renameOp = builder.newNameColumn(renameNode);
    renameOp.setOriginalColumnName(oldName);
    renameOp.setNewColumnName(var);
    renameOp.setOutputColumnName(var);

    // put it on the top of the result XAT.
    if (result.getRoot() != null) {
        renameNode.addChild(result.getRoot());
        result.getRoot().setParent(renameNode);
        result.setRoot(renameNode);
    } else {
        result.setRoot(renameNode);
    }

    // add to the binding hashtable.
    VarStep value = new VarStepImp(renameOp.getOriginalColumnName());

    level--;
    return result;
}
/**
 * Determines the type of XPath (child, descendant-or-self)
 * Contains an axis variable
 * Calls the NodeTest object
 * The result is either a NameTest or NodeType object visit call
 * <p>
 * The LocationStep will be converted into one Navigate operator with optional multiple select operators.
 * <p>
 * The axies have:
	AXISSPECIFIERS = {"ancestor", 
				    "ancestor-or-self",
				    "attribute",
				    "child",
				    "descendant",
				    "descendant-or-self",
				    "following",
				    "following-sibling",
				    "namespace",
				    "parent",
				    "preceding",
				    "preceding-sibling",
				    "self"};
 *
 * @return A Vector of multiple elements: 1) Step, 2-x) XATTrees for predicates.
 * @param lstep LocationStep
 * @since 1.0
 */
public Object visit(LocationStep lstep) {
    level++;
    if (isDebug())
        System.out.println(indent() + "LocationStep");

    // prepare the result.
    Vector result = new Vector();

    // create the Step.
    NavigationStep step = null;

    // get the stepname.
    String name = null;
    // there could be two kinds of the test: NameTest and NodeType.
    // NameTest for att, child. NodeType for text.
    NodeTest test = lstep.getTest();

    if (test instanceof NodeType) {
        if (((NodeType) test).getType() == 3) { // it's a text().
            step = new TextStepImp();
        } else if (((NodeType) test).getType() == 0) { // it's descedent "//" step.
            step = new DescendantStepImp(); // also it's axis is 5.
        }

    } else if (test instanceof NameTest) {

        name = ((NameTest) test).getName();

        if (name == null) // it's a '*'
            name = "*";

        // get the axis
        switch (lstep.getAxis()) {
            case 2 : // attribute
                step = new AttStepImp(name);
                break;
            case 3 : // child
                step = new NodeStepImp(name);
                break;
            case 4 : // descendant
                step = new DescendantStepImp();
                break;
            default :
                System.err.println("ERROR: Unhandled Location Step Axis:" + lstep.getAxis());
                // temporarily make a dummy step.
                step = new NodeStepImp(lstep.getAxis() + ":" + name);
        }
    } else { // an UN handled step type.
        System.err.println("ERROR: Unhandled Location Step Type:" + test.toString());
        // temporarily make a dummy step.
        step = new NodeStepImp(test.toString());
    }

    // add the step to the result.
    result.addElement(step);

    // NOTE: We get the output column name for the to be created Nav operator.
    // which is the this.currentNodeColumnName.
    // remember the currentNodeColumnName for this LocationStep.
    // so that, it will not be overwritten by later Nav.
    String columnName = this.currentNodeColumnName;

    /** handle the predicates, and make them into select operators.*/
    QuiltExpression[] predicates = lstep.getPredicates();

    for (int i = 0; i < predicates.length; i++) {
        level++;
        if (isDebug())
            System.out.println(indent() + "Predicate");
        // create the subtree for the predicates.
        XATDag subtree = new XATDagImp();
        if (predicates[i] instanceof Operator) {
            // Then here, we use the kept column name.
            subtree = createSelect((Operator) predicates[i], columnName);
        } else if (predicates[i] instanceof Nav) {
            subtree = createSelect((Nav) predicates[i], columnName);
        }
        // add the subtree to the result.
        result.addElement(subtree);
        level--;
    }

    level--;
    return result;
}
/**
 * A new Navigate node is created.
 * NavigationStep objects are assigned to the Navigate node
 * <p>
 * NavCollection Operator is created instead of NavigationImp.
 * <p>
 * In the ForBinding, we will add a NavigateImp (unnesting) to iterate the answers.
 * <p>
 * If the LocationStep has a filter, then a rename (dummy) is created at the root, so that, we know
 * which column is the focus.
 * <p>
 * For the locationstep, it will return a vector.
 * Also, for the E, it returns a XATDag or a String for the Var.
 *
 * @return XATDag
 * @param nav Nav
 * @since 1.0
 */
public Object visit(Nav nav) {
    level++;
    if (isDebug())
        System.out.println(indent() + "NAV");

    // default mode of the navigates is context-sensitive.
    boolean sensitive = true;

    Object head = null;
    //visit everything under the navigate
    head = visit(nav.getE());
    //there are two types of the results.
    //either a XATDag or a VarStep.
    EntryTree et = findEntryPointAndCreateHeaderTree(head);
    NavigationStep entry = et.getEntry();
    XATDag result = et.getTree();

    Step[] steps = nav.getSteps();

    // Used to keep track of the last ColumnName generated by the Nav.
    String navColumnName = "ERROR";
    String context = entry.getStepValue();
    if (context == null) {
        if (isDebug())
            System.err.println("Context Insensitive Mode!");
        // get the context from the global variable.
        context = this.xPathContext;
    }

    //visit the steps
    if (steps.length == 0) { //there is no location steps.
        // Empty Navigation will create a Dummay Rename Node.
        // create a dummy rename operator.
        // it rename the name into a same name.
        DagNode renameNode = new DagNodeImp();
        NameColumn renameOp = builder.newNameColumn(renameNode);
        renameOp.setOriginalColumnName(entry.getStepValue());
        renameOp.setNewColumnName(entry.getStepValue());
        renameOp.setOutputColumnName(entry.getStepValue());
        // put rename Op on top of the result.
        result.insertRoot(renameNode);
    } else {

        // NOTE: Regarding to Merging multiple navigates into one navigates with multiple steps,
        // we don't do it in the parser traverse, instead, we put it into the rewriting rules.
        for (int i = sensitive ? 0 : 1; i < steps.length; i++) {
            // create the Navigate node.
            DagNode navNode = new DagNodeImp();
            Navigate navOp = builder.newNavUnnest(navNode);
            // set the entry point. 
            navOp.setEntryPoint(new NavigationStep[] { entry });
            // set the output column name and binding name.
            // NOTE: We generate the new column name before we visit the content of the LocationStep
            //       is for the reason of handling the CurrentNode in the Filter Expression.
            this.numNodes++;
            navOp.setBinding("$col" + numNodes);
            navOp.setOutputColumnName("$col" + numNodes);
            // NOTE: We use a global variable to keep track of the "CurrentNode" context for the filter expression.
            // NOTE: We remeber the old currentNodeColumnName, and after the visit, we restore the currentNodeColumnName
            String oldColumnName = this.currentNodeColumnName;
            this.currentNodeColumnName = "$col" + numNodes;
            // Get a vector with Step + XATTrees (for how to get to such step).
            Vector tmp = (Vector) visit(steps[i]);
            // NOTE: Restore the original NodeColumnName.
            this.currentNodeColumnName = oldColumnName;
            // Get the destination step from the result.
            NavigationStep dest = (NavigationStep) tmp.elementAt(0);
            // Set the destination.
            navOp.setDestination(new NavigationStep[] { dest });
            // keep the binding destination.
            NavigationStep[] bindDest = new NavigationStep[] { entry, dest };
            // update the entry point.
            entry = new VarStepImp(navOp.getOutputColumnName());

            //put the navOp in the result tree, on the top.
            result.insertRoot(navNode);

            // The output column name for the navigate operator.
            navColumnName = navNode.getOperator().getOutputColumnName();

            // remember the outputcolumn of the navigate operator.
            DagNode renameNode = new DagNodeImp();
            NameColumn renameOp = builder.newNameColumn(renameNode);
            renameOp.setNewColumnName(navColumnName);
            renameOp.setOriginalColumnName(navColumnName);
            renameOp.setOutputColumnName(navColumnName);

            //add the predicate XATTrees on the top of the result tree.
            //FOR each predicate tree, we need to traverse the tree and add a groupby
            //operator for each function which is context sensitive, and also
            //the groupby is on the output column before the navigate operator, which is
            //the entry point of the navigate operator.
            for (int j = 1; j < tmp.size(); j++) {
                XATDag predTree = (XATDag) tmp.elementAt(j);
                preprocessPredicateTree(predTree, navOp.getEntryPoint()[0].getStepValue());
                predTree = mergeTreeWithSources(predTree, result);
                result = predTree;
            }
            //add the rename operator on top of the result tree.
            //the dummy rename operator is used to let the caller know which column to keep track of.
            if (tmp.size() > 1) {
                //put the dummy name holder.
                result.insertRoot(renameNode);

            }
        }
    }

    //visit efunc, the optional last function at the end.
    if (nav.getEfunc() != null) {
        XATDag tail = (XATDag) visit(nav.getEfunc());
        //add the tail on top of the operators.
        tail = mergeTreeWithSources(tail,result);
        result = tail;
    }

    if (steps.length != 0) {
        //also create the aggregate operator to aggregate the results into one collection.
        DagNode gbNode = new DagNodeImp();
        Combine aggOp = builder.newAggregate(gbNode);
        aggOp.setOutputColumnName(navColumnName);

        XATOperator gbOp = createGBOp(aggOp, context);
        gbOp.setNode(gbNode);
        gbNode.setOperator(gbOp);

        //Adding:
        //for filters, we need have groupby for the currentnodecolumn anyway
        //eg. for $b in doc/book[title = $a/title], we need gb on book even for $a/title
        //so redo preprocessPredicateTree if necessary.
        if ((this.currentNodeColumnName != null)
            && (!this.currentNodeColumnName.equals(context))) {
            DagNode gbNode2 = new DagNodeImp();
            XATOperator gbOp2 = createGBOp(gbOp, this.currentNodeColumnName);
            gbOp2.setNode(gbNode2);
            gbNode2.setOperator(gbOp2);
			gbNode = gbNode2;
        }

        result.insertRoot(gbNode);
    }

    level--;
    return result;
}
/**
 * This is the first pattern that is visited
 * The QuiltQuery object contains the entire query
 * The visitor pattern visits the body of the QuiltQuery object,
 * which is the entire Quilt query
 * For each function definitions it will create a new XAT tree.
 * Hence, QuiltQuery will return a vector of XAT trees.
 * Where the first one is the root Tree and rest of them are functions.
 *
 * @return Hashtable of XATTrees. Hashed by function name and "main".
 * @param q QuiltQuery
 * @since 1.0
 */
public Object visit(QuiltQuery q) throws FactoryException {
    try {
        /** create the factory */
        OperatorFactory factory = OperatorFactory.newInstance();
        /** instantiate the operator builder */
        builder = factory.newOperatorBuilder();
    } catch (FactoryException e) {
        /** 
        * there was a problem instantiating the factory
        * so we have to stop generating the tree. 
        */
        throw e;
    }
    level++;
    if (isDebug())
        System.out.println(indent() + "Query");

    // create the Tree Vector.
    this.xatTrees = new Hashtable();

    // prepare the binding table computer. 
    BindingTableComputer bindComp = new BindingTableComputer();

    // get all the function definitions.
    xacute.quilt.FunDef[] funDefs = q.getfunDefs();
    for (int i = 0; i < funDefs.length; i++) {
        // add each function as a separate XAT tree into the Tree vector.
        XATDag funDefTree = (XATDag) visit(funDefs[i]);
        // If the query is not FLWU expression, then we add Expose operator.
        if (!(funDefs[i].getBody() instanceof FLWUExpression)) {
            addExpose(funDefTree);
        }
        // validate the generated tree.
        //funDefTree.validate();
        // clean up the generated Tree.  
        if(System.getProperty("CLEANXAT")!=null)
		    XATCleanUp.doCleanUp(funDefTree);
        // compute its binding table.
        funDefTree.setBindingTable(bindComp.computeBindingTable(funDefTree));
        // remember this tree.
        xatTrees.put(funDefs[i].getName(), funDefTree);
    }

    // visit the body, which is the root of the main XAT, and set the root.
    XATDag mainXAT = (XATDag) visit(q.getBody());

    // put the main query.
    xatTrees.put("main", mainXAT);

    // If the query is not FLWU expression, then we add Expose operator.
    if (!(q.getBody() instanceof FLWUExpression)) {
        addExpose(mainXAT);
    }

    // validate the generated tree.
    mainXAT.validate();
    // clean up the generated Tree.  
	if(System.getProperty("CLEANXAT")!=null)
		XATCleanUp.doCleanUp(mainXAT);
    // set its BindingTable.
    mainXAT.setBindingTable(bindComp.computeBindingTable(mainXAT));

    level--;
    return xatTrees;
}
/**
 * Called when a Rename Update is encountered
 * 
 * @return XATDag
 * @param RenameUpdate re
 * @since 1.0
 */

public Object visit(RenameUpdate re) {

    XATDag result = new XATDagImp();

    DagNode renameNode = new DagNodeImp();
    Rename renameOp = new RenameImp(renameNode, updateContext, re.getvarnameT());
    renameOp.setName(re.getContent().getStringContent());
    renameNode.setOperator(renameOp);

    result.setRoot(renameNode);

    return result;
}
/**
 * Called when a Replace Update is encountered
 * 
 * @param ReplaceUpdate rep
 * @return XATDag
 * @since 1.0
 */
public Object visit(ReplaceUpdate rep) {
    DagNode replaceNode = new DagNodeImp();
    Replace replaceOp = null;

    // if there appears  an update function in the replace
    if (rep.getUpdateFunction() != null) {
        String params[] = { "one", "two" };
        int updateFunction = 99;

        params[0] = rep.getUpdateFunction().getstringT();
        params[1] = rep.getUpdateFunction().getContent().getStringContent();

        if (rep.getUpdateFunction().getnewitemT().equals("new_attribute"))
            updateFunction = Insert.NEW_ATTRIBUTE;
        else if (rep.getUpdateFunction().getnewitemT().equals("new_ref"))
            updateFunction = Insert.NEW_REF;

        replaceOp =
            new ReplaceImp(
                replaceNode,
                updateContext,
                rep.getvarnameT(),
                updateFunction,
                params);
    } else {
        replaceOp = new ReplaceImp(replaceNode, updateContext, rep.getvarnameT());
        // if this is just a simple PCDATA replacement
        if (rep.getContent().getStringContent() != null)
            replaceOp.setPCDATAContent(rep.getContent().getStringContent());

        else {
            // get the  output column name from the content
            XATDag subtree = (XATDag) visit(rep.getContent().getContent());
            replaceOp.setReplacingVariable(
                subtree.getRoot().getOperator().getOutputColumnName());
            // make the subtree a branch of the insertOp.
            replaceNode.addChild(subtree.getRoot());
            subtree.getRoot().setParent(replaceNode);
        }
    }

    replaceNode.setOperator(replaceOp);

    XATDag result = new XATDagImp();
    result.setRoot(replaceNode);

    return result;
}
/**
 * Visit sequence expression will just link all the expressions together.
 *
 * @return XATDag
 * @param seq xacute.quilt.SequenceExpression
 * @since 1.0
 */
public Object visit(SequenceExpression seq) {
    level++;
    if (isDebug())
        System.out.println(indent() + "SequenceExpression");
    // prepare the result.
    XATDag result = new XATDagImp();

    // iterate through the list of the contents and create the trees.
    QuiltExpression[] contents = seq.getContent();
    // process the content.
    Vector con = processContent(contents);
    // get the cols and xattree.
    Vector cols = (Vector) con.elementAt(0);
    XATDag conTree = (XATDag) con.elementAt(1);

    // if there less than two elements in the cols, just return the conTree.
    if (cols.size() < 2) {
        level--;
        return conTree;
    } else { // otherwise create the XMLUnion operator.
        DagNode xunionNode = new DagNodeImp();
        XMLUnion xunionOp = new XMLUnionImp(xunionNode);

        // set the output column name.
        numNodes++;
        xunionOp.setOutputColumnName("$col" + numNodes);

        // put the list of the columns/text/double in it.
        for (int i = 0; i < cols.size(); i++) {
            // we assume only the column name in it.
            // in the future for more general case, we need to update the xunionOp to handle
            // more than only ColumnNames.
            xunionOp.addColumnName(cols.elementAt(i).toString());
        }

        // put on top of the conTree.
        if (conTree.getRoot() != null) {
            xunionNode.addChild(conTree.getRoot());
            conTree.getRoot().setParent(xunionNode);
        }

        // set the root of the result.
        result.setRoot(xunionNode);
    }

    level--;
    return result;
}
/**
 * Called when a sortyby command is called in a query
 *
 * @return XATDag.
 * @param sort SortClause
 * @since 1.0
 */
public Object visit(SortClause sort) {
    level++;

    if (isDebug())
        System.out.println(indent() + "SortClause");

    XATDag result = (XATDag) visit(sort.getE());

    level--;
    return result;
}
/**
 * The tag name data is found in the TagNameable object
 * A tag node is created for the taggerTree (Tag node data structure)
 * 
 * @return String.
 * @param tag TagNameable
 * @since 1.0
 */
public Object visit(TagNameable tag) {
    level++;
    if (isDebug())
        System.out.println(indent() + "TagNameable");

    String result = tag.getTagName();

    level--;

    return result;
}
/**
 * It will return a Vector of 1) expression and 2) subtree of required operators.
 *
 * @return Vector of UniMinusExpression and subXATTree.
 * @param andOp xacute.quilt.UniOpMinus
 * @since 1.0
 */
public Object visit(UnOpMinus minusOp) {
    level++;
    if (isDebug())
        System.out.println(indent() + "UnOpMinus");

    //prepare the Subtree.
    XATDag subtree = new XATDagImp();

    //initialize a new BinAndExpression
    // handle the result, split the expression and also the subtree parts.
    UniMinusExpression minusExp =
        new UniMinusExpressionImp(handleExpression(visit(minusOp.getE()), subtree));

    // prepare the result vector.
    Vector result = new Vector();
    result.addElement(minusExp);
    result.addElement(subtree);

    level--;
    return result;
}
/**
 * It will return a Vector of 1) expression and 2) subtree of required operators.
 *
 * @return Vector of UniNotExpression and subXATTree.
 * @param andOp xacute.quilt.UniOpNot
 * @since 1.0
 */
public Object visit(UnOpNot notOp) {
    level++;
    if (isDebug())
        System.out.println(indent() + "UnOpNot");

    //prepare the Subtree.
    XATDag subtree = new XATDagImp();

    //initialize a new BinAndExpression
    // handle the result, split the expression and also the subtree parts.
    UniNotExpression notExp =
        new UniNotExpressionImp(handleExpression(visit(notOp.getE()), subtree));

    // prepare the result vector.
    Vector result = new Vector();
    result.addElement(notExp);
    result.addElement(subtree);

    level--;
    return result;
}
/**
 * Visit a Boolean value
 *
 * @return Boolean
 * @param valBool xacute.quilt.ValueBool
 * @since 1.0
 */
public Object visit(ValueBool valBool) {
    level++;
    if (isDebug())
        System.out.println(indent() + "ValueBool");

    Boolean result = new Boolean(valBool.getValue());

    level--;
    return result;
}
/**
 * Visit a Boolean value
 *
 * @return Double
 * @param valNum xacute.quilt.ValueNum
 * @since 1.0
 */
public Object visit(ValueNum valNum) {
    level++;
    if (isDebug())
        System.out.println(indent() + "ValueNum");

    Double result = new Double(valNum.getValue());

    level--;
    return result;
}
/**
 * Returns the ValueString object as a string
 * Used for source nodes and select nodes.
 * Also sets the source node in the appropriate tree.
 *
 * @return String.
 * @param val xacute.quilt.ValueString
 * @since 1.0
 */
public Object visit(ValueString val) {
    level++;
    if (isDebug())
        System.out.println(indent() + "ValueString ");

    String result = val.returnString();

    level--;
    return result;
}
/**
 * Visits a binding variable
 * Contains the entry point for a navigation node
 *
 * @return VarStep.
 * @param var Var
 * @since 1.0
 */
public Object visit(Var var) {
    level++;
    if (isDebug())
        System.out.println(indent() + "VAR:" + var.getTagName());

    VarStep varNav = new VarStepImp(var.getTagName());

    level--;
    return varNav;
}
/**
* Called by the FLWR and FLWU.
* 
* @param Upper is the upper XAT above the FOR/LET bindings.
* @param Lower is the lower XAT below the FOR/LET bindings.
* @since 1.0
*/
private void visitBindings(Binding[] bindings, XATDag upper, XATDag lower) {

    // need the split the binding trees into letbindings and forbindings.
    Vector forbindings = new Vector();
    Vector letbindings = new Vector();

    for (int i = 0; i < bindings.length; i++) {
        if (bindings[i] instanceof ForBinding)
            forbindings.addElement(bindings[i]);
        else
            letbindings.addElement(bindings[i]);
    }

    // create the let binding subtree.
    XATDag letTree = new XATDagImp();

	String[] letBindings = new String[letbindings.size()];

    // so for the letbinding has a source operator, will link to the productNode.
    // for the let binding without a source will link to the letTree.
    for (int i = letbindings.size() - 1; i >= 0; i--) {
        // get the let binding XAT.
        XATDag letBindTree = (XATDag) visit(letbindings.elementAt(i));
        // merge it.
        //letTree = mergeTreeWithSources(letTree, letBindTree);
        letBindings[0] = "$"+ ((LetBinding)letbindings.elementAt(i)).getName();
		DagPushdownLET pushdownLET = new DagPushdownLET(upper,lower,letBindTree,letBindings);
		lower = pushdownLET.getMergeTree();
    }

    // well, so until now, the letTree is created properly.
    // the letTree should be append to the bottom of the updateclause.
    //lower = mergeTreeWithSources(lower, letTree);
	//PushdownLET pushdownLET = new PushdownLET(upper,lower,letTree,letBindings);
	//lower = pushdownLET.getMergeTree();
	
    // now let's finish the missing part, which is the forbindings.
    // make sure we have the for bindings
    if (forbindings.size() > 0) {
        // ** create the tree from the inside out. **

        // create the rest of the For operators inside out, besides the first one.
        XATDag lastXtre = lower;
        for (int i = forbindings.size() - 1; i > 0; i--) {
            // create a new for binding tree.
            XATDag forTree = new XATDagImp();

            // get the current binding.
            ForBinding bind = (ForBinding) forbindings.elementAt(i);

            // create the For operator.
            DagNode forNode = new DagNodeImp();
            FOR forOp = builder.newFor(forNode);
            forNode.setOperator(forOp);

            // set the bind variable name.
            forOp.setOutputColumnName("$" + bind.getName());
            forOp.addColumnName("$" + bind.getName());

            // make it the root of the forTree.
            forTree.setRoot(forNode);
            XATDag forBinding = (XATDag) visit(bind);
            if (forBinding.getRoot() != null) {
                forNode.addChild(forBinding.getRoot());
                forBinding.getRoot().setParent(forNode);
            }

            // set internal tree.
            forOp.setXtreFunctions(lastXtre);

            // update the lastXtre's root node to have parent as the ForNode.
            //lastXtre.getRoot().setParent(forNode);
			((DagNode)lastXtre.getRoot()).deleteAllParents();
            // update the reference.
            lastXtre = forTree;
        }

        // create the first For operator.
        DagNode forNode = new DagNodeImp();
        FOR forOp = builder.newFor(forNode);
        forNode.setOperator(forOp);
        // get the binding.
        ForBinding bind = (ForBinding) forbindings.elementAt(0);
        // set the output column name.
        forOp.setOutputColumnName("$" + bind.getName());
        forOp.addColumnName("$" + bind.getName());

        // set the internal tree.
        forOp.setXtreFunctions(lastXtre);
        // update the lastXtre's root node to have parent as the ForNode.
        //lastXtre.getRoot().setParent(forNode);
        // add the for binding blow the FOR node.
        XATDag forBinding = (XATDag) visit(bind);
        // add the forBinding below the forNode.
        XATDag forTree = new XATDagImp();
        forTree.setRoot(forNode);
        forTree = mergeTreeWithSources(forTree, forBinding);
        // add the forNode to the upper
        upper = mergeTreeWithSources(upper, forTree);

    } else { // missing the for binding, means only the update and the let clause.
        // need to connect the let binding with the upper.
        upper = mergeTreeWithSources(upper, lower);
    }

}
}
