package edu.wpi.cs.dsrg.xmldb.xat.component.xatrewrite.common;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.*;

import edu.wpi.cs.dsrg.utils.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;

/**
 * <pre>
 * ReWriteTraversal is the traversal strategy used in the rewrite steps
 * It is responsible for iterating over an XAT algebra tree and finding 2 nodes
 * to evaluate equivalence rules on.  
 *
 * The class is instantiated by passing in an XATTree to traverse, a ReWriteRules object that
 * will be used to evaluate the rules and a Vector containing names (strings) 
 * of XATOperators (fully quantified) to rewrite. doPushDown is then called with a direction
 * to rewrite rules.  
 * 
 * The types of XATOperators to evaluate are determined before starting the traversal.  A Vector
 *  is passed into the constructor
 * This reWriteOrder vector also determines the order of operators to evaluate.
 *
 * The traversal starts by iterating over the tree and finding all Source operators. Then, starting with
 * the first operator, travels up the tree until the first XATOperator (x1) whose type corresponds to the 
 * first class name in the reWriteOrder Vector is found. Operators in their final
 * position are not used.  Next,depending on the direction the user
 * wishes to rewrite rules, the system will either take the parent or child of x1.  A call is then
 * made to ReWriteRules.evaluateRules passing in the 2 operators.  
 *
 * Once the rule is evaluted for 2 nodes, the ReWriteResult object is interpreted and 
 * If the operator is in its finalPosition (from the result object), the hashtable that keeps track
 * of which operators have completed evaluation is updated and the next Operator is found.  If the
 * operator is not in its final position, the new child or parent (depending on direction) is used
 * to evalute another rule.
 *
 * The process continues until all operators are in their final position.  
 * </pre>
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 * @see ReWriteResult
 * @see XATNode
 * @see XATOperator
 * @see XATTree
 */

public class ReWriteTraversal extends XATQueryObjectImp {
	/** 
	 * The XATTree to rewrite
	 * @since 1.0
	 */
	protected XATTree reWriteTree;
	/**
	 * <pre>
	 * The HashTabe stores the list of operators in the tree and whether they
	 * are in their final position or not. 
	 * The key is the operator, the value is a Boolean that represents whether the operator
	 * is in its final position.  
	 * </pre>
	 * @since 1.0	 
	 */
	protected Hashtable reWriteList;
	/** 
	 * Contains the equivalence rules
	 * @since 1.0
	 */
	protected ReWriteRules reWriteRule;
	/** 
	 * List of Strings representing the order and name of XATOperators to 
	 * rewrite.
	 * @since 1.0
	 */
	protected Vector reWriteOrder;
	/** 
	 * Constant to rewrite rules downward.
	 * @since 1.0
	 */
	public final static int DOWN = 1;
	/** 
	 * Constant to rewrite rules upward.
	 * @since 1.0
	 */
	public final static int UP = 0;
	/**
	 * Out put the tree after each rewrtie step or not.
	 * @since 1.0
	 */
	protected boolean isStepThrough;
	/** 
	 * The tree in current rewrite step.
	 * @since 1.0
	 */

	protected java.util.Vector stepTrees;
	/** 
	 * A hashtable to keep track of the parents for each leaf node returned 
	 * by the getAllSourceNodes().
	 * @since 1.0
	 */
	Hashtable nodeAndTree;
	/** 
	 * Indicate if the XAT is changed or not.
	 * @since 1.0
	 */
	private boolean isChanged;
	/**
	 * ReWriteTraversal constructor.
	 * @param tree	The tree to rewrite.
	 * @param rule 	The class contains equivalance rules.
	 * @param order The vector contains strings of the types of operators to optimize.
	 * @since 1.0
	 * @see ReWriteRules
	 * @see XATTree
	 */
	public ReWriteTraversal(XATTree tree, ReWriteRules rule, Vector order) {

		reWriteTree = tree;
		reWriteRule = rule;
		reWriteOrder = order;
		isStepThrough = false;
		reWriteList = new Hashtable();
		nodeAndTree = new Hashtable();
		isChanged = false;

	}
	/**
	 * ReWriteTraversal constructor.
	 * This method is specificly used in GUI.
	 * @param tree	The tree to rewrite.
	 * @param rule 	The class contains equivalance rules.
	 * @param order The vector contains strings of the types of operators to optimize.
	 * @param step  Identify if the ShowTreeCheckBox is selected or not.
	 * @since 1.0
	 * @see ReWriteRules
	 * @see XATTree
	 */

	public ReWriteTraversal(
		XATTree tree,
		ReWriteRules rule,
		Vector order,
		boolean step) {

		reWriteTree = tree;
		reWriteRule = rule;
		reWriteOrder = order;
		isStepThrough = step;
		reWriteList = new Hashtable();
		stepTrees = new Vector();
		nodeAndTree = new Hashtable();
		isChanged = false;
	}
	/**
	 * <pre>
	 * This method uses the XATtree passed into the constructor, and 
	 * traverses over the tree and finds 2 nodes to try to swap.  
	 * The types and order of operators to pushDown are defined in the 
	 * constructor. 
	 * </pre>
	 * @return XATTree The result XAT tree after pushDown.
	 * @param direction The direction to push nodes.
	 * @since 1.0
	 * @see XATTree
	 * @see ReWriteResult
	 * @see XATOperator
	 */
	public XATTree doPushdown() {
		Vector sourceSet = new Vector();
		Vector reWriteOrder;
		Vector childSet = new Vector();
		boolean pushedDown = false;
		ReWriteResult result = new ReWriteResult();
		String operatorToPush;

		XATNode currentNode;
		XATOperator currentOp;

		Debugger d1 = new Debugger(reWriteTree);

		// Initially the tree is unchanged, isChanged = false
		// if nodes are swapped, or deleted, isChanged = true
		setIsChanged(false);

		//set the debugging options for the reWriteRule class
		reWriteRule.setDebug(this.isDebug());

		//find all of the source nodes for the mapping query	
		getAllSourceNodes(sourceSet, reWriteTree.getRoot(), reWriteTree);
		if (isDebug()) {
			System.out.println(sourceSet.size() + " leaf nodes found!");
		}

		for (int i = 0; i < this.reWriteOrder.size(); i++) {

			//get the name of the operator to pushDown
			operatorToPush = (String) this.reWriteOrder.elementAt(i);
			if (isDebug()) {
				System.out.println(
					"----------Now Pushing down "
						+ operatorToPush.toString()
						+ "--------");
			}

			currentNode = new XATNodeImp();

			for (int x = sourceSet.size() - 1; x >= 0; x--) {
				//get the next source node found
				XATNode curSourceNode = (XATNode) sourceSet.elementAt(x);
				//get the tree corresponding to this source set.
				XATTree curTree = (XATTree) nodeAndTree.get(curSourceNode);
				//find the next operator of type operatorToPush that has not been pushed down
				currentNode =
					findNextNode(curTree, curSourceNode, operatorToPush);

				while (currentNode != null) {
					do {
						/*
						*	evalute the equivalence rules and perform the swap if possible	
						*/
						try {
							String timeBefore =
								"" + (System.currentTimeMillis());
							if (statsPresent) {
								/** update the number of times rules applied */
								stats.updateValue(
									"NUMBER_OF_TIMES_RULES_APPLIED",
									this);
							}
							XATNode parent = currentNode;
							XATNode child = null;

							//if we want to pushUp, then we evaluate the currentNode with its parent
							//if we want to pushDown, then we evaluate the currentNode with its child
							try {
								childSet = getChildren(parent);
								if (childSet.size() != 1) {
									System.out.println(
										"Error: Non-linear tree is not rewritable for now.");
									break;
								}
								child = (XATNode) childSet.elementAt(0);
								result =
									(ReWriteResult) reWriteRule.evaluateRules(
										parent,
										child);
							} catch (ArrayIndexOutOfBoundsException e) {
							}
							if (statsPresent) {
								String timeAfter =
									"" + (System.currentTimeMillis());
								String timeRun =
									""
										+ (Long.parseLong(timeAfter)
											- Long.parseLong(timeBefore));
								stats.updateValue(
									"TIME_SPENT_EVALUATING_RULE",
									this,
									timeRun);
								stats.updateValue(
									"TIME_SPENT_EVALUATING_RULES_TOTAL",
									this);
							}
							//if the nodes were swapped, then we need to make sure look at the new node
							if (result.isSwapped()) {
								//currentNode = new XATNodeImp();
								setIsChanged(true);
								XATNode temp = parent;
								parent = child;
								child = temp;
								//then update the childset
								//getChildren(childSet, currentNode);
							}
							if (result.isFinalPosition()) {
								setOperatorPushedDown(parent.getOperator(), true);
							}
							//if the node parent should be deleted, then do so
							if (result.isDeleteParentNode()) {
								//have a problem when the node to delete is the root
								//XIN: problem fixed.
								setIsChanged(true);
								curTree.deleteNode(parent);
								if (isDebug()) {
									System.out.println("Deleted " + parent.toString());
								}
							}
							//if the child node should be deleted, then do so
							if (result.isDeleteChildNode()) {
								setIsChanged(true);
								curTree.deleteNode(child);
								if (isDebug()) {
									System.out.println("Deleted " + child);
								}
							}

							//if the currentNode was moved, then dont change it
							if (result.isMoved()) {
							}
							//if the currentNode was moved, then dont change it
							if (result.isReplaced()) {
								setIsChanged(true);
							}
							//update currentNode
							currentNode = parent;
						} catch (Exception e) {
							//System.out.println(e.getMessage());
							//no rules found so set the node to be pushedDown
							try {
								setOperatorPushedDown(currentNode.getOperator(),true);
							} catch (Exception e2) {
								System.out.println(e2.getMessage());
							}
						}
						// validate the curTree, make sure the p-c relationship consistent.
						// curTree.validate();
					} while (!isOperatorPushedDown(currentNode.getOperator()));
					// find up further.
					currentNode = findNextNode(curTree, curSourceNode, operatorToPush);
				}
			}

			//before returning the tree, now look for Taggers and Aggregates we can delete
			//System.out.println("Cleaning up trees");
			//cleanUpTree();
			reWriteList = new Hashtable();
			//if the user wishes to step through the rewrite process, this add a clone of the tree to the stepTrees vector
			if (getIsStepThrough()) {
				stepTrees.addElement(((XATTree) reWriteTree).clone());
			}
			if (isDebug()) {
				d1.printTree();
			}
		}
		return reWriteTree;
	}

	public XATTree doPushup() {
		//currently we did not have pushup rewrite rules. this code is not tested yet.
		Vector sourceSet = new Vector();
		Vector reWriteOrder;
		Vector childSet = new Vector();
		boolean pushedDown = false;
		ReWriteResult result = new ReWriteResult();
		String operatorToPush;

		XATNode currentNode;
		XATOperator currentOp;

		int y;
		Debugger d1 = new Debugger(reWriteTree);

		// Initially the tree is unchanged, isChanged = false
		// if nodes are swapped, or deleted, isChanged = true
		setIsChanged(false);

		//set the debugging options for the reWriteRule class
		reWriteRule.setDebug(this.isDebug());

		//find all of the source nodes for the mapping query	
		getAllSourceNodes(sourceSet, reWriteTree.getRoot(), reWriteTree);
		if (isDebug()) {
			System.out.println(sourceSet.size() + " leaf nodes found!");
		}

		for (int i = 0; i < this.reWriteOrder.size(); i++) {

			//get the name of the operator to pushDown
			operatorToPush = (String) this.reWriteOrder.elementAt(i);
			if (isDebug()) {
				System.out.println(
					"----------Now Pushing down "
						+ operatorToPush.toString()
						+ "--------");
			}

			currentNode = new XATNodeImp();

			for (int x = sourceSet.size() - 1; x >= 0; x--) {
				//get the next source node found
				XATNode curSourceNode = (XATNode) sourceSet.elementAt(x);
				//get the tree corresponding to this source set.
				XATTree curTree = (XATTree) nodeAndTree.get(curSourceNode);
				//find the next operator of type operatorToPush that has not been pushed down
				currentNode =
					findNextNode(curTree, curSourceNode, operatorToPush);

				while (currentNode != null) {
					do {
						/*
						*	evalute the equivalence rules and perform the swap if possible	
						*/
						try {
							String timeBefore =
								"" + (System.currentTimeMillis());
							if (statsPresent) {
								/** update the number of times rules applied */
								stats.updateValue(
									"NUMBER_OF_TIMES_RULES_APPLIED",
									this);
							}
							XATNode parent = null;
							XATNode child = null;

							//if we want to pushUp, then we evaluate the currentNode with its parent
							parent = currentNode.getParent();
							if (parent == null) {
								System.out.println(
									"Error: Root node can not be push up!");
								break;
							}
							result =
								(ReWriteResult) reWriteRule.evaluateRules(
									parent,
									currentNode);

							if (statsPresent) {
								String timeAfter =
									"" + (System.currentTimeMillis());
								String timeRun =
									""
										+ (Long.parseLong(timeAfter)
											- Long.parseLong(timeBefore));
								stats.updateValue(
									"TIME_SPENT_EVALUATING_RULE",
									this,
									timeRun);
								stats.updateValue(
									"TIME_SPENT_EVALUATING_RULES_TOTAL",
									this);
							}
							//if the nodes were swapped, then we need to make sure look at the new node
							if (result.isSwapped()) {
								//currentNode = new XATNodeImp();
								setIsChanged(true);
								currentNode = parent;
							}
							if (result.isFinalPosition()) {
								setOperatorPushedDown(
									currentNode.getOperator(),
									true);
							}
							//if the node parent should be deleted, then do so
							if (result.isDeleteParentNode()) {
								//have a problem when the node to delete is the root
								setIsChanged(true);
								curTree.deleteNode(parent);
								if (isDebug()) {
									System.out.println(
										"Deleted " + parent.toString());
								}
							}
							//if the child node should be deleted, then do so
							if (result.isDeleteChildNode()) {
								setIsChanged(true);
								curTree.deleteNode(child);
								if (isDebug()) {
									System.out.println("Deleted " + child);
								}
							}

							//if the currentNode was moved, then dont change it
							if (result.isMoved()) {
							}
							//update currentNode
							currentNode = parent;

						} catch (Exception e) {
							//System.out.println(e.getMessage());
							//no rules found so set the node to be pushedDown
							try {
								setOperatorPushedDown(
									currentNode.getOperator(),
									true);
							} catch (Exception e2) {
								System.out.println(e2.getMessage());
							}
						}
						// validate the curTree, make sure the p-c relationship consistent.
						// curTree.validate();
					} while (!isOperatorPushedDown(currentNode.getOperator()));
					// find up further.
					currentNode =
						findNextNode(curTree, curSourceNode, operatorToPush);
				}
			}

			//before returning the tree, now look for Taggers and Aggregates we can delete
			//System.out.println("Cleaning up trees");
			//cleanUpTree();
			reWriteList = new Hashtable();
			//if the user wishes to step through the rewrite process, this add a clone of the tree to the stepTrees vector
			if (getIsStepThrough()) {
				stepTrees.addElement(((XATTree) reWriteTree).clone());
			}
			if (isDebug()) {
				d1.printTree();
			}
		}
		return reWriteTree;
	}

	/**
	 * This method repeatly call doPushdown until there is no more changes in the Pushdown.
	 * Assume the this.rewriteTree never used again.
	 * @return The XATTree after rewrite.
	 * @param direction The direction of rewrite.
	 * @since 1.0
	 * @see XATTree
	 */

	public XATTree doRewrite(int direction) {
		if (direction == ReWriteTraversal.DOWN)
			do {
				this.reWriteTree = doPushdown();
			} while (getIsChanged());
		else {
			do {
				this.reWriteTree = doPushup();
			} while (getIsChanged());
		}
		return this.reWriteTree;
	}
	/**
	 * Find next node need to be pushed down.
	 * <pre>
	 * This method will traverse the tree, starting from paramNode and 
	 * ending as it finds a node of type that has not been pushedDown. 
	 * This node is then returned. If no such node is found, it returns null.
	 * </pre>
	 * @return XATNode The first match that has not been pushed down.
	 * @param paramNode The node to start looking from.
	 * @param toMatch The type of operator to match.
	 * @since 1.0
	 * @see XATNode
	 */
	protected XATNode findNextNode(
		XATTree tree,
		XATNode paramNode,
		String toMatch) {
		XATNode parentNode = paramNode;
		String parentClass;

		while (parentNode != tree.getRoot()) { // while it's not the root.
			//if the parent is not null 
			if (parentNode != null) {
				//make sure its a navigation node we are looking at
				parentClass = parentNode.getOperator().getClass().getName();
				if (parentClass.equals(toMatch)) {
					//and the node is not pushed down
					if (isOperatorPushedDown(parentNode.getOperator())
						== false) {
						return parentNode;
					}
				}
				//make parentNode = its parent to recursively climb the tree
				parentNode = parentNode.getParent();
			}
			//the parent was null
			else {
				return null;
			}
		}
		parentClass = parentNode.getOperator().getClass().getName();
		if (parentClass.equals(toMatch)) {
			//and the node is not pushed down
			if (isOperatorPushedDown(parentNode.getOperator()) == false) {
				return parentNode;
			}
		}
		return null;
	}
	/**
	 * <pre>
	 * Accepts an XATNode paramNode and an XATOperator. 
	 * The tree is traversed starting at paramNode and ends when it finds 
	 * a node of type type that has not been pushedDown.  This node is then returned
	 * If no such node is found, it returns null
	 * </pre>
	 * @return XATNode The first match that has not been pushed down.
	 * @param paramNode The node to start looking from.
	 * @param toMatch The type of operator to match.
	 * @since 1.0
	 * @see XATNode
	 */
	protected XATNode findNextNode(XATNode paramNode, String toMatch) {

		XATNode parentNode = paramNode;
		String parentClass;

		// if should only reach the root of this Node.

		while (true) {
			//if the parent is not null
			if (parentNode != null) {
				//make sure its a navigation node we are looking at
				parentClass = parentNode.getOperator().getClass().getName();
				if (parentClass.equals(toMatch)) {
					//and the node is not pushed down
					if (isOperatorPushedDown(parentNode.getOperator())
						== false) {
						return parentNode;
					}

				}
				//make parentNode = its parent to recursively climb the tree
				parentNode = parentNode.getParent();

			}
			//the parent was null
			else {

				return null;

			}
		}

	}
	/**
	 * Find all of the Source operators in an XATTree.
	 * <pre>
	 * This method first look at all of the root's children and adds them to the vector
	 * if that nodes' operator is an XML source .  The function is then
	 * recurvsively called on each of root's children until there are no more nodes to look at.  
	 * </pre>
	 * @return void 
	 * @param sourceSet stores all of the Source operators found  
	 * @param root The root of a tree to iterate over
	 * @since 1.0
	 * @see XATTree
	 * @see XATNode
	 */
	protected void getAllSourceNodes(
		Vector sourceSet,
		XATNode root,
		XATTree tree) {
		Vector childrenVector = getChildren(root);
		;

		XATNode currentNode;

		if (root.getOperator() instanceof FOR) {
			// recursively call this function for the subtree of the FOR operator.
			getAllSourceNodes(
				sourceSet,
				((FOR) root.getOperator()).getXtreFunctions().getRoot(),
				((FOR) root.getOperator()).getXtreFunctions());
		}

		//loop recursively until there are no children of a node
		if (childrenVector.size() == 0) {
			//XIN: add the current node.
			sourceSet.addElement(root);
			nodeAndTree.put(root, tree);
			return;

		} else {
			for (int x = 0; x < childrenVector.size(); x++) {
				currentNode = (XATNode) childrenVector.elementAt(x);
				//determine the current node's operator's class type
				//if (currentNode.getOperator() instanceof Source) {
				//	SourceImp s1 = (SourceImp) currentNode.getOperator();

				//add the source node to the source set
				//sourceSet.addElement(currentNode);

				//}else
				//recursively call this function
				getAllSourceNodes(sourceSet, currentNode, tree);
			}
		}
		return;
	}
	/**
	 * getChildren accepts a destination Vector and an XATNode
	 * it then puts all of the root's children into this vector
	 * @return void
	 * @param destinationVector 
	 * @param root
	 * @since 1.0
	 * @see XATNode
	 */
	protected Vector getChildren(XATNode root) {

		XATNode[] childrenArray = root.getChildren();
		//empty the destinationVector to ensure that only the root's children are in the Vector
		Vector childrenVector = new Vector();
		for (int x = 0; x < childrenArray.length; x++) {
			childrenVector.addElement((XATNode) childrenArray[x]);
		}
		return childrenVector;
	}
	/**
	 * Get the tree in current rewrite step.
	 * @return java.util.Vector
	 * @since 1.0
	 */
	public java.util.Vector getStepTrees() {
		return stepTrees;
	}
	/**
	 * Get the value of the propertity 'DEBUG_ReWriteTraversal'.
	 * @return boolean True: With debug information. False: No debug information.
	 * @since 1.0
	 */
	public boolean isDebug() {
		return System.getProperty("DEBUG_ReWriteTraversal") != null;
	}
	/**
	 * Get the value of the propertity 'isChanged'.
	 * @return boolean True: The tree has been changed. False: No change at all.
	 * @since 1.0
	 */
	public boolean getIsChanged() {
		return isChanged;
	}
	/**
	 * Check if an operator need any push down or not.
	 * This method will look at the reWriteList and 
	 * return whether the operator is in its final position of not.
	 * <pre>
	 * If the operator does not appear in the list, the operator is added
	 * and the value is set to false
	 * </pre>
	 * @return boolean True: if the operator is in its final position. False: otherwise.
	 * @param toCheck The XATOperator to look up in the reWriteList.
	 * @since 1.0
	 * @see XATOperator
	 */
	protected boolean isOperatorPushedDown(XATOperator op) {

		//if the param node is contained within the hashtable
		if (reWriteList.containsKey(op)) {
			//then returns its value
			return ((Boolean) reWriteList.get(op)).booleanValue();
		}
		//if the element does not exist in the hashtable, then add it with a false value
		else {
			Boolean b1 = new Boolean(false);
			reWriteList.put(op, b1);
			return b1.booleanValue();
		}
	}
	/**
	 * Get the value of the propertity 'StepThrough'.
	 * @return boolean True: Out put the tree after each rewrtie step. False: No debug information.
	 * @since 1.0
	 */
	public boolean getIsStepThrough() {
		return isStepThrough;
	}
	/**
	 * I temporarily keep this method.
	 * But it should be deprecated!
	 * Using System.getProperty("DEBUG_ReWriteTraversal") instead.
	 * Creation date: (8/23/2002 4:56:17 PM)
	 * @param newDebug boolean
	 */
	public void setDebug(boolean newDebug) {
		if (newDebug)
			System.setProperty("DEBUG_ReWriteTraversal", "OK");
		// there is no way to unset this property.
	}
	/**
	 * This method will be called by the constructor. It will set the default properties
	 * that the ALL query objects will support. Each subclass should override
	 * this method and add their own unique properties, but should also call
	 * super.setDefaultProperties() so that the common properties may be set.
	 * @return void
	 * @since 1.0
	 */
	protected void setDefaultProperties() {
		super.setDefaultProperties();
		addProperty("NUMBER_OF_TIMES_RULES_APPLIED", "0");
		addProperty("TIME_SPENT_EVALUATING_RULE", "0");
		addProperty("TIME_SPENT_EVALUATING_RULES_TOTAL", "0");
	}
	/**
	 * Get the value of the propertity 'isChanged'.
	 * @return boolean True: The tree has been changed. False: No change at all.
	 * @param newIsChanged True: The tree is changed. False: otherwise.
	 * @since 1.0
	 */

	public void setIsChanged(boolean newIsChanged) {
		isChanged = newIsChanged;
	}
	/**
	 * Modifies this operators value in the reWriteList according to the
	 * val parameter.  This is used after a rule has been evaluated and the result
	 * specifies that the operator is now in its final position
	 * @return void
	 * @param toSet The XATOperator to set final position.
	 * @param val True: if the operator is now in its final position. False: otherwise.
	 * @since 1.0
	 */
	protected void setOperatorPushedDown(
		XATOperator op,
		boolean isPushedDown) {
		reWriteList.put(op, new Boolean(isPushedDown));
	}
	/**
	 * Get the tree in current rewrite step.
	 * @return void
	 * @param newStepTrees java.util.Vector
	 * @since 1.0
	 */

	public void setStepTrees(java.util.Vector newStepTrees) {
		stepTrees = newStepTrees;
	}
}
