package edu.wpi.cs.dsrg.xmldb.xat.component.xatrewrite.common.dag;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.*;
import java.lang.reflect.*;

import edu.wpi.cs.dsrg.utils.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.*;

/**
 * This class is the super class for all equivalence rule classes. 
 * <pre>
 * An object of this class is created with a Vector of Strings containing the order
 * of classes to look at when trying to find a rule for 2 operators. 
 *
 * Calls are then made to evaluateRules with 2 nodes that the user wishes to evalute. 
 * The TwoParamVisitor class from the visitorpattern package is used to determine the correct 
 * equivalence rule to use for a given pair of XATOperators. This is done by iterating over 
 * the classes provided in the classVector and calling the visit(Object, Object) method of each class. 
 * 
 * If a rule is not found in a class, the next class is looked at.  If no rules are found for a given
 * pair, an ReWriteRulesNoRulesFound exception is thrown.
 * 
 * Equivalence rules are defined in subclasses named after the parent operator.  ReWriteNavigationRules
 * defines the equivalence rules for Navigate Operators.  ReWriteOrderByRules defines the
 * equivalence rules for OrderBy Operators and so on. 
 * 
 * A rule is a method in the strucure of visit(XATOperator parent, XATOperator child).  The rule
 * for swapping a Navigate with a Source is defined in ReWriteNavigationRules.visit(Navigate, Source). 
 * Each rule will perform modififactions neccessary for the swap of the two operators if such a swap is
 * possible.  If the nodes can be swapped, the rule will do so. 
 * See each rule for the logic behind that rule.  Each rule returns a ReWriteResult object. 
 * This object will tell the calling program if the Operators were be swapped, is the parent in its
 * final position, and should either node be deleted.  
 * </pre>
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 * @see  ReWriteResult
 */

public class ReWriteRules extends TwoParamVisitor {

	/** 
	 * ReWriteResult that describes the results from performing the Equivalence Rule
	 * @since 1.0
	 */
	protected ReWriteResult result;
	/** 
	 * Vector that contains the list and order of rule classes to look at.
	 * @since 1.0
	 */
	protected Vector classVector;
	/** 
	 * BindingTable that is used to store the variables and the associated columnPaths
	 * @since 1.0
	 */
	protected BindingTable variableTable;
	/** 
	 * Hashtable to keep the rule statistics of this rewrite rules
	 * In the format of "method obj, Integer count".
	 * This is an object for all the instances of this class.
	 * @since 1.0
	 */
	protected static Hashtable statis;
	/**
	 * Default ReWriteRules constructor.
	 * @since 1.0
	 */
	public ReWriteRules() {

		result = new ReWriteResult();
		classVector = new Vector();
		variableTable = new BindingTable();

	}
	/**
	 * ReWriteRules constructor.
	 * @param sourceList The list of operator types being considered to apply this rewrite rule.
	 * @since 1.0  
	 * @see BindingTable 
	 */

	public ReWriteRules(Vector sourceList) {
		result = new ReWriteResult();
		classVector = sourceList;
		variableTable = new BindingTable();

	}
	/**
	 * Constructor that accepts a Vector sourceList that is set to classVector 
	 * and a Hashtable that is set to variableTable. The result object is
	 * instantiated.  
	 * @param sourceList The list of operator types being considered to apply this rewrite rule.
	 * @param BindingTable The binding table of given XAT.
	 * @since 1.0  
	 * @see BindingTable 
	 */
	public ReWriteRules(Vector sourceList, BindingTable varList) {
		result = new ReWriteResult();
		classVector = sourceList;
		variableTable = varList;
	}
	/**
	 * Convert an array of NavigationStep into String.
	 * This method accepts an array of NavigationSteps and will convert these 
	 * steps into one long string.
	 * If a VarStep is found, it will be replaced by the actual column name(s) as found
	 * in this.variableTable
	 * The array is formatted with "/" between steps.  A "/" is added at the end of the string as well
	 * @return String The result corresponding step string.
	 * @param step NavigationStep from which to create a string.
	 * @since 1.0
	 * @see NavigationStep
	 */

	protected String createString(NavigationStep[] step) {
		String returnString = new String();
		String stepValue = "";
		NavigationStep[] varValue;

		for (int i = 0; i < step.length; i++) {
			if (step[i] instanceof VarStep) {
				varValue =
					getVariableTable().getVariableValue(step[i].getStepValue());
				for (int j = 0; j < varValue.length; j++) {
					stepValue = stepValue + varValue[j].getStepValue() + "/";
				}

			} else {
				stepValue = step[i].getStepValue() + "/";
			}

			returnString = returnString + stepValue;
			stepValue = "";

		}

		return returnString;
	}
	public Object dispatch() {
		return null;
	}
	public Object dispatch(Object object) {
		return null;
	}
	public Object dispatch(Object object1, Object object2) {
		/*
		        System.out.println("In dispatch Object, Object in super class");
		        System.out.println("Object 1 is: " + object1.toString());
		        System.out.println("Object 2 is: " + object2.toString());
		*/
		return null;
	}
	/**
	 * Evaluate the rules over 2 XATOperators.  
	 * <pr>
	 * This method uses java.lang.reflect methods to instantiate the classes
	 * in this.classVector.  The method iterates over the classes, instantiates the class,
	 * then calls the visit function with the operators of parent and child.
	 *
	 * If no rule is found, the next class is instantiated and the process repeats itself.  
	 * the function returns true if the function cannot be pushedDown further
	 * otherwise it returns false
	 * </pr>
	 * @return ReWriteResult An object containing a summary of what was performed by the rule.
	 * @param parent The parent node that we will try to find a rule for.
	 * @throws ReWriteRulesNoRulesFound Exception thrown when the system has looked at
	 * all classes in the classVector and is unable to find a rule for parent and child.
	 * @since 1.0
	 * @see ReWriteResult
	 */
	public ReWriteResult evaluateRules(XATNode parent)
		throws ReWriteRulesNoRulesFound {

		String className;
		Class ruleClass;
		ReWriteRules p1;
		/*
		        System.out.println(
		            "Trying to find a rule for: "
		                + parent.getOperator().toString()
		                + " and "
		                + child.getOperator().toString());
		*/
		className = parent.getOperator().getClass().toString();

		for (int x = 0; x < classVector.size(); x++) {

			try {
				className = (String) classVector.elementAt(x);
				ruleClass = Class.forName(className);
				//create a new instance of the class from the classVector
				p1 = (ReWriteRules) ruleClass.newInstance();
				//set the variableTable of the new rule Class
				p1.setVariableTable(this.variableTable);
				//set the debug value for each new rule Class
				p1.setDebug(this.isDebug());
				//call the visit(Object, Object) of that class
				result = (ReWriteResult) p1.visit(parent.getOperator());
				//return the result if it is not null
				if (result != null) {
					//make the last rule class looked at the first class to make it faster
					String lastClass = (String) classVector.elementAt(x);
					classVector.removeElement(lastClass);
					classVector.insertElementAt(lastClass, 0);
					return result;
				} else {
					//otherwise look at the next class
				}
			} catch (Exception e) {
				System.out.println(e.toString());
			}
		}

		throw new ReWriteRulesNoRulesFound(
			"No rules were found for " + parent.getOperator());

	}
	/**
	 * Evaluate the rules over 2 XATOperators.  
	 * <pr>
	 * This method uses java.lang.reflect methods to instantiate the classes
	 * in this.classVector.  The method iterates over the classes, instantiates the class,
	 * then calls the visit function with the operators of parent and child.
	 *
	 * If no rule is found, the next class is instantiated and the process repeats itself.  
	 * the function returns true if the function cannot be pushedDown further
	 * otherwise it returns false
	 * </pr>
	 * @return ReWriteResult An object containing a summary of what was performed by the rule.
	 * @param parent The parent node that we will try to find a rule for.
	 * @param child The child node that we will try to find a rule for.
	 * @throws ReWriteRulesNoRulesFound Exception thrown when the system has looked at
	 * all classes in the classVector and is unable to find a rule for parent and child.
	 * @since 1.0
	 * @see ReWriteResult
	 */

	public ReWriteResult evaluateRules(XATNode parent, XATNode child)
		throws ReWriteRulesNoRulesFound {

		String className;
		Class ruleClass;
		ReWriteRules p1;
		/*
		        System.out.println(
		            "Trying to find a rule for: "
		                + parent.getOperator().toString()
		                + " and "
		                + child.getOperator().toString());
		*/
		className = parent.getOperator().getClass().toString();

		for (int x = 0; x < classVector.size(); x++) {

			try {
				className = (String) classVector.elementAt(x);
				ruleClass = Class.forName(className);
				//create a new instance of the class from the classVector
				p1 = (ReWriteRules) ruleClass.newInstance();
				//set the variableTable of the new rule Class
				p1.setVariableTable(this.variableTable);
				//set the debug value for each new rule Class
				p1.setDebug(this.isDebug());
				//call the visit(Object, Object) of that class
				result =
					(ReWriteResult) p1.visit(
						parent.getOperator(),
						child.getOperator());
				//return the result if it is not null
				if (result != null) {
					//make the last rule class looked at the first class to make it faster
					String lastClass = (String) classVector.elementAt(x);
					classVector.removeElement(lastClass);
					classVector.insertElementAt(lastClass, 0);
					return result;
				} else {
					//otherwise look at the next class
				}
			} catch (Exception e) {
				System.out.println(e.toString());
			}
		}

		throw new ReWriteRulesNoRulesFound(
			"No rules were found for "
				+ parent.getOperator()
				+ " and "
				+ child.getOperator());

	}
	/**
	 * Return the statistics of the rewrite rules.
	 * @since 1.0  
	 */

	public static java.util.Hashtable getStatis() {
		return statis;
	}
	/**
	 * Insert the method's description here.
	 * 
	 * @return BindingTable
	 */
	public BindingTable getVariableTable() {
		return variableTable;
	}
	/**
	 * Get the value of the propertity 'DEBUG_NavTaggerCancelout'.
	 * @return boolean True: With debug information. False: No debug information.
	 * @since 1.0
	 */
	public boolean isDebug() {
		return System.getProperty("DEBUG_ReWriteRules") != null;
	}
	/**
	 * A flag to check whether to get the statistics from rewrite rules.
	 * Creation date: (9/16/2002 5:13:23 PM)
	 * @return boolean
	 */
	public boolean isGatherStatistics() {
		return System.getProperty("ReWriteRules_GatherStatistics") != null;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (9/16/2002 5:13:23 PM)
	 * @param newDebug boolean
	 */
	public void setDebug(boolean newDebug) {
		if (newDebug)
			System.setProperty("DEBUG_ReWriteRules", "OK");
	}
	/**
	 * Insert the method's description here.
	 * 
	 * @param newVariableTable java.util.BindingTable
	 */
	public void setVariableTable(BindingTable newVariableTable) {
		variableTable = newVariableTable;
	}
	/** If there is an inheretance between the visited objects, it will only match the highest class. */
	public Object visit(Object object) {
		if (object == null)
			return dispatch();
		Method m = getMethod(object.getClass());
		try {
			if (isGatherStatistics()) {
				if (ReWriteRules.statis == null) {
					ReWriteRules.statis = new Hashtable();
				} else {
					Integer count = (Integer) ReWriteRules.statis.get(m);
					if (count == null) {
						ReWriteRules.statis.put(m, new Integer(1));
					} else {
						ReWriteRules.statis.put(
							m,
							new Integer(count.intValue() + 1));
					}
				}
			}

			return m.invoke(this, new Object[] { object });
		} catch (InvocationTargetException e) {
			e.getTargetException().printStackTrace();
			return dispatch(object);
		} catch (IllegalAccessException e) {
			return dispatch(object);
		} catch (IllegalArgumentException e) {
			return dispatch(object);
		}
	}
	/**
	* We add funtionality to print out the rules used in this rewrite.
	* If there is an inheretance between the visited objects, 
	* it will only match the highest class. 
	*/
	public Object visit(Object obj1, Object obj2) {
		if (obj1 == null && obj2 == null)
			return dispatch();

		if (obj1 == null)
			return visit(obj2);

		if (obj2 == null)
			return visit(obj1);

		Method m = getMethod(obj1.getClass(), obj2.getClass());
		if (m == null) {
			// then get the dispatch object.
			try {
				m =
					getClass().getMethod(
						"dispatch",
						new Class[] { Object.class, Object.class });
			} catch (NoSuchMethodException nsme) {
				System.out.println("Cannot Happend!");
			}
		}
		try {
			if (isGatherStatistics()) {
				if (ReWriteRules.statis == null) {
					ReWriteRules.statis = new Hashtable();
				} else {
					Integer count = (Integer) ReWriteRules.statis.get(m);
					if (count == null) {
						ReWriteRules.statis.put(m, new Integer(1));
					} else {
						ReWriteRules.statis.put(
							m,
							new Integer(count.intValue() + 1));
					}
				}
			}
			return m.invoke(this, new Object[] { obj1, obj2 });
		} catch (InvocationTargetException e) {
			return dispatch(obj1, obj2);
		} catch (IllegalAccessException e) {
			return dispatch(obj1, obj2);
		} catch (IllegalArgumentException e) {
			return dispatch(obj1, obj2);
		}
	}
}
