package edu.wpi.cs.dsrg.xmldb.xat.component.xattreegeneration;
/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//	Rainbow 1.0 - XQuery Processing.
//
//	Copyright 2003, All Rights Reserved.
//
//	Permission to use, copy, modify and distribute this software and
//	its documentation for educational, research and non-profit
//	purposes, without fee, and without a written agreement is hereby
//	granted, provided that this copyright notice appear in all copies.
//
//	Permission to incorporate this software into commercial products
//	may be obtained directly from the authors.
//
//	This software is provided "as is" without express or implied warranty.
//	In no event will the authors be held liable for any damages arising
//	from the use of this software.
//
//	Authors:
//	Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//	Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//	Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//	Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//		Computer Science Department
//		Worcester Polytechnic Institute
//		Worcester, MA 01609 USA
//
//		http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//	Send bug reports, bug fixes, enhancements, requests, or comments to
//	the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import edu.wpi.cs.dsrg.utils.*;
import java.util.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.operatorfactory.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.xatrewrite.common.*;

/**
 * This class visits the XAT Tree constructed in ParserTraverse 
 * and get rid of For operaters. Thus we can get rid of inner loops
 * in the XAT Tree and make a flat algebra tree.
 * <p>
 * To understand this code, one should know how the decorrelation
 * works and what our algorithm is.  It would also be helpful to 
 * know what a visitor pattern is because that is what we use to
 * traverse through the query.
 * <p>
 * Currently, we need to 2 round of decorrelation to get rid of 
 * all For loops.
 * <p>
 * The Main issues in decorrelation is:
 * 1. empty collection problem
 * 2. count bugs.
 *
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 * @see edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern
 */

/**
 * songwang
 *
 * The new decor algorithm consistent with paper.
 */

public class QueryDecorrelation extends InheritableVisitor {

	/**
	* record the context FOR bindings for the groupby operator generation.
	* The Strings will be the for variables.
	* @since 1.0
	*/
	protected String[] contextForBindings;

	/**
	* record the decorrelated XAT Tree.
	* This tree will have less one FOR after calling.
	* @since 1.0
	*/
	protected XATTree decorrTree;

	/**
	* A global point to denote where is the current parent we are working with.
	* 
	* @since 1.0
	*/
	protected XATNode currentParent;
	protected XATNode forNode;
	protected boolean hasPassedReturn;
	protected boolean hasPassed2ndReturn;

	/**
	* used to create operators
	* 
	* @since 1.0
	*/
	protected OperatorBuilder builder;

	protected ConsumedVisitor consumedVisitor;

	protected QueryDecorrelation() {
		super();
	}
	/**
	 * Class Constructor with XAT Tree to be decorrelated.
	 * This is the operating method to be called by outside.
	 * @since 1.0
	 */
	public QueryDecorrelation(XATTree tree) {
		if (tree != null) {
			try {
				/** create the factory */
				OperatorFactory factory = OperatorFactory.newInstance();
				/** instantiate the operator builder */
				builder = factory.newOperatorBuilder();
			} catch (FactoryException e) {
				/** 
				* there was a problem instantiating the factory
				* so we have to stop generating the tree. 
				*/
				System.err.println(e);
			}
			//used to check if the for binding variable is used by a XATNode
			this.consumedVisitor = new ConsumedVisitor();

			decorrTree = (XATTree) tree.clone();

			//start decorrelating
			//having a loop, each time decor one FOR only
			this.currentParent = null;
			this.forNode = null;
			while (hasFor(decorrTree.getRoot(), null)) {
				//step1: decor the subtree of the forNode first
				QueryDecorrelation subtreeDecor =
					new QueryDecorrelation(
						((FOR) this.forNode.getOperator()).getXtreFunctions());
				((FOR) this.forNode.getOperator()).setXtreFunctions(
					subtreeDecor.getDecorTree());
				//step2: decor this forNode, you will never meet another for in the subtree
				decor();
			}
			//recompute the binding table.
			decorrTree.setBindingTable(
				new BindingTableComputer().computeBindingTable(decorrTree));
			//after decorrelation, cleanup the XATTree.
			if (System.getProperty("CLEANXAT") != null)
			XATCleanUp.doCleanUp(decorrTree);
		} else {
			decorrTree = new XATTreeImp();
			decorrTree.setBindingTable(
				new BindingTableComputer().computeBindingTable(decorrTree));
		}
	}

	/**
	 * Find out if the tree rooted at the node has For underneath.
	 * If there is no For Node existing, then return false.
	 * If there is For Node existing, then return true;
	 * At the same time, this.currentParent is set.
	 * @param node XATNode: the root of the xat
	 * @param parent XATNode: the parent of XATNode node.
	 * @since 1.0
	 */
	private boolean hasFor(XATNode node, XATNode parent) {
		if (node.getOperator() instanceof FORImp) {
			this.currentParent = parent;
			this.forNode = node;
			this.hasPassedReturn = false;
			this.hasPassed2ndReturn = false;
			this.contextForBindings =
				((FOR) node.getOperator()).getColumnNames();
			return true;
		}

		XATNode[] children = node.getChildren();

		for (int i = 0; i < children.length; i++) {
			if (hasFor(children[i], node))
				return true;
		}
		return false;
	}

	/**
	 * Completes the decorrelation algorithm on the XAT
	 *
	 * @param parent XATNode: parent of the first FOR in XAT Tree.
	 * @since 1.0
	 */
	private void decor() {
		//1. calculate the stop position of the For variable 
		//in the subtree of FOR. Put stop XATNode into Hashtable
		//If one XATNode consume the binding variable(Navigate, select, gb, ...)
		//Or it's another FOR and the subtree has a XATNode using this for binding variable.
		//It is USING node.
		//If one XATNode is the lowest USING node.
		//It is the stop XATNode.

		//2. Check if the LOJ is needed at this FOR position.
		//If there is select not on this variable till first tag, merge...
		//If there is tag, merge... upper this FOR till root or a LOJ.

		//3. visit the subtree of the FOR XATNode.
		//If the XATNode is stop node...generate CP
		//If the XATNode is context sensitive node...generate GB
		//If the XATNode is other Node ...push down

		//4. recursively call the decor with new parent of for.

		if (!isConsumeFor()) {
			//basic cases:
			//at this point, there is no XATNode in the subtree of the FOR
			//that consuming the binding variable.
			//so if the subtree is not null, make the CP otherwise delete the FOR Node
			if (((FOR) this.forNode.getOperator()).getXtreFunctions().getRoot()
				== null) {
				XATNode childOfFor = this.forNode.getChild(0);
				decorrTree.addNode(childOfFor, this.currentParent);
				this.currentParent.deleteChild(this.forNode);
			} else {
				XATNode childOfFor = this.forNode.getChild(0);
				//create a cartesian product node
				XATNode cpNode = new XATNodeImp();
				CartesianProduct cartProd = builder.newCartesianProduct(cpNode);
				cpNode.setOperator(cartProd);
				//set the subtree of for as a child of the cartesian product
				//The order of the children is essential for the cp. because it will be used by merge.
				decorrTree.addNode(childOfFor, cpNode);
				decorrTree.addNode(
					((FOR) this.forNode.getOperator())
						.getXtreFunctions()
						.getRoot(),
					cpNode);
				// now, add the cartesian product to the tree
				decorrTree.addNode(cpNode, this.currentParent);
				this.currentParent.deleteChild(this.forNode);
			}
		} else {
			//recursive case: according to upper explanation. steps are:
			//step 1:check if the LOJ is needed to be added here.
			//step 2:pushdown the FOR depending on different node.
			//step 3:recursively call decor()
			if (hasSelectBelow(((FOR) this.forNode.getOperator())
				.getXtreFunctions()
				.getRoot())
				&& hasTagAbove()) {
				//create an OuterJoin node
				XATNode outerJoin = new XATNodeImp();
				LeftOuterJoin loj = builder.newLeftOuterJoin(outerJoin, null);
				loj.setOutputColumnName(
					((FOR) this.forNode.getOperator()).getOutputColumnName());
				outerJoin.setOperator(loj);
				//now, insert the outerjoin node to the tree
				decorrTree.addNode(outerJoin, this.currentParent);
				this.currentParent.deleteChild(this.forNode);
				//Then link the magic branche to the left.
				XATTree magicBranch = new XATTreeImp();
				magicBranch.setRoot(this.forNode.getChild(0));
				decorrTree.addNode(
					((XATTree) magicBranch.clone()).getRoot(),
					outerJoin);
				//Then link the original tree to the right.
				decorrTree.addNode(this.forNode, outerJoin);
				this.currentParent = outerJoin;
			}
			//visit root of the subtree of For
			visit(
				(Object) ((FOR) this.forNode.getOperator())
					.getXtreFunctions()
					.getRoot()
					.getOperator());
			//recursively call this function.
			//Note: this.currentParent and this.forNode should be changed
			//in above visit() statement. and decor() is called in visit() also.
		}
	}
	private boolean hasSelectBelow(XATNode root) {
		//If there exist a select operator inside the subtree of for
		//and this select is inside(appear in subtree of) the "return" blockop
		//and this select is above tag or merge or loj
		//we will returned a TRUE and return.
		//This is a recursive function. 

		if (!this.hasPassedReturn)
			return false;

		if (root == null)
			return false;
		if (!this.hasPassed2ndReturn) {
			if (root.getOperator() instanceof TaggerImp) {
				return false;
			}

			if (root.getOperator() instanceof MergeImp) {
				return false;
			}
			//for the nature Join changed from Merge
			if (root.getOperator() instanceof NatureJoinImp) {
				return false;
			}
			//for the count and may also needed for other aggregate functions
			if ((root.getOperator() instanceof FunctionImp)
				&& (((Function) root.getOperator())
					.getName()
					.equalsIgnoreCase("count"))) {
				return false;
			}
		}

		if ((root.getOperator() instanceof SelectImp))
			return true;

		for (int i = 0; i < root.getChildren().length; i++) {
			if (hasSelectBelow(root.getChild(i)))
				return true;
		}

		return false;
	}

	private boolean hasTagAbove() {
		//Node: We should only adding LOJ for those Tagger, Merge and etc
		//that belongs to THIS RETURN clause.
		XATNode parent = this.currentParent;
		boolean backToSelfReturn = false;
		while (parent != null) {
			if ((parent.getOperator() instanceof NameColumnImp)
				&& ("return"
					.equals(
						((NameColumnImp) parent.getOperator())
							.getBlockDesc()))) {
				backToSelfReturn = true;
			}
			if (!this.hasPassed2ndReturn || backToSelfReturn) {
				if (parent.getOperator() instanceof TaggerImp)
					return true;
				if (parent.getOperator() instanceof MergeImp)
					return true;
				if (parent.getOperator() instanceof NatureJoinImp)
					return true;
				if ((parent.getOperator() instanceof FunctionImp)
					&& (((Function) parent.getOperator())
						.getName()
						.equalsIgnoreCase("count"))) {
					return true;
				}
			}
			if (parent.getOperator() instanceof LeftOuterJoinImp)
				return false;
			parent = parent.getParent();
		}
		return false;
	}

	private boolean isConsumeFor() {
		HashSet consumed = new HashSet();
		XATNode rootOfSubTree =
			((FOR) this.forNode.getOperator()).getXtreFunctions().getRoot();
		consumed = consumedCol(rootOfSubTree);
		//Here we assume a For clause only has one binded variable.
		if (consumed.contains(this.contextForBindings[0]))
			return true;
		else
			return false;
	}
	private boolean isConsumeFor(XATNode rootOfSubTree) {
		HashSet consumed = new HashSet();
		consumed = consumedCol(rootOfSubTree);
		//Here we assume a For clause only has one binded variable.
		if (consumed.contains(this.contextForBindings[0]))
			return true;
		else
			return false;
	}

	//Return all the columns used in the XAT tree recursively.
	private HashSet consumedCol(XATNode node) {
		HashSet consumed = new HashSet();
		try {
			if (node == null)
				return consumed;
			if ((!(node.getOperator() instanceof LeftOuterJoinImp))
				&& (!(node.getOperator() instanceof NatureJoinImp)))
				consumed = (HashSet) consumedVisitor.visit(node);
			XATNode[] children = node.getChildren();
			if (children.length > 0) {
				// append all the children's consumed columns.
				for (int i = 0; i < children.length; i++) {
					HashSet temp = (HashSet) consumedCol(children[i]);
					if (temp != null)
						consumed.addAll(temp);
				}
			}
			//for FOR op
			if (node.getOperator() instanceof FORImp)
				consumed.addAll(
					(HashSet) consumedCol(((FOR) node.getOperator())
						.getXtreFunctions()
						.getRoot()));

			//for GB op
			if (node.getOperator() instanceof GroupByImp)
				consumed.addAll(
					(HashSet) consumedCol(((GroupBy) node.getOperator())
						.getXATTree()
						.getRoot()));
		} catch (java.lang.NullPointerException e) {
			System.out.println();
		}

		return consumed;
	}
	/**
	 * Access method of field context of outter for bindings.
	 *
	 * @return array of String
	 * @since 1.0
	 */
	private String[] getForContext() {
		return contextForBindings;
	}
	/**
	 * Create GroupBy operator at the place of an XAT operator and put the
	 * XAT operator as the root of the inner tree of the new GroupBy operator.
	 * 
	 * @return edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
	 * @param op edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator
	 * @since 1.0
	 */
	XATOperator createGBOp(XATOperator op) {

		GroupBy gbOp = builder.newGroupBy(new XATNodeImp());
		gbOp.setOutputColumnName(op.getOutputColumnName());

		// set the groupBy's input columns.
		gbOp.setColumnNames(getForContext());

		// create the subtree of the operator.
		XATTree subTree = new XATTreeImp();

		// disconnect the op's node from it's children.
		XATNode opNode = (XATNode) op.getNode().clone();
		opNode.setChildren(new XATNodeImp[0]);

		// add the op's node to the subtree.
		subTree.setRoot(opNode);

		gbOp.setXATTree(subTree);

		return gbOp;
	}

	public Object dispatch() {
		//System.out.println(indent() + "NULL");
		return null;
	}
	/**
	 * Called when an object hasn't been implemented
	 * by the visitor pattern
	 * default handling of the unknow operators.
	 */
	public Object dispatch(Object obj) {

		XATOperator opr = (XATOperator) obj;

		// do nothing here. Because we modify on the cloned input tree directly.	

		if (isDebug())
			System.out.println(
				"Operator of type "
					+ opr.getClass().getName()
					+ " is not handled!");
		if (isDebug())
			System.out.println(
				"This node is simply added to the Decorrelated Tree as is.");

		return null;
	}
	/**
	 * Access method of field decorrelate tree.
	 *
	 * @return XATTree
	 * @since 1.0
	 */
	public XATTree getDecorTree() {
		return this.decorrTree;
	}
	/**
	 * Get the value of the propertity 'DEBUG_QueryDecorrelation'.
	 * @return boolean True: With debug information. False: No debug information.
	 *
	 * @since 1.0
	 */
	public boolean isDebug() {
		return System.getProperty("DEBUG_QueryDecorrelation") != null;
	}
	/**
	 * Pushing down function for context insensitive operator having single child
	 * 
	 * @return java.lang.Object
	 * @param contextInsensetiveOp edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.XATOperator
	 * @since 1.0
	 */
	public Object visit(XATOperator contextInsensetiveOp) {
		//just push for op down.
		if (isDebug())
			System.out.println("pushing for node down\n");

		XATNode subTreeRoot = contextInsensetiveOp.getNode();
		//if this.forNode is the root of the subtree, this.currentParent should be null
		//we need to set the root of the tree then.
		if (this.currentParent != null) {
			decorrTree.addNode(subTreeRoot, this.currentParent);
			this.currentParent.deleteChild(this.forNode);
		} else {
			decorrTree.setRoot(subTreeRoot);
		}

		//push FOR under the root node. here we assume there only one child existing.
		//for multiple children, we have another visit()
		if (subTreeRoot.getChildrenSize() > 0) {
			XATNode child = subTreeRoot.getChild(0);
			((FOR) this.forNode.getOperator()).getXtreFunctions().setRoot(
				child);
			subTreeRoot.deleteChild(child);
			decorrTree.addNode(this.forNode, subTreeRoot);
		} else {
			((FOR) this.forNode.getOperator()).getXtreFunctions().setRoot(null);
			decorrTree.addNode(this.forNode, subTreeRoot);
		}

		this.currentParent = subTreeRoot;
		decor();
		return null;
	}
	/**
	 * Pushing down function for context insensitive operator having multiple children
	 * 
	 * @return java.lang.Object
	 * @param contextInsensetiveOp edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.XATMultiSourceOperator
	 * @since 1.0
	 */
	public Object visit(XATMultiSourceOperator contextInsensetiveOp) {
		//just push for op down.
		if (isDebug())
			System.out.println("multiple pushing for node down\n");

		XATNode subTreeRoot = contextInsensetiveOp.getNode();

		//if this.forNode is the root of the subtree, this.currentParent should be null
		//we need to set the root of the tree then.
		if (this.currentParent != null) {
			decorrTree.addNode(subTreeRoot, this.currentParent);
			this.currentParent.deleteChild(this.forNode);
		} else {
			decorrTree.setRoot(subTreeRoot);
		}

		//clone and push FOR under the subroot node.
		XATNode[] children = subTreeRoot.getChildren();
		for (int i = 0; i < subTreeRoot.getChildrenSize(); i++) {
			XATNode child = children[i];
			if (isConsumeFor(child)) {
				XATTree forTree = new XATTreeImp();
				forTree.setRoot(this.forNode);
				XATTree newForTree = (XATTree) forTree.clone();
				XATNode newForTreeRoot = newForTree.getRoot();

				((FOR) newForTreeRoot.getOperator())
					.getXtreFunctions()
					.setRoot(
					child);
				subTreeRoot.deleteChild(child);
				decorrTree.addNode(newForTreeRoot, subTreeRoot);
				this.currentParent = subTreeRoot;
				this.forNode = newForTreeRoot;
				decor();
			}
		}
		return null;
	}
	/**
	 * Pushing down function for NatureJoin
	 * 
	 * @return java.lang.Object
	 * @param natureJoin edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.NatureJoin
	 * @since 1.0
	 */
	public Object visit(NatureJoin natureJoin) {
		//just push for op down.
		if (isDebug())
			System.out.println("natureJoin pushing for node down\n");

		XATNode subTreeRoot = natureJoin.getNode();

		//if this.forNode is the root of the subtree, this.currentParent should be null
		//we need to set the root of the tree then.
		if (this.currentParent != null) {
			decorrTree.addNode(subTreeRoot, this.currentParent);
			this.currentParent.deleteChild(this.forNode);
		} else {
			decorrTree.setRoot(subTreeRoot);
		}

		XATNode[] children = subTreeRoot.getChildren();
		if (children.length > 2)
			System.out.println(
				"natureJoin decor wrong. check the children of Merge");

		//reorder the children XATNode[], make sure that the first child need 
		//the magic branch pushing down. OR all branches do not need to push down.
		int firstChildConsumeFor = -1;
		boolean[] isConsume = new boolean[children.length];
		for (int i = 0; i < children.length; i++) {
			isConsume[i] = false;
			if (isConsumeFor(children[i])) {
				isConsume[i] = true;
				if (firstChildConsumeFor == -1)
					firstChildConsumeFor = i;
			}
		}

		if ((firstChildConsumeFor != -1) && (firstChildConsumeFor != 0)) {
			XATNode tempXATNode = children[0];
			children[0] = children[firstChildConsumeFor];
			children[firstChildConsumeFor] = tempXATNode;
			isConsume[0] = true;
			isConsume[firstChildConsumeFor] = false;
		}

		//clone and push FOR under the subroot node.
		for (int i = 0; i < subTreeRoot.getChildrenSize(); i++) {
			XATNode child = children[i];
			if (isConsume[i]) {
				XATTree forTree = new XATTreeImp();
				forTree.setRoot(this.forNode);
				XATTree newForTree = (XATTree) forTree.clone();
				XATNode newForTreeRoot = newForTree.getRoot();

				((FOR) newForTreeRoot.getOperator())
					.getXtreFunctions()
					.setRoot(
					child);
				subTreeRoot.deleteChild(child);
				decorrTree.addNode(newForTreeRoot, subTreeRoot);
				this.currentParent = subTreeRoot;
				this.forNode = newForTreeRoot;
				decor();
			} else { //delete and add, only to keep the correct order of the child of the NatureJoin
				subTreeRoot.deleteChild(child);
				decorrTree.addNode(child, subTreeRoot);
			}
		}
		return null;
	}
	/**
	 * We will change Merge operator to NatureJoin Operator and introducing visit(NatureJoin)
	 * to decide the Left and Right branch. The tuple order is decided by the L / R.
	 *
		 * @return java.lang.Object
	 * @param mergeOp edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.MergeOp
	 * @since 1.0
	 */
	public Object visit(Merge mergeOp) {
		//just push for op down.
		if (isDebug())
			System.out.println("merge pushing for node down\n");
		XATNode subTreeRoot = mergeOp.getNode();

		//change Merge to NatureJoin
		//1. if there only one child under the merge. delet this Merge
		if (subTreeRoot.getChildrenSize() == 1) {
			XATNode child = subTreeRoot.getChild(0);
			((FOR) this.forNode.getOperator()).getXtreFunctions().setRoot(
				child);
			decor();
		}
		//2. if there is more than 2 children under the merge, change to 
		// a serial of Nature Join. At this point, the position of child doesn't matter. 
		else if (subTreeRoot.getChildrenSize() > 2) {
			XATTree tempTree = new XATTreeImp();
			tempTree.setRoot(subTreeRoot);

			NatureJoin natureJoin = builder.newNatureJoin(subTreeRoot, null);
			natureJoin.setOutputColumnName(mergeOp.getOutputColumnName());
			subTreeRoot.setOperator(natureJoin);
			XATNode[] children = subTreeRoot.getChildren();

			subTreeRoot.setChildren(new XATNode[] {
			});
			tempTree.addNode(children[0], subTreeRoot);

			XATNode tempParent = subTreeRoot;
			for (int i = 0; i < children.length - 2; i++) {
				XATNode natureJoinNode = new XATNodeImp();
				NatureJoin natureJoinChild =
					builder.newNatureJoin(natureJoinNode, null);
				natureJoinChild.setOutputColumnName(
					mergeOp.getOutputColumnName());
				natureJoinNode.setOperator(natureJoinChild);
				natureJoinNode.setChildren(new XATNode[] {
				});

				tempTree.addNode(natureJoinNode, tempParent);

				tempTree.addNode(children[i + 1], natureJoinNode);
				if (i == children.length - 3) {
					//last child for original mergeOp
					tempTree.addNode(children[i + 2], natureJoinNode);
				}

				tempParent = natureJoinNode;

			}

			//recursively call visit(NatureJoin)
			visit((Object) natureJoin);

		}
		//3. if there is only 2 children, then change the merge to one Nature Join
		else {
			NatureJoin natureJoin = builder.newNatureJoin(subTreeRoot, null);
			natureJoin.setOutputColumnName(mergeOp.getOutputColumnName());
			subTreeRoot.setOperator(natureJoin);

			//recursively call visit(NatureJoin)
			visit((Object) natureJoin);
		}
		return null;
	}

	/**
	 * Called when an aggregate function exists in the tree.
	 * A GroupBy node replaces the Function node, and push down for
	 * 
	 * @return java.lang.Object
	 * @param function edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.FunctionImp
	 * @since 1.0
	 */
	public Object visit(Function function) {

		if (isDebug())
			System.out.println("Function added to decorrelation tree");

		XATNode subTreeRoot = function.getNode();

		//if this.forNode is the root of the subtree, this.currentParent should be null
		//we need to set the root of the tree then.
		if (this.currentParent != null) {
			decorrTree.addNode(subTreeRoot, this.currentParent);
			this.currentParent.deleteChild(this.forNode);
		} else {
			decorrTree.setRoot(subTreeRoot);
		}

		if (this.getForContext() != null) {
			if (this.getForContext().length != 0) {
				// for the function there will be a groupby operator created with the FOR context,
				// and then put the function as the subtree of it.
				// in terms of the decorrelated tree, the function node will be deleted from its parent.
				// and then the groupby will take its position.
				XATOperator gbOp = createGBOp(function);
				gbOp.setNode(subTreeRoot);
				// replace the aggregate operator with the gbNode.
				subTreeRoot.setOperator(gbOp);
			}
		}

		if (subTreeRoot.getChildrenSize() > 0) {
			XATNode child = subTreeRoot.getChild(0);
			((FOR) this.forNode.getOperator()).getXtreFunctions().setRoot(
				child);
			subTreeRoot.deleteChild(child);
			decorrTree.addNode(this.forNode, subTreeRoot);
		} else {
			((FOR) this.forNode.getOperator()).getXtreFunctions().setRoot(null);
			decorrTree.addNode(this.forNode, subTreeRoot);
		}

		this.currentParent = subTreeRoot;
		decor();
		return null;
	}

	/**
	 * Rename operator
	 * Specially to set the return Rename node
	 * used for deciding the necessary of LOJ
	 * 
	 * @return java.lang.Object
	 * @param renNode edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.RenameImp
	 * @since 1.0
	 */
	public Object visit(NameColumn renNode) {

		if (isDebug())
			System.out.println("NameColumn added to decorrelation tree\n");

		if ("return".equals(renNode.getBlockDesc())) {
			//if passed the second return, that is belong to another for below
			if (this.hasPassedReturn) {
				this.hasPassed2ndReturn = true;
			} else {
				this.hasPassedReturn = true;
			}
		}

		XATNode subTreeRoot = renNode.getNode();

		//if this.forNode is the root of the subtree, this.currentParent should be null
		//we need to set the root of the tree then.
		if (this.currentParent != null) {
			decorrTree.addNode(subTreeRoot, this.currentParent);
			this.currentParent.deleteChild(this.forNode);
		} else {
			decorrTree.setRoot(subTreeRoot);
		}

		//push FOR under the root node. here we assume there only one child existing.
		//for multiple children, we have another visit() 
		if (subTreeRoot.getChildrenSize() > 0) {
			XATNode child = subTreeRoot.getChild(0);
			((FOR) this.forNode.getOperator()).getXtreFunctions().setRoot(
				child);
			subTreeRoot.deleteChild(child);
			decorrTree.addNode(this.forNode, subTreeRoot);
		} else {
			((FOR) this.forNode.getOperator()).getXtreFunctions().setRoot(null);
			decorrTree.addNode(this.forNode, subTreeRoot);
		}

		this.currentParent = subTreeRoot;
		decor();
		return null;
	}
	/**
	 * for the groupby there will be a groupby operator created with the FOR context,
	 * and then put the original groupby as the subtree of it.
	 * in terms of the decorrelated tree, the old groupby node will be deleted from its parent.
	 * and then the new groupby will take its position.
	 * 
	 * @return java.lang.Object
	 * @param groupby edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.GroupByImp
	 * @since 1.0
	 */
	public Object visit(GroupBy groupby) {

		if (isDebug())
			System.out.println("Combine node added to decorrelation tree");

		XATNode subTreeRoot = groupby.getNode();

		//if this.forNode is the root of the subtree, this.currentParent should be null
		//we need to set the root of the tree then.
		if (this.currentParent != null) {
			decorrTree.addNode(subTreeRoot, this.currentParent);
			this.currentParent.deleteChild(this.forNode);
		} else {
			decorrTree.setRoot(subTreeRoot);
		}

		if (this.getForContext() != null) {
			if (this.getForContext().length != 0) {

				XATOperator gbOp = createGBOp(groupby);
				gbOp.setNode(groupby.getNode());

				// replace the aggregate operator with the gbNode.
				groupby.getNode().setOperator(gbOp);
			}
		}

		if (subTreeRoot.getChildrenSize() > 0) {
			XATNode child = subTreeRoot.getChild(0);
			((FOR) this.forNode.getOperator()).getXtreFunctions().setRoot(
				child);
			subTreeRoot.deleteChild(child);
			decorrTree.addNode(this.forNode, subTreeRoot);
		} else {
			((FOR) this.forNode.getOperator()).getXtreFunctions().setRoot(null);
			decorrTree.addNode(this.forNode, subTreeRoot);
		}

		this.currentParent = subTreeRoot;
		decor();
		return null;
	}
	/**
	 * Pushdown function for orderby operator
	 * 
	 * @return java.lang.Object
	 * @param orderBy edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.OrderByImp
	 * @since 1.0
	 */
	public Object visit(OrderBy orderBy) {

		if (isDebug())
			System.out.println("Orderby added to decorrelation tree");

		XATNode subTreeRoot = orderBy.getNode();

		//if this.forNode is the root of the subtree, this.currentParent should be null
		//we need to set the root of the tree then.
		if (this.currentParent != null) {
			decorrTree.addNode(subTreeRoot, this.currentParent);
			this.currentParent.deleteChild(this.forNode);
		} else {
			decorrTree.setRoot(subTreeRoot);
		}

		if (this.getForContext() != null) {
			if (this.getForContext().length != 0) {

				XATOperator gbOp = createGBOp(orderBy);
				gbOp.setNode(orderBy.getNode());

				// replace the aggregate operator with the gbNode.
				orderBy.getNode().setOperator(gbOp);
			}
		}

		if (subTreeRoot.getChildrenSize() > 0) {
			XATNode child = subTreeRoot.getChild(0);
			((FOR) this.forNode.getOperator()).getXtreFunctions().setRoot(
				child);
			subTreeRoot.deleteChild(child);
			decorrTree.addNode(this.forNode, subTreeRoot);
		} else {
			((FOR) this.forNode.getOperator()).getXtreFunctions().setRoot(null);
			decorrTree.addNode(this.forNode, subTreeRoot);
		}

		this.currentParent = subTreeRoot;
		decor();
		return null;
	}
	/**
	 * for the aggregate function there will be a groupby operator
	 * created with the FOR context, and then put the aggregate 
	 * function as the subtree of it. In terms of the decorrelated tree, 
	 * the aggregate node will be deleted from its parent. and then the 
	 * groupby will take its position.
	 *
	 * @return java.lang.Object
	 * @param aggregate edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.CombineImp
	 * @since 1.0
	 */
	public Object visit(Combine aggregate) {

		if (isDebug())
			System.out.println("Combine node added to decorrelation tree");

		XATNode subTreeRoot = aggregate.getNode();

		//if this.forNode is the root of the subtree, this.currentParent should be null
		//we need to set the root of the tree then.
		if (this.currentParent != null) {
			decorrTree.addNode(subTreeRoot, this.currentParent);
			this.currentParent.deleteChild(this.forNode);
		} else {
			decorrTree.setRoot(subTreeRoot);
		}

		if (this.getForContext() != null) {
			if (this.getForContext().length != 0) {
				XATOperator gbOp = createGBOp(aggregate);
				gbOp.setNode(aggregate.getNode());

				// replace the aggregate operator with the gbNode.
				aggregate.getNode().setOperator(gbOp);
			}
		}

		if (subTreeRoot.getChildrenSize() > 0) {
			XATNode child = subTreeRoot.getChild(0);
			((FOR) this.forNode.getOperator()).getXtreFunctions().setRoot(
				child);
			subTreeRoot.deleteChild(child);
			decorrTree.addNode(this.forNode, subTreeRoot);
		} else {
			((FOR) this.forNode.getOperator()).getXtreFunctions().setRoot(null);
			decorrTree.addNode(this.forNode, subTreeRoot);
		}

		this.currentParent = subTreeRoot;
		decor();
		return null;
	}
}
