/* Kweelt - a framework to query XML documents
 * Copyright (C) 2000,  Arnaud Sahuguet (Arnaud.Sahuguet@polytechnique.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package xacute.quilt;

import xacute.common.Node;
import xacute.common.NodeList;
import xacute.common.QuiltException;


import java.util.Vector;

/**
 * This class implements a dereferencing step
 * in <i>Quilt</i>, <i>i.e.</i> a construct of
 * the following type :
 * <code align="center">@spouse-></code>
 * (where the deferencing step itself is
 * <code>-></code>)
 * which returns the elements of the document
 * that have an <code>ID</code> attribute equal
 * to the value of the <code>spouse</code>
 * attribute of the context-node.
 *
 * A <code>DerefStep</code> is NOT a <code>QuiltExpression</code>.
 *
 * This construct can also be enhanced through
 * <i>hints</i> in the following way:
 * <code align="center">->{person@name}</code>
 * which means that we look up only for
 * <name> attributes within <code>person</code>
 * elements.
 * Several hints can be mentioned, separated by
 * commas.
 */
public class DerefStep
    extends Step
{

    private DerefPair[] hints = new DerefPair[0];

    DerefStep() {}

    void setHints(Vector v)
    {
	this.hints = new DerefPair[v.size()];
	v.toArray(this.hints);
    }

    void setPredicates(Vector v)
    {
	this.predicates = new QuiltExpression[v.size()];
	v.toArray(this.predicates);
    }

    public String pprint()
    {
	StringBuffer buf = new StringBuffer();
	buf.append("->");
	if (this.hints.length > 0)
	    {
		buf.append("{");
		DerefPair p = this.hints[0];
		buf.append(p.getElt() + "@" + p.getAtt());
		for(int i=1; i < this.hints.length; i++)
		    {
			p = this.hints[i];
			buf.append(", " + p.getElt() + "@" + p.getAtt());
		    }
		buf.append("}");
	    }
	if (this.predicates.length > 0)
	    for(int k = 0; k < this.predicates.length; k++)
		buf.append("[" + this.predicates[k] + "]");
	return buf.toString();
    }

    NodeList evalStep(Node n, EvalContext con)
	throws QuiltException
    {
	NodeList nl = n.deref(n.stringValue(), this.hints);
	for (int k=0; k < this.predicates.length; k++)
	    nl = nl.applyPredicate( new Predicate(this.predicates[k], con) );
	return nl;
    }

    public QuiltExpression[] returnQE() {
      return predicates;
    }

}


