/* Kweelt - a framework to query XML documents
 * Copyright (C) 2000,  Arnaud Sahuguet (Arnaud.Sahuguet@polytechnique.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package xacute.quilt;

import xacute.common.Constants;
import xacute.common.NodeList;
import xacute.common.Node;
import xacute.common.QuiltException;

import java.util.Vector;
import java.util.TreeMap;

/**
 * This class implements <i>Quilt</i>'s
 * <b>FOR</b> / <b>LET</b> /
 * <b>WHERE</b> / <b>RETURN</b> expressions.
 */
public class FLWRExpression extends QuiltExpression
{

    private Binding[] bindings;

    private QuiltExpression where;

    public QuiltExpression getWhere()
    {
        return where;
    }

    private QuiltExpression e;

    private SortClause[] sort;

    public SortClause[] getSort()
    {
        return sort;
    }

    FLWRExpression(
        Vector bindings,
        QuiltExpression where,
        QuiltExpression e,
        Vector sort)
    {
        this.bindings = new Binding[bindings.size()];
        bindings.toArray(this.bindings);
        this.where = where;
        this.e = e;
        if (sort != null)
            {
            this.sort = new SortClause[sort.size()];
            sort.toArray(this.sort);
        } else
            this.sort = new SortClause[0];
    }

    Value eval(EvalContext con) throws QuiltException
    {
        if (con.getNode() != null)
            throw new RuntimeException("xacute.quilt.FLWRExpression.eval() with a non-null node");
        Vector result = new Vector();
        BindingTree.buildTreeAndPaths(
            this.bindings,
            con,
            this.where,
            this.sort,
            result);
        return this.evalInOrder(result, con);
    }

    public String pprint()
    {
        StringBuffer buf = new StringBuffer();
        buf.append("( ");
        for (int i = 0; i < this.bindings.length; ++i)
            buf.append(this.bindings[i].pprint() + "\n");
        if (where != null)
            buf.append("WHERE " + this.where.pprint() + "\n");
        buf.append("RETURN " + this.e.pprint());
        for (int i = 0; i < this.sort.length; i++)
            buf.append(" " + this.sort[i].pprint());
        buf.append(" )");
        return buf.toString();
    }

    private Value evalInOrder(Vector tm, EvalContext con) throws QuiltException
    {
        // The call returns sorted results.
        // Depending on the nature of the sort, the result will be returned
        // forward or backward.
        BindingTreePath[] paths = new BindingTreePath[tm.size()];
        tm.toArray(paths);
        NodeList nl = con.getNodeFactory().makeEmptyNodeList();
        if (this.sort.length == 0)
            {
            for (int k = 0; k < paths.length; k++)
                {
                con.getEnv().setBindings(this.bindings, paths[k]);
                nl = nl.APPEND(this.e.eval(con).toNodeList(con));
            }
            if (paths.length > 0)
                {
                // System.err.println("in FLWR");
                // modified by Xin. for the purpose of SequenceExpression.
                // con.getEnv().rmBindings(this.bindings);
            }
            return new ValueNodeList(nl);
        } else
            {
            NodeListTreeMap sortMap = new NodeListTreeMap(new MyCompare(this.sort));
            for (int k = 0; k < paths.length; k++)
                {
                con.getEnv().setBindings(this.bindings, paths[k]);
                Node n = this.e.eval(con).toNode(con);
                //con.getEnv().rmBindings(this.bindings);
                EvalContext newCon = new EvalContext(n, -1, -1, con.getEnv());
                newCon.setNodeFactory(con.getNodeFactory());
                String[] multiKey = new String[this.sort.length];
                for (int j = 0; j < this.sort.length; j++)
                    multiKey[j] = this.sort[j].eval(newCon).toString(con);
                sortMap.put(multiKey, n);
            }
            Vector sortedResult = sortMap.values();
            for (int k = 0; k < sortedResult.size(); k++)
                nl = nl.APPEND((Node) sortedResult.elementAt(k));
            return new ValueNodeList(nl);
        }
    }

    public QuiltExpression getE()
    {
        return e;
    }
    public Binding[] getBindings()
    {
        return bindings;
    }

}