/* Kweelt - a framework to query XML documents
 * Copyright (C) 2000,  Arnaud Sahuguet (Arnaud.Sahuguet@polytechnique.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package xacute.quilt;

import xacute.common.NodeFactory;

import org.xml.sax.DocumentHandler;
import java.io.*;

import gnu.getopt.*;

public class Main
{

    public final static String TOP     = "+---------------------------------------------+";
    public final static String VERSION = "| Kweelt version 1.04 / Bergerac / 2000-09-10 |";
    public final static String CREDITS = "| Copyright 2000, Arnaud Sahuguet             |";
    public final static String EMAIL   = "| MAIL: Arnaud.Sahuguet@polytechnique.org     |";
    public final static String URL     = "| URL: http://db.cis.upenn.edu/Kweelt         |";
    public final static String BOTTOM  = "+---------------------------------------------+";
    
    public static void main(String args[])
	throws Exception
    {
	boolean parseTime  = false;
	boolean execTime   = false;
	boolean verbose    = false;
	boolean parseOnly  = false;
	boolean printQuery = false;
	String nodeFactory = "xacute.impl.xdom.NodeFactoryXerces";
	String handlerName = "xacute.util.OutputHandler";
	int benchmark      = 1;
	String query       = null;
	
	int c;
	String arg;
	LongOpt[] longopts = new LongOpt[12];
	StringBuffer sb = new StringBuffer();

	longopts[0] = new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h', "to get the option details", null);
	longopts[1] = new LongOpt("factory", LongOpt.REQUIRED_ARGUMENT, null, 1, "to specify the NodeFactory to use", "fully qualified NodeFactory (e.g. xacute.impl.xdom.NodeFactoryXerces)");
	longopts[2] = new LongOpt("parse_time", LongOpt.NO_ARGUMENT, null, 2, "to get the time spent parsing the query", null);
	longopts[3] = new LongOpt("exec_time", LongOpt.NO_ARGUMENT, null, 3, "to get the time spent evaluating the query", null);
	longopts[4] = new LongOpt("verbose", LongOpt.NO_ARGUMENT, null, 4, "to get some trace information", null);
	longopts[5] = new LongOpt("output", LongOpt.REQUIRED_ARGUMENT, null, 5, "to specify which output handler to use", "fully qualified hanlder (e.g. xacute.util.HashHandler)");
	longopts[6] = new LongOpt("parse_only", LongOpt.NO_ARGUMENT, null, 6, "to parse the query and skip the evaluation", null);
	longopts[7] = new LongOpt("print_query", LongOpt.NO_ARGUMENT, null, 7, "to print the parsed query", null);
	longopts[8] = new LongOpt("benchmark", LongOpt.REQUIRED_ARGUMENT, null, 8, "to evaluate the query multiple times (useful for benchmarks", "the number of iterations");
	longopts[9] = new LongOpt("query", LongOpt.REQUIRED_ARGUMENT, null, 9, "to input the query from the command line", "the query in ASCII");
	longopts[10] = new LongOpt("version", LongOpt.NO_ARGUMENT, null, 'v', "to get the version of the program", null);
	longopts[11] = new LongOpt("check-classpath", LongOpt.NO_ARGUMENT, null, 10, "to check which packages (optentially needed by Kweelt) are available from the CLASSPATH", null);
	
	Getopt g = new Getopt("xacute.quilt.Main", args, "h::v::", longopts);
	g.setOpterr(false);
	
	while ((c=g.getopt()) != -1)
	    switch(c)
		{
		case 'h':
		    help(longopts);
		    System.exit(0);
		    break;
		case 1:
		    nodeFactory = g.getOptarg();
		    break;
		case 2:
		    parseTime = true;
		    break;
		case 3:
		    execTime = true;
		    break;
		case 4:
		    verbose = true;
		    break;
		case 5:
		    handlerName = g.getOptarg();
		    break;
		case 6:
		    parseOnly = true;
		    break;
		case 7:
		    printQuery = true;
		    break;
		case 8:
		    benchmark = new Integer( g.getOptarg() ).intValue();
		    break;
		case 9:
		    query = g.getOptarg();
		    break;
		case 10:
		    checkClasspath();
		    System.exit(0);
		    break;
		case 'v':
		    System.out.println( TOP + "\n" + VERSION + "\n" + CREDITS + "\n" + EMAIL + "\n" + URL + "\n" + BOTTOM );
		    System.exit(0);
		    break;
		case '?':
		    throw new Exception("The option '" + args[g.getOptind()-1] + "' is not valid");
		default:		   
		    throw new Exception( args[g.getOptind()] );
		}

	if (g.getOptind() < args.length)
	    query = readQuery( args[g.getOptind()] );
	else if (query == null)
	    {
		help(longopts);
		throw new Exception("You must specify a query");
	    }
	
	
	QuiltParser parser = new QuiltParser();
	QuiltQuery q = null;
	long parse_t = -1;
	long exec_t = -1;
	
	long l = System.currentTimeMillis();
	try
	    {
		q = parser.parseQuery(query);
		parse_t = System.currentTimeMillis() - l;
		if (printQuery) System.out.println(q.pprint());
		if (parseOnly)
		    {
			if (parseTime) System.out.println("Parse time: " + parse_t);
			System.exit(0);
		    }
	    }
	catch(Exception e)
	    {
		e.printStackTrace(System.out);
		System.exit(-1);
	    }

	DocumentHandler h = null;
	try
	    {
		h = (DocumentHandler) Class.forName(handlerName).newInstance();
	    }
	catch (Exception e)
	    {
		e.printStackTrace(System.out);
		System.exit(-1);
	    }

	NodeFactory nf = null;
	try
	    {
		nf = (NodeFactory) Class.forName(nodeFactory).newInstance();
	    }
	catch (Exception e)
	    {
		e.printStackTrace(System.out);
		System.exit(-1);
	    }
	
	EvalContext con = new EvalContext();
	con.setNodeFactory( nf );
	l = System.currentTimeMillis();
	q.eval(h, con);
	exec_t = System.currentTimeMillis() - l;

	System.out.println(h.toString());
	if (parseTime) System.out.println("Parse time     (ms): " + parse_t);
	if (execTime)  System.out.println("Execution time (ms): " + exec_t );
	
	
	System.exit(0);
	
    }

    public static void help(LongOpt[] options)
    {
	StringBuffer buf = new StringBuffer("usage:\n");
	for(int i=0; i < options.length; i++)
	    {
		buf.append( "--" + options[i].getName() + " ");
		if (options[i].getHasArg() == LongOpt.OPTIONAL_ARGUMENT)
		    buf.append("[ <" + options[i].getArgDesc() + ">]");
		if (options[i].getHasArg() == LongOpt.REQUIRED_ARGUMENT)
		    buf.append("<" + options[i].getArgDesc() + ">");
		buf.append("\n\t" + options[i].getOptionDesc() + "\n");
	    }
	System.out.println(buf.toString());    
    }

    private static String readQuery(String file)
	throws Exception
    {
	DataInputStream dis = new DataInputStream(new FileInputStream(file));
	StringBuffer buf = new StringBuffer();
	String s;
	while ( (s=dis.readLine()) != null)
	    buf.append(s + "\n");
	return buf.toString();
    }

    private static void checkClasspath()
    {
	String[][] pkgs = {
	    { "Java 2", "mandatory", "java.util.TreeMap" },			
	    { "W3C DOM", "mandatory", "org.w3c.dom.Document" },			
	    { "SAX", "mandatory", "org.xml.sax.InputSource"},
	    { "Xerces DOM/SAX parsers", "mandatory", "org.apache.xerces.parsers.DOMParser" },	
	    { "Sun Project X DOM/SAX parsers", "optional", "javax.xml.parsers.DocumentBuilderFactory" },
	    { "Oracle DOM parser", "optional", "oracle.xml.parser.v2.DOMParser" },
	    { "AElfred SAX parser", "optional", "com.microstar.xml.SAXDriver"},
	    { "Apache Regular Expression", "needed by KSP + useful for user-defined functions", "org.apache.regexp.RE" },
	    { "Servlet API", "needed by KSP", "javax.servlet.http.HttpServletRequest" },
	    { "Cocoon", "needed to compile KSP", "org.apache.cocoon.processor.Processor" }		
	};
	System.out.println("\nLooking for available packages:\n");
	for(int i=0; i < pkgs.length; i++)
	    {
		System.out.println("(*) package " + pkgs[i][0] + " (" + pkgs[i][1] + ")");
		try
		    {
			Class.forName(pkgs[i][2]);
			System.out.println("\t" + "FOUND\n");
		    }
		catch(Exception e)
		    {
			System.out.println("\tNOT FOUND\n");
		    }
	    }
	return;
    }
    
}


