/* Kweelt - a framework to query XML documents
 * Copyright (C) 2000,  Arnaud Sahuguet (Arnaud.Sahuguet@polytechnique.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package xacute.quilt;

import xacute.common.Constants;
import xacute.common.Node;
import xacute.common.NodeList;
import xacute.common.QuiltException;


/**
 * This class represents <i>Quilt</i>'s numeric
 * basic values.
 * <br> We store this value as a float, even if
 *      it is an integer.
 * <br> Whenever a <i>Quilt</i>'s string value 
 *      can be converted to a numeric value, 
 *      it <b>must</b> be done !
 */
public class ValueNum
    extends ValueCompared
{

    private float value;

    /**
     * Constructs a <code>ValueNum</code>
     * from a <code>String</code>.
     */
    public ValueNum(String s)
    {
	this((new Float(s)).floatValue());
    }
    
    /**
     * Constructs a <code>ValueNum</code>
     * from a float number.
     */
    public ValueNum(float f)
    {
	this.value = f;
    }
    
    public void toSAX(org.xml.sax.DocumentHandler handler) 
	throws org.xml.sax.SAXException
    {
	Tools.reportString(handler, this.toString(null));
    }
    
    public boolean toBool(EvalContext con)
    {
	return (this.value > 0);
    }
    

    public String pprint()
    {
	return "" + this.value;
    }

    public String toString(EvalContext con)
    {
	return "" + this.value;
    }

    /**
     * Accessor to the numeric value.
     */
    public float toNum(EvalContext con)
    {
	return this.value;
    }

    public String printResult()
    {
	return this.toString();
    }

    public Value filter(QuiltExpression[] filters, EvalContext con)
	throws QuiltException
    {
	for (int k=0; k < filters.length; k++)
	    if (!filters[k].eval(con).toBool(con))
		return new ValueNodeList(con.getNodeFactory().makeEmptyNodeList()); // choice of semantics
	return this;
    }

    /**
     * Applies an arithmetic operation to a
     * <code>ValueNum</code>.
     * @param op a constant (from <code>xacute.xmltree.Constants</code>)
     *           representing the arithmetic operation
     * @param n the right-hand side of the operation
     */
    void applyOperation(int op, ValueNum n, EvalContext con)
	throws QuiltException
    {
	switch (op) {
	case '+':
	    this.value += n.toNum(con);
	    break;
	case '-':
	    this.value -= n.toNum(con);
	    break;
	case '*':
	    this.value *= n.toNum(con);
	    break;
	case '/':
	    this.value = this.value / n.toNum(con);
	    break;
	case '%':
	    this.value = this.value % n.toNum(con);
	    break;
	default:
	    throw new QuiltException("Unknown arithmetic operator : " + (char) op);
	}
    }

    // Num -> Bool -> boolean
    public boolean compare(int op, ValueBool b, EvalContext con)
	throws QuiltException
    {
	return b.compare( Constants.COMPOPS_REVERSE[op], new ValueBool(this.toBool(con)), con);
    }
    
    // Num -> Num -> boolean
    public boolean compare(int op, ValueNum f, EvalContext con)
	throws QuiltException
    {
	switch (op) {
	case Constants.LT:
	    return (this.value < f.value);
	case Constants.GT:
	    return (this.value > f.value);
	case Constants.GEQ:
	     return (this.value >= f.value);
	case Constants.LEQ:
	    return (this.value <= f.value);
	case Constants.EQ:
	    return (this.value == f.value);
	case Constants.NEQ:
	    return (this.value != f.value);
	default:
	    throw new QuiltException("Unknown comparison operator: " + op);
	}
    }

    // Num -> String -> boolean
    public boolean compare(int op, ValueString s, EvalContext con)
	throws QuiltException
    {
	return s.compare( Constants.COMPOPS_REVERSE[op], this, con );
    }

    // Num -> Node -> boolean 
    public boolean compare(int op, ValueNode n, EvalContext con)
	throws QuiltException
    {
	String s = n.toString(con);
	return compare(op, new ValueString(s), con);
    }

    // Num -> NodeList -> boolean
    public boolean compare(int op, ValueNodeList nl, EvalContext con)
	throws QuiltException
    {
	String s = nl.toString(con);
	return compare(op, new ValueString(s), con);
    }
    
/**
 * Insert the method's description here.
 * Creation date: (3/28/2002 7:33:49 PM)
 * @return int
 */
public float getValue() {
	return value;
}
}

