/* Kweelt - a framework to query XML documents
 * Copyright (C) 2000,  Arnaud Sahuguet (Arnaud.Sahuguet@polytechnique.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */



/**
 * DOMUtil.java
 *
 *
 * Created: Tue Jun 27 17:00:32 2000
 *
 * Some useful method to provide services not directly offered by DOM:
 * - ancestors
 * - descendants
 * - etc.
 * The other methods are simply rerouted to DOM.
 *
 * Just use the methods instead of DOM
 *
 *
 * @author Arnaud Sahuguet
 * @version
 */

package xacute.util;

import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;

import java.util.Hashtable;
import java.util.Vector;

import xacute.common.Hint;


public class DOMUtil
{

    public static NodeList reverse( NodeList nl )
    {
	DOMNodeList result = new DOMNodeList();
	for(int i=nl.getLength()-1; i >= 0; i--)
	    result.addElement(nl.item(i));
	return result;
    }

    public static NodeList getSelf( Node n )
    {
	DOMNodeList result = new DOMNodeList();
	result.addElement(n);
	return result;
    }    

    public static NodeList getAttributes( Node n )
    {
	NamedNodeMap nodeMap = n.getAttributes();
	DOMNodeList result = new DOMNodeList();
	for(int i=0; i<nodeMap.getLength(); i++)
	    result.addElement( nodeMap.item(i) );
	return result;
    }
    
    public static NodeList getChildNodes( Node n )
    {
	return n.getChildNodes();
    }
    
    public static NodeList getAncestors( Node n )
    {
	Node current = n;
	DOMNodeList nl = new DOMNodeList();
	while( (current = current.getParentNode()) != null)
	    {
		nl.addElement( current );
	    }
	return nl;
    }

    public static NodeList getAncestorsOrSelf( Node n )
    {
	Node current = n;
	DOMNodeList nl = new DOMNodeList();
	nl.addElement(current);
	while( (current = current.getParentNode()) != null)
	    {
		nl.addElement( current );
	    }
	return nl;
    }

    public static NodeList getDescendants( Node n )
    {
	DOMNodeList result = new DOMNodeList();
	NodeList children = n.getChildNodes();
	for(int i=0; i<children.getLength(); i++)
	    {
		result.addElement(children.item(i));
		result.append( getDescendants(children.item(i)) );
	    }
	return result;
    }
    
    public static NodeList getDescendantsOrSelf( Node n )
    {
	DOMNodeList result = new DOMNodeList();
	result.addElement(n);
	result.append( getDescendants(n) );
	return result;
    }

    
    /* the sibling, in reverse document order */
    public static NodeList getPrecedingSiblings( Node n )
    {
	Node cur = n;
	DOMNodeList result = new DOMNodeList();
	while( (cur = cur.getPreviousSibling()) != null)
	    result.addElement(cur);
	return result;
    }
    
    /* the sibling, in reverse document order */
    public static NodeList getFollowingSiblings( Node n )
    {
	Node cur = n;
	DOMNodeList result = new DOMNodeList();
	while( (cur = cur.getNextSibling()) != null)
	    result.addElement(cur);
	return result;
    }
    
    /* all that follows but is not a descendant */
    public static NodeList getFollowing( Node n )
    {
	DOMNodeList result = new DOMNodeList();
	NodeList ancestorsOrSelf = getAncestorsOrSelf(n);
	for(int i=0;i<ancestorsOrSelf.getLength(); i++)
	    {
		NodeList followingSiblings = getFollowingSiblings( ancestorsOrSelf.item(i) );
		for(int j=0; j<followingSiblings.getLength(); j++)
		    result.append( getDescendantsOrSelf( followingSiblings.item(j) ));
	    }
	return result;
    }

    public static NodeList getParent( Node n )
    {
	DOMNodeList result = new DOMNodeList();
	if (n.getParentNode() != null) result.addElement( n.getParentNode() );
	return result;
    }
    
    /* all that precedes but is not an ancestor */
    public static NodeList getPreceding( Node n )
    {
	DOMNodeList result = new DOMNodeList();
	NodeList ancestorsOrSelf = getAncestorsOrSelf(n);
	for(int i=0;i<ancestorsOrSelf.getLength(); i++)
	    {
		NodeList precedingSiblings = getPrecedingSiblings( ancestorsOrSelf.item(i) );
		for(int j=0; j<precedingSiblings.getLength(); j++)
		    result.append( getDescendantsOrSelf( precedingSiblings.item(j) ));
	    }
	return result;
    }


    public static String getNodeName( Node n )
    {
	return n.getNodeName();
    }

    public static String getNodeValue( Node n )
    {
	return n.getNodeValue();
    }
    

    public static Node cloneNode( Node n )
    {
	return n.cloneNode(true);
    }

    public static boolean isElementNode( Node n )
    {
	return (n.getNodeType() == org.w3c.dom.Node.ELEMENT_NODE);
    }

    public static boolean isAttributeNode( Node n )
    {
	return (n.getNodeType() == org.w3c.dom.Node.ATTRIBUTE_NODE);
    }
    
    public static boolean isTextNode( Node n )
    {
	return (n.getNodeType() == org.w3c.dom.Node.TEXT_NODE);
    }

    public static boolean isPINode( Node n )
    {
	return (n.getNodeType() == org.w3c.dom.Node.PROCESSING_INSTRUCTION_NODE);
    }

    public static boolean isCommentNode( Node n )
    {
	return (n.getNodeType() == org.w3c.dom.Node.COMMENT_NODE);
    }

    public static boolean isRootNode( Node n )
    {
	return (
		(n.getNodeType() != org.w3c.dom.Node.ATTRIBUTE_NODE)
		&&
		(n.getParentNode() == null)
		);
    }
    
    /* two nodes are comparable if they belong to the same document */
    public static boolean canCompareNodes( Node n1, Node n2 )
    {
	return (n1.getOwnerDocument() == n2.getOwnerDocument());
    }

    public final static int AFTER = 1;
    public final static int BEFORE = -1;
    public final static int SAME = 0;


    /* make it more efficient: do not use reverse */
    public static int compareNodes( Node n1, Node n2 )
    {
	if(n1==n2) return SAME;
	if (isAttributeNode(n1))
	    {
		Node parent = n1.getParentNode();
		// System.out.println(n1 + " and its parent " + parent);
		return compareNodes( parent, n2);
	    }
	if (isAttributeNode(n2))
	    {
		Node parent = n2.getParentNode();
		// System.out.println(n2 + " and its parent " + parent);
		return compareNodes(n1, parent);
	    }
	
			       
	NodeList n1l = getAncestorsOrSelf(n1);
	NodeList n2l = getAncestorsOrSelf(n2);

	boolean DEBUG = false;
	if (DEBUG)
            {
                for(int i=0; i < n1l.getLength(); i++)
                    System.out.println( n1l.item(i) + " ");
                System.out.println();
                for(int i=0; i < n2l.getLength(); i++)
                    System.out.println( n2l.item(i) + " ");
		System.out.println("--------------------");
            }

	
	int i;
	int s1 = n1l.getLength();
	int s2 = n2l.getLength();
	for(i=0; i < Math.min(s1, s2); i++)
	    if (!n1l.item(s1-1-i).equals(n2l.item(s2-1-i))) break;
	i--;
	
	// s-1-i is the node first node that is common
	// s-1-i == 0 => subtree
	
	if (i == s1-1)
	    return BEFORE;
	    
	if (i == s2-1)
	    return AFTER;
	
	
	
	int i1 = getChildIndex(n1l.item(s1-1-i),n1l.item(s1-1-i-1));
	int i2 = getChildIndex(n2l.item(s2-1-i),n2l.item(s2-1-i-1));

		
	if ((i1-i2)>0)
	    return AFTER;
	if ((i1-i2)<0)
	    return BEFORE;
	return SAME;
	
    }

    private static int getChildIndex( Node parent, Node n )
    {
	NodeList children = parent.getChildNodes();
	for(int i=0; i<children.getLength(); i++)
	    if (children.item(i)==n) return i;
	return -1;
    }

    private static void printAll(NodeList nl)
    {
	System.out.println("All:");
	for(int i=0; i<nl.getLength(); i++)
	    System.out.println( nl.item(i).getNodeName() );
    }

    public static Node getDocumentRoot( Node n )
    {
	return n.getOwnerDocument();
    }

    // could be improved a lot.
    // once we have found a match for a given id, we can stop looking for this id
    // onc we have a found a match for all ids, we can stop.
    public static NodeList getElementsByID( Node n, String[] ids, Hint[] hints )
    {
	// System.err.println( n.getClass().getName() );
	// System.err.println( n.getOwnerDocument().getDocumentElement().getClass().getName() );
	Element root = (Element) n.getOwnerDocument().getDocumentElement();
	NodeList nl = null;
	Hashtable table = new Hashtable();
	for(int i=0; i < hints.length; i++)
	    {
		//System.err.println("Looking for elements " + hints[i].getElt()
		//		   + " with attribute " + hints[i].getAtt() );
		nl = root.getElementsByTagName( hints[i].getElt() );
		for(int j=0; j < nl.getLength(); j++)
		    {
			Element elt = (Element) nl.item(j);
			for(int k=0; k < ids.length; k++)
			    if ( ids[k].compareTo( elt.getAttribute( hints[i].getAtt() ) ) == 0 )
				table.put( ids[k], elt );
		    }
	    }
	DOMNodeList result = new DOMNodeList();
	for(int k=0; k < ids.length; k++)
	    if (table.containsKey(ids[k])) result.addElement( (Node) table.get(ids[k]) );
	return result;
    }
        
} // DOMUtil




