/* Kweelt - a framework to query XML documents
 * Copyright (C) 2000,  Arnaud Sahuguet (Arnaud.Sahuguet@polytechnique.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


/**
 *
 * This is an implementation of the xacute.common.Node interface that
 * relies on DOM.
 *
 **/

package xacute.util;

import xacute.common.Constants;
import xacute.common.ExpandedName;
import xacute.common.Hint;
import xacute.common.Node;
import xacute.common.NodeList;
import xacute.common.NodeTest;
import xacute.common.Predicate;
import xacute.common.QuiltException;
import xacute.util.DOMUtil;

import java.util.StringTokenizer;
import java.util.Vector;

import org.xml.sax.DocumentHandler;
import org.xml.sax.helpers.AttributeListImpl;


public abstract class NodeDefaultImpl
    implements Node
{

    /* abstract classes that need to be implemented by subclasses */
    public abstract NodeList createNodeList( org.w3c.dom.NodeList nl );
    
    public abstract NodeList createNodeList( org.w3c.dom.Node n );
    
    public abstract NodeList createNodeList( Node n );
    
    public abstract NodeList createNodeList();

    public abstract Node createNode( org.w3c.dom.Node n );
    
    /* what the class offers */

    /* the underlying DOM node */
    private org.w3c.dom.Node domNode;

    protected org.w3c.dom.Node getDOMNode() { return domNode; }
    
    /* constructor for package classes ONLY */
    public NodeDefaultImpl(org.w3c.dom.Node domNode)
    {
	this.domNode = domNode;
    }

    /**
     * Compares two nodes for physical equality.
     **/
    public boolean equals(Node n)
    {
	/* different implementations => false */
	if (!(n instanceof NodeDefaultImpl)) return false;
	
	if (!DOMUtil.canCompareNodes( this.domNode, ((NodeDefaultImpl)n).getDOMNode() ))
	    return false;

	return (DOMUtil.compareNodes(this.domNode, ((NodeDefaultImpl)n).getDOMNode() )==0);
    }

    public int compareTo(Node n)
    {
        try
            {
                org.w3c.dom.Node n1 = this.getDOMNode();
                org.w3c.dom.Node n2 = ((NodeDefaultImpl)n).getDOMNode();
                return DOMUtil.compareNodes(n1, n2);
                
            }
        catch(Exception e)
            {
                throw new RuntimeException("In xacute.util.NodeDefaultImpl.compareTo: " + e.toString());
            }
    }

    /**
     *
     * Can be overridden by subclasses.
     * For instance, should be redefined for shallow nodes.
     *
     **/
    protected NodeList findAxis(int axis)
    {
	switch(axis)
	    {
	    case Constants.ANCESTOR_AXIS:
		return createNodeList( DOMUtil.getAncestors(this.domNode) );
	    case Constants.ANCESTOR_SELF_AXIS:
		return createNodeList( DOMUtil.getAncestorsOrSelf(this.domNode) );
	    case Constants.ATTRIBUTE_AXIS:
		return createNodeList( DOMUtil.getAttributes(this.domNode) );
	    case Constants.CHILD_AXIS:
		return createNodeList( DOMUtil.getChildNodes(this.domNode) );
	    case Constants.DESCENDANT_AXIS:
		return createNodeList( DOMUtil.getDescendants(this.domNode) );
	    case Constants.DESCENDANT_SELF_AXIS:
		return createNodeList( DOMUtil.getDescendantsOrSelf(this.domNode) );
	    case Constants.PRECEDING_AXIS:
		return createNodeList( DOMUtil.getPreceding(this.domNode) );
	    case Constants.FOLLOWING_AXIS:
		return createNodeList( DOMUtil.getFollowing(this.domNode) );
	    case Constants.FOLLOWING_SIBLING_AXIS:
		return createNodeList( DOMUtil.getFollowingSiblings(this.domNode) );
	    case Constants.PRECEDING_SIBLING_AXIS:
		return createNodeList( DOMUtil.getPrecedingSiblings(this.domNode) );
	    case Constants.PARENT_AXIS:
		return createNodeList( DOMUtil.getParent(this.domNode) );
	    case Constants.SELF_AXIS:
		return createNodeList( DOMUtil.getSelf(this.domNode) );
	    case Constants.NAMESPACE_AXIS:
		throw new RuntimeException("class giza.domimpl.NodeImpl: not yet implemented");
	    default:
		return createNodeList();
	    }
    }

    /* 
       We get the nodes from the axis and then we apply the test.
       Could be optimized by pushing down the test?
     */
    public NodeList findAxis(int axis, NodeTest test)
    {
	NodeList tmpResult = this.findAxis(axis);
	if (test == null)
	    return tmpResult;
	NodeList result = createNodeList();
	for(int i=0; i < tmpResult.size(); i++)
	    if (test.isTrue(tmpResult.get(i), axis))
		result.APPEND(tmpResult.get(i));
	return result;
    }

    public NodeList findAxis(int axis, NodeTest test, Predicate[] predicates)
	throws QuiltException
    {
	NodeList result = this.findAxis( axis, test );
	for(int i=0; i < predicates.length; i++)
	    result = result.applyPredicate(predicates[i]);
	return result;
    }
    
    public Node getDocumentRoot()
    {
	return createNode( DOMUtil.getDocumentRoot(this.domNode) );
    }
    
    public String getName()
    {
	return DOMUtil.getNodeName(this.domNode);
    }

    public ExpandedName expand(String qName)
    throws QuiltException
    {
	throw new RuntimeException("xacute.impl.giza.NodeImpl.expand() NYI");
    }

    public ExpandedName getExpandedName()
    {
	throw new RuntimeException("xacute.impl.giza.NodeImpl.getQName() NYI");
    }
    
    public String getQName()
    {
	return getName();
    }


    
    public boolean isOfPrincipalType(int axis)
    {
	switch(axis)
	    {
	    case Constants.ATTRIBUTE_AXIS:
		return this.isOfType(Constants.ATTRIBUTE_NODE);
	    case Constants.NAMESPACE_AXIS:
		return this.isOfType(Constants.NAMESPACE_NODE);
	    default:
		return this.isOfType(Constants.ELEMENT_NODE);
	    }
    }



    /** maybe we should have a method that returns the type instead **/
    public boolean isOfType(int type)
    {
	switch(type)
	    {
	    case Constants.NODE_TYPE:
		return true;
	    case Constants.ROOT_NODE:
		return DOMUtil.isRootNode(this.domNode);
	    case Constants.ELEMENT_NODE:
		return DOMUtil.isElementNode(this.domNode);
	    case Constants.TEXT_NODE:
		return DOMUtil.isTextNode(this.domNode);
	    case Constants.ATTRIBUTE_NODE:
		return DOMUtil.isAttributeNode(this.domNode);
	    default:
		throw new RuntimeException("Other node types NYI");
	    }
    }

    /*
      text      -> text value
      attribute -> value
      element   -> concatenation of text values of descendants
    */

    private String stringValue(boolean deep)
    {
	if (this.isOfType(Constants.ELEMENT_NODE) && deep)
	    {
		StringBuffer buf = new StringBuffer();
		NodeList nl = this.findAxis(Constants.DESCENDANT_AXIS);
		for(int i=0; i<nl.size(); i++)
		    {
			Node cur = nl.get(i);
			if(cur.isOfType(Constants.TEXT_NODE))
			    buf.append( cur.stringValue() );
		    }
		return buf.toString();
	    }
	if (this.isOfType(Constants.TEXT_NODE))
	    {
		return DOMUtil.getNodeValue(this.domNode);
	    }
	if (this.isOfType(Constants.ATTRIBUTE_NODE))
	    {
		return DOMUtil.getNodeValue(this.domNode);
	    }
	return "giza.domimpl.NodeImpl.stringValue() NYI";
	//	throw new RuntimeException("giza.domimpl.NodeImpl.stringValue() NYI");
    }

    public String stringValue() { return stringValue(true); }
    public String stringValue_shallow() { return stringValue(false); }
    
    
    public NodeList deref(String value, Hint[] hints)
    {
	// IDREF  -> 0 or 1 item
	// IDREFS -> 0 or more item
	// do we preserve order here ?
	String[] ids = split(value);  
	org.w3c.dom.NodeList nl = DOMUtil.getElementsByID( this.domNode, ids, hints );
	NodeList result = createNodeList();
        for(int i=0; i < nl.getLength(); i++)
	    result = result.UNION( createNodeList( nl.item(i) ) );
	return result;
    }

    private final static String[] split(String value)
    {
	StringTokenizer tok = new StringTokenizer(value, " ");
	Vector tmp = new Vector();
	while( tok.hasMoreElements() )
	    tmp.add( tok.nextToken() );
	String[] result = new String[tmp.size()];
	tmp.toArray(result);
	return result;
    }
    
    /*
      Send back the sax event corresponding to a Node.
      Depending on the nature of the node, we do different things:
    */
    private void toSAX( DocumentHandler handler, boolean deep)
	throws org.xml.sax.SAXException
    {
	if (this.isOfType(Constants.ELEMENT_NODE))
	    {
		org.w3c.dom.Node n = this.domNode;
		
		/* compute attributes */
		NodeList atts = this.findAxis(Constants.ATTRIBUTE_AXIS);
		AttributeListImpl attributes = new AttributeListImpl();
		for(int i=0; i < atts.size(); i++)
		    {
			org.w3c.dom.Node attNode = ((NodeDefaultImpl)atts.get(i)).getDOMNode();
			attributes.addAttribute( DOMUtil.getNodeName(attNode),
						 "",
						 DOMUtil.getNodeValue(attNode));
		    }
		/* open tag */
		handler.startElement( DOMUtil.getNodeName(n), attributes );
		
		/* process the sub-tree */
		if (deep)
		    {
			NodeList nl = this.findAxis(Constants.CHILD_AXIS);
			for(int i=0; i<nl.size(); i++)
			    nl.get(i).toSAX( handler );
		    }
		/* close the tag */
		handler.endElement( DOMUtil.getNodeName(n) );
		return;
	    }
	if (this.isOfType(Constants.TEXT_NODE))
	    {
		char[] value = DOMUtil.getNodeValue(this.domNode).toCharArray();
		handler.characters(value, 0, value.length);
		return;
	    }
	if (this.isOfType(Constants.ATTRIBUTE_NODE))
	    {
		char[] value = DOMUtil.getNodeValue(this.domNode).toCharArray();
		handler.characters(value, 0, value.length);
		return;
	    }
	char[] value = new String("giza.domimpl.NodeImpl.trigger() NYI").toCharArray();
	handler.characters(value, 0, value.length);
	return;
    }

    public void toSAX(DocumentHandler handler)
    	throws org.xml.sax.SAXException
    {
	toSAX( handler, true );
    }

    public void toSAX_shallow(DocumentHandler handler)
    	throws org.xml.sax.SAXException
    {
	toSAX( handler, false );
    }
    
    
} // NodeImpl

   

