/* Kweelt - a framework to query XML documents
 * Copyright (C) 2000,  Arnaud Sahuguet (Arnaud.Sahuguet@polytechnique.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package xacute.util;

import xacute.common.Constants;
import xacute.common.Node;
import xacute.common.NodeList;
import xacute.common.Predicate;
import xacute.common.QuiltException;

import java.util.AbstractList;
import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Arrays;

public /*abstract*/ class NodeListDefaultImpl
    implements NodeList
{
    protected AbstractList list;

    /**
     * These are the only required subclass-specific methods: the
     * subclass is responsible for allocating the object that will
     * play the role of the abstrat list.  One can also redefine all
     * the methods below if one thinks they can be implemented in a
     * better, for a given XML back-end.
     **/

    /*
    protected abstract void init();
    protected abstract NodeListDefaultImpl newInstance();
    */

    protected void init()
    {
	this.list = new ArrayList();
    }
    
    protected NodeListDefaultImpl newInstance()
    {
	return new NodeListDefaultImpl();
    }
    
    protected NodeListDefaultImpl() { init(); }

    // List operations
    public Node get(int k) { return (Node) this.list.get(k); }
    
    public boolean isEmpty() { return this.list.isEmpty(); }

    public int size() {	return this.list.size(); }

    /**
     *
     * stringValue of the NodeList (i.e. concatenation of Node
     * individual string values).
     *
     **/

    public String stringValue()
    {
	StringBuffer buf = new StringBuffer();
	for(int i=0; i<this.size(); i++)
		buf.append( this.get(i).stringValue() );
	return buf.toString();
    }
    
    public String stringValue_shallow()
    {
	StringBuffer buf = new StringBuffer();
	for(int i=0; i<this.size(); i++)
	    buf.append( this.get(i).stringValue_shallow() );
	return buf.toString();
    }

    // boolean methods
        /**
     * Compare 2 nodelists for equality Meaning: There exists u1 in
     * L1, there exists u2 in L2 ... such that u1.stringValue() =
     * u2.stringValue() If one or both of the the lists are empty,
     * then false.
     *
     * Strategy: We build a hash of one list and we iterate through
     * the other one.  We stop as soon as we find a match.
    **/
    public boolean isEqualTo( NodeList nl )
    {
	if ((this.size()==0) || (nl.size()==0))
	    return false;
	
	NodeList nl1 = this;
	NodeList nl2 = nl;
	
	Hashtable table = new Hashtable();
	for(int i=0; i<nl2.size(); i++)
	    table.put( NodeHelper.normalize(nl2.get(i).stringValue()), Boolean.TRUE);

	for(int i=0; i<nl1.size(); i++)
	    if (table.containsKey( NodeHelper.normalize(nl1.get(i).stringValue()) ) )
		return true;

	return false;
    }

    /**
     * Meaning: There exists u1 in L1, there exists u2 in L2 ... such
     * that u1.stringValue() != u2.stringValue()
     *
     * Strategy: We build the two lists of string values.  We sort
     * them and look for differences
     **/
    public boolean isDifferentFrom( NodeList nl )
    {
	if ((this.size()==0) || (nl.size()==0))
	    return false;

	NodeList nl1 = this;
	NodeList nl2 = nl;

	String[] lv1 = new String[nl1.size()];
	for(int i=0; i<nl1.size(); i++)
	    lv1[i] = NodeHelper.normalize(nl1.get(i).stringValue());
	Arrays.sort(lv1);

	String[] lv2 = new String[nl2.size()];
	for(int i=0; i<nl2.size(); i++)
	    lv2[i] = NodeHelper.normalize(nl2.get(i).stringValue());
	Arrays.sort(lv2);

	int i1 = 0;
	int i2 = 0;
	while ( (i1<lv1.length) && (i2<lv2.length) )
	    {
		String s1 = lv1[i1]; String s2 = lv2[i2];
		if (!s1.equals(s2)) return true;
		i1++; i2++;
		int i = i1;
		while ((i1<lv1.length) && (lv1[i1].equals(s1))) i1++;
		while ((i2<lv2.length) && (lv2[i2].equals(s2))) i2++;
	    }
	return false;
    }
    
    
    // structural methods

    /**
     *
     * add
     *
     * THIS IS THE ONLY METHOD THAT DOES NOT REQUIRE TO GET THE RESULT.
     * CAN BE USED ONLY BY SUBCLASSES THOUGH.
     *
     **/
    protected NodeListDefaultImpl add( Node n )
    {
	this.list.add(n);
	return this;
    }

    public NodeList APPEND( Node n )
    {
	this.list.add(n);
	return this;
    }

    public NodeList APPEND( NodeList nl )
    {
	for(int i=0; i < nl.size(); i++)
	    this.list.add(nl.get(i));
	return this;
    }
    
    /**
     *
     * Apply predicates -- one at a time -- on each element of the
     * nodelist.  Returns a new nodelist as a result.
     *
     * THE RESULT IS RETURNED. DO NOT COUNT ON SIDE-EFFECTS.
     *
     **/
    public NodeList applyPredicate( Predicate p )
	throws QuiltException
    {
	NodeListDefaultImpl result = this.newInstance();
	int size = this.size();
	for(int i=0; i<size; i++)
	    if (p.isTrue( this.get(i), i, size)) 
		result.add(this.get(i));
	return result;
    }

    /**
     *
     * Both node lists NL1 and NL2 are sorted. If get(i) in NL1 is in
     * the result, then get(i+1) is also in the result. Once we have
     * found a first match, we know that the remaining items will also
     * be part of the result.
     *
     * An exception is raised when the nodelists cannot be compared
     * (e.g. nodes from different documents).
     *
     * THE RESULT IS RETURNED. DO NOT COUNT ON SIDE-EFFECTS.
     *
     *
     **/
    public NodeList AFTER( NodeList nl)
	throws QuiltException
    {
	NodeList nl1 = this;
	NodeList nl2 = nl;
	NodeListDefaultImpl result = this.newInstance();
	boolean fillTheRest = false;
	
	int i1=0;
	try
	    {
		for(; i1<nl1.size(); i1++)
		    {
			Node n1 = nl1.get(i1);
			for(int i2=0; i2<nl2.size(); i2++)
			    {
				Node n2 = nl2.get(i2);
				if (n1.compareTo(n2) == Constants.AFTER)
				    throw new JobDoneException();
			    }
		    }
	    }
	catch(JobDoneException e)
	    {
		for(int i=i1; i<nl1.size(); i++)
		    result.add(nl1.get(i));
	    }
	return result;
    }
    
    /**
     *
     * Both node lists NL1 and NL2 are sorted. If get(i) in NL1 is in
     * the result, then get(i-1) is also in the result. Once we have
     * found a first match, we know that the remaining items will also
     * be part of the result (assuming we iterate in reverse order).
     *
     *
     * THE RESULT IS RETURNED. DO NOT COUNT ON SIDE-EFFECTS.
     *
     **/
    public NodeList BEFORE( NodeList nl)
	throws QuiltException
    {
	NodeList nl1 = this;
	NodeList nl2 = nl;
	NodeListDefaultImpl result = this.newInstance();
	NodeListDefaultImpl tmp = this.newInstance();
	boolean fillTheRest = false;
	
	int i1 = nl1.size()-1;
	try
	    {
		for(; i1 >= 0; i1--)
		    {
			Node n1 = nl1.get(i1);
			for(int i2=nl2.size()-1; i2 >= 0; i2--)
			    {
				Node n2 = nl2.get(i2);
				if (n1.compareTo(n2)==Constants.BEFORE)
				    throw new JobDoneException();
			    }
		    }
	    }
	catch(JobDoneException e)
	    {
		for(int i=i1; i >= 0; i--)
		    result.add(nl1.get(i));
	    }
	for(int i=result.size()-1; i >= 0; i--)
	    tmp = tmp.add( result.get(i) );
	return tmp;
    }

    /**
     * Like difference.
     *
     *
     * THE RESULT IS RETURNED. DO NOT COUNT ON SIDE-EFFECTS.
     *
     **/
    public NodeList EXCEPT( NodeList nl)
	throws QuiltException
    {
	NodeList nl1 = this;
	NodeList nl2 = nl;
	NodeListDefaultImpl result = this.newInstance();
	int i1 = 0;
	int i2 = 0;
	while ( (i1<nl1.size()) && (i2<nl2.size()))
	    {
		Node n1 = nl1.get(i1);
		Node n2 = nl2.get(i2);
		int comparison = n1.compareTo(n2);
		switch (comparison)
		    {
		    case Constants.BEFORE:
			result.add(nl1.get(i1)); i1++; break;
		    case Constants.AFTER:
			i2++; break;
		    case Constants.SAME:
			i1++; i2++; break;
		    }
	    }
	return result;
    }
    
    /**
     * We assume both lists to be sorted.
     *
     * THE RESULT IS RETURNED. DO NOT COUNT ON SIDE-EFFECTS.
     *
     **/
    public NodeList INTER( NodeList nl)
	throws QuiltException
    {
	NodeList nl1 = this;
	NodeList nl2 = nl;
	NodeListDefaultImpl result = this.newInstance();
	int i1 = 0;
	int i2 = 0;
	while ( (i1<nl1.size()) && (i2<nl2.size()))
	    {
		Node n1 = nl1.get(i1);
		Node n2 = nl2.get(i2);
		int comparison = n1.compareTo(n2);
		switch (comparison)
		    {
		    case Constants.BEFORE:
			i1++; break;
		    case Constants.AFTER:
			i2++; break;
		    case Constants.SAME:
			result.add(nl1.get(i1)); i1++; i2++; break;
		    }
	    }
	return result;
    }
    
    /**
     *
     * Physical duplicates are removed
     *
     * THE RESULT IS RETURNED. DO NOT COUNT ON SIDE-EFFECTS.
     *
     **/


    /*
    public NodeList UNION( NodeList nl)
    {
	return this.APPEND(nl);
    }
    */
    
    /* NEEDS TO BE FIXED */
    public NodeList UNION( NodeList nl)
    {
	try
	    {
		NodeList nl1 = this;
		NodeList nl2 = nl;
		NodeListDefaultImpl result = this.newInstance();
		int i1 = 0;
		int i2 = 0;
		while ( (i1 < nl1.size()) && (i2 < nl2.size()))
		    {
			Node n1 = nl1.get(i1);
			Node n2 = nl2.get(i2);
			int comparison = n1.compareTo(n2);
			switch (comparison)
			    {
			    case Constants.BEFORE:
				result.add(nl1.get(i1)); i1++;
				break;
			    case Constants.AFTER:
				result.add(nl2.get(i2)); i2++;
				break;
			    case Constants.SAME:
				result.add(nl1.get(i1)); i1++; i2++;
				break;
			    default:
				throw new RuntimeException("invalid comparison result" + comparison);
			    }
		    }
		if (i1==nl1.size())
		    for(int i=i2; i < nl2.size(); i++)
			result.add( nl2.get(i));
		if (i2==nl2.size())
		    for(int i=i1; i < nl1.size(); i++)
			result.add( nl1.get(i));
		return result;
	    }
	catch(Exception e)
	    {
		System.err.println("Ordered Union created an exception " + e);
		System.err.println("Unordered Union used instead " + e);
		
		return this.APPEND(nl);
	    }
    }
    
    
    /**
     *
     * removes value-based duplicates (stringValue())
     *
     * THE RESULT IS RETURNED. DO NOT COUNT ON SIDE-EFFECTS.	  
     *
     **/
    public NodeList freeDuplicateValues()
    {
	Hashtable table = new Hashtable();
	NodeListDefaultImpl result = this.newInstance();
	for(int i=0; i<this.size(); i++)
	    {
		String sValue = this.get(i).stringValue();
		if (!table.containsKey(sValue))
		    result.add(this.get(i));
		table.put(sValue, Boolean.TRUE);
	    }
	return result;
    }
  

} // NodeListDefaultImpl

