/* Kweelt - a framework to query XML documents
 * Copyright (C) 2000,  Arnaud Sahuguet (Arnaud.Sahuguet@polytechnique.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package xacute.util;

import xacute.common.Constants;
import xacute.common.ExpandedName;
import xacute.common.Hint;
import xacute.common.Node;
import xacute.common.NodeTest;
import xacute.common.NodeList;
import xacute.common.Predicate;
import xacute.common.QuiltException;

import org.xml.sax.DocumentHandler;
import org.xml.sax.AttributeList;
import org.xml.sax.helpers.AttributeListImpl;

/**
 * A <code>QuiltElementNode</code> is the result an element creation in a <code>RETURN</code> clause.
 * It has children and attributes.
 *
 * Like for other Quilt nodes, we restrict navigation.
 *
 **/

public class QuiltElementNode
    implements Node
{
    private String tagName;
    private AttributeList atts;
    private Node[] childNodes;

    public QuiltElementNode(String tagName, AttributeList atts, Node[] childNodes)
    {
	this.tagName = tagName;
	this.atts = atts;
	this.childNodes = childNodes;
    }

    public boolean equals(Node n)
    {
	return (this == n);
    }

    public int compareTo(Node n)
    {
	// this is an arbitrary choice
	if (n instanceof QuiltElementNode)
	    return Constants.AFTER;
	return Constants.NOT_COMPARABLE;
    }
    
    public NodeList findAxis( int axis, NodeTest test)
    {
	NodeList nl = new NodeListDefaultImpl();
        switch (axis)
	    {
	    case Constants.ANCESTOR_AXIS:
		return nl;
	    case Constants.CHILD_AXIS:
		for(int i=0; i < childNodes.length; i++)
		    if (test.isTrue( childNodes[i], axis ))
			       nl = nl.APPEND( childNodes[i] );
		return nl;
	    case Constants.DESCENDANT_AXIS:
	    case Constants.FOLLOWING_AXIS:
	    case Constants.PRECEDING_AXIS:
	    case Constants.FOLLOWING_SIBLING_AXIS:
	    case Constants.PRECEDING_SIBLING_AXIS:
		throw new RuntimeException("xacute.util.QuiltTextNode.findAxis(): NYI");
	    case Constants.PARENT_AXIS:
		return nl;
	    case Constants.ATTRIBUTE_AXIS:
		for(int i=0; i < atts.getLength(); i++)
		    nl = nl.APPEND( new QuiltAttributeNode( atts.getName(i), atts.getValue(i), this ) );
		break;
	    case Constants.NAMESPACE_AXIS:
		throw new RuntimeException("xacute.util.QuiltElementNode.findAxis(): invalid navigation");
	    case Constants.ANCESTOR_SELF_AXIS:
		throw new RuntimeException("xacute.util.QuiltElementNode.findAxis(): invalid navigation");
	    case Constants.DESCENDANT_SELF_AXIS:
		nl = nl.APPEND(this);
		for(int i=0; i < childNodes.length; i++)
		    nl = nl.APPEND( childNodes[i].findAxis( axis, test) );
		break;
	    case Constants.SELF_AXIS:
		nl = nl.APPEND(this);
		break;
	    }
	NodeList result = new NodeListDefaultImpl();
	for(int i=0; i < nl.size(); i++)
	    if (test.isTrue(nl.get(i), axis))
		result = result.APPEND(nl.get(i));
	return result;
    }

    public NodeList findAxis(int axis, NodeTest nt, Predicate[] predicates)
    	throws QuiltException
    {
	NodeList result = this.findAxis(axis, nt);
	for(int i=0; i < predicates.length; i++)
	    result = result.applyPredicate(predicates[i]);
	return result;
    }

    public Node getDocumentRoot()
	throws QuiltException
    {
	throw new QuiltException("A QuiltTextNode does not belong to any document");
    }
    
    public ExpandedName getExpandedName()
    	throws QuiltException
    {
	return new ExpandedName( this.tagName );
    }

    public ExpandedName expand(String qName)
	throws QuiltException
    {
	return new ExpandedName( this.tagName );
    }
    
    public String getQName()
    {
	return this.tagName;
    }
    
    public boolean isOfPrincipalType(int axis)
    {
	if (axis == Constants.ATTRIBUTE_AXIS) return false;
	if (axis == Constants.NAMESPACE_AXIS) return false;
	return true;
    }
    
    public boolean isOfType(int type)
    {
	return ( (type == Constants.ELEMENT_NODE)
		 ||
		 (type == Constants.NODE_TYPE)
		 );
    }
    
    public String stringValue()
    {
	StringBuffer buf = new StringBuffer();
	for(int i=0; i < childNodes.length; i++)
	    buf.append( childNodes[i].stringValue() );
	return buf.toString();
    }

    public String stringValue_shallow()
    {
	return "";
    }

    
    public NodeList deref(String idref, Hint[] hints) { return new NodeListDefaultImpl(); }

    
    public void toSAX(DocumentHandler handler)
	throws org.xml.sax.SAXException
    {
	handler.startElement(this.tagName, this.atts);
	for(int i=0; i < childNodes.length; i++)
	    {
		childNodes[i].toSAX(handler);
	    }
	handler.endElement(this.tagName);
    }

    public void toSAX_shallow(DocumentHandler handler)
	throws org.xml.sax.SAXException
    {
	handler.startElement(this.tagName, this.atts);
	handler.endElement(this.tagName);
    }

}





